/* ***** BEGIN LICENSE BLOCK *****
   - Version: MPL 1.1/GPL 2.0/LGPL 2.1
   -
   - The contents of this file are subject to the Mozilla Public License Version
   - 1.1 (the "License"); you may not use this file except in compliance with
   - the License. You may obtain a copy of the License at
   - http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS IS" basis,
   - WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
   - for the specific language governing rights and limitations under the
   - License.
   -
   - The Original Code is CheckYesss.
   -
   - The Initial Developer of the Original Code is
   - Thomas Reitmayr <treitmayr@yahoo.com>.
   - Portions created by the Initial Developer are Copyright (C) 2005-2008
   - the Initial Developer. All Rights Reserved.
   -
   - Contributor(s):
   - -
   -
   - Alternatively, the contents of this file may be used under the terms of
   - either the GNU General Public License Version 2 or later (the "GPL"), or
   - the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
   - in which case the provisions of the GPL or the LGPL are applicable instead
   - of those above. If you wish to allow use of your version of this file only
   - under the terms of either the GPL or the LGPL, and not to allow others to
   - use your version of this file under the terms of the MPL, indicate your
   - decision by deleting the provisions above and replace them with the notice
   - and other provisions required by the LGPL or the GPL. If you do not delete
   - the provisions above, a recipient may use your version of this file under
   - the terms of any one of the MPL, the GPL or the LGPL.
   -
   - ***** END LICENSE BLOCK ***** */

const CLASS_ID = Components.ID("{e0e24620-22c8-11da-8cd6-0800200c9a66}");
const CLASS_NAME = "Checkyesss YESSS Manager Component";
const CONTRACT_ID = "@mozdev.org/checkyesss/yesss_manager;1";

const TO_HTTP = 120;              // 120 seconds until aborting stalled access
const TO_MIN_PAUSE = 10;          // 10 minutes pause between scheduled accesses
const TO_BLOCKED = 60;            // 60 seconds pause if access was blocked
const TO_MIN_CALLUPD = 1;         // 1 hours min. pause for calls (debug: set to 0!)
const TO_CALLUPD = 7*24;          // 7 days regular interval for calls

const MAX_CARDS = 10;

const DEBUG = 0;


/***************************************************************************/

/**
/* ArrayEnumerator courtesy of Brendan Eich <brendan@mozilla.org>
 */
function ArrayEnumerator(array) {
  this.array = array;
  this.index = 0;
}
ArrayEnumerator.prototype = {
  hasMoreElements: function () { return this.index < this.array.length; },
  getNext: function () { return (this.index < this.array.length) ?
                                this.array[this.index++] : null; }
};
// nsISupportsArrayEnumerator
//
function nsISupportsArrayEnumerator(array) {
  this.array = array;
  this.index = 0;
}
nsISupportsArrayEnumerator.prototype = {
  hasMoreElements: function() { return this.index < this.array.Count(); },
  getNext: function () { return (this.index < this.array.Count()) ?
                       this.array.GetElementAt(this.index++) : null; }
};

/**
/* StringEnumerator
 */
function StringEnumerator(array) {
  this.array = array;
  this.index = 0;
}
StringEnumerator.prototype = {
  hasMore: function() { return this.index < this.array.length; },
  getNext: function () { return (this.index < this.array.length) ?
                                this.array[this.index++] : null; }
};

/* This might make it into some future version...
var loader = Components.classes["@mozilla.org/moz/jssubscript-loader;1"].
                        getService(Components.interfaces.mozIJSSubScriptLoader);
loader.loadSubScript("chrome://checkyesss/content/common/includeArrays.js");
*/

/***************************************************************************/

/**
/* 
 */
function AccountDetails(aId, aUser, aPassword, aPlan, aPeriodReset, aEnabled,
                        aMayStorePwd) {
  this._id = aId;
  this._user = aUser;            // user name (=phone number)
  this._pw = ((typeof(aPassword) == "string") && (aPassword.length > 0)) ?
               aPassword : null;          // password
  this._mayStorePwd = aMayStorePwd;   // may the password be stored in PW-Mng?
  this._periodreset = aPeriodReset; // type of period reset
  this._manualperiod = 0;        // time an accounting period should start
  this._en = aEnabled;           // account enabled
  this._rateplan = aPlan;        // actual plan
  this._actual_rate = 69;        // default 6.9c/min
  this._credit = "";             // current credit ("euro.cent")
  this._feeMinRate = 0;          // fee relevant for the minimum fee
  this._credit_till = "";        // credit valid till
  this._total_duration = 0;
  this._total_sms = 0;
  this._total_calls = 0;
  this._dateCallsLoaded = 0;     // last time calls where updated
  this._error = "Daten noch nicht geladen";
  this._forceupdate = false;     // force update!!
  this._quickLinkPending = false;
  this._quickLinkWindow = null;
  this._quickLinkTarget = "";
  this._reloadAvailableCalls = false;
  this._intl_packet_type = "";
  this._intl_packet_till = 0;
  this._sms_packet_type = "";
  this._sms_packet_till = 0;
}

/******************************************************************************
 * CheckyesssYesssManager component
 ******************************************************************************/

function CheckyesssYesssManager() {
  this.wrappedJSObject = this;
  this._start();
}


CheckyesssYesssManager.prototype = {
  
  _providerId:   "YS001",
  
  _yesssServer:   "www.yesss.at",
  _yesssLogin:    "https://www.yesss.at/kontomanager.php",
  _yesssCData:    "https://www.yesss.at/kontomanager/kundendaten.php",
  _yesssCalls:    "https://www.yesss.at/kontomanager/letzte_gespraeche.php?page=%",
  _yesssDeposits: "https://www.yesss.at/kontomanager/aufladungen.php",
  _yesssLogout:   "https://www.yesss.at/kontomanager/logout.php",
  
  _yesssVCalls:   "https://www.yesss.at/kontomanager_v/gespraeche.php?page=%",
  _yesssVCallsP:  "https://www.yesss.at/kontomanager_v/prepaid.php?page=%",
  _yesssVLogout:  "https://www.yesss.at/kontomanager_v/logout.php",
  
  _yesssWFLogin:  "https://www.yesss.at/webfon/kontomanager.php",
  _yesssWFCalls:  "http://www.ewave.at/cgi-bin/onlineegn3.py",
  
  _yesssILogin:   "https://www.yesss.at/diskont-surfen/kontomanager.php",
  _yesssIConns:   "https://www.yesss.at/diskont-surfen/kontomanager/datentransfers.php?page=%",
  _yesssILogout:  "https://www.yesss.at/diskont-surfen/kontomanager/logout.php",
  
  _blockingUrlRegExp: "^https?://(?:www\\.)?yesss\\.at(?:/\\S+)?/(?:webfon|diskont-surfen/)?kontomanager(?:_v)?/(?!logout\\.php)",
  
  _timeHour:   3600000,
  _timeMinute: 60000,
  _timeSecond: 1000,
  
  _observerTopic: null,
  _userRegExp: null,
  
  _interval: 24,                     // interval [hours]
  _suspendUntil: 0,                  // suspend automatic update from internet
  
  _accountsArray: null,
  _accountsChanged: false,
  
  _webAccess: null,
  _accessAccount: -1,
  _lastAccess: 0,     // date of last access [ms] (not counting logout)
  _checkPending: false,
  
  _firstTime: true,
  
  _checkTimer: null,
  _obs: null,
  _utils: null,
  _userAgent: null,
  _managerCenter: null,
  
  
  /**************************************************************/
  
  _start: function() {
    // set components
    this._utils = Components.classes["@mozdev.org/checkyesss/util_service;1"].
                getService(Components.interfaces.cyICheckyesssUtil);
    
    this._checkTimer = Components.classes["@mozilla.org/timer;1"].
                         createInstance(Components.interfaces.nsITimer); 
    
    this._obs = Components.classes["@mozilla.org/observer-service;1"].
                  getService(Components.interfaces.nsIObserverService);
    
    this._managerCenter = Components.classes["@mozdev.org/checkyesss/manager_center;1"].
                            getService(Components.interfaces.cyICheckyesssCenter);
    
    var providerProperties = this._managerCenter.getProviderProperties(this._providerId);
    this._observerTopic = providerProperties.getPropertyAsString("topic");
    this._userRegExp = providerProperties.getPropertyAsString("validUserRegExp");
    
    //add observer topics
    this._obs.addObserver(this, "xpcom-shutdown", false);
    
    this._webAccess = Components.classes["@mozdev.org/checkyesss/web_access;1"].
                        createInstance(Components.interfaces.cyICheckyesssWeb);
    this._webAccess.requestTimeout = TO_HTTP;
    
    try {
      this._userAgent = Components.classes["@mozilla.org/network/protocol;1?name=http"].
                          getService(Components.interfaces.nsIHttpProtocolHandler).
                            userAgent;
      this._userAgent += " Checkyesss/0.6";
    }
    catch (e) {
      this._userAgent = null;
    }
    
    if (DEBUG) {
      // prepaid
      //this._yesssLogin    = "http://localhost/cy/kontomanager.php.html";
      this._yesssLogin    = "http://localhost/cy/kundendaten.php.html";
      //this._yesssLogin    = "http://localhost/cy/login-fail.php.html";
      this._yesssLogout   = "http://localhost/cy/logout.html";
      //this._yesssCalls    = "http://localhost/cy/letzte_gespraeche%.php.html";
      this._yesssCalls    = "http://localhost/cy/gespraeche%.php.html";
      this._yesssDeposits = "http://localhost/cy/aufladungen.php.html";
/*      // postpaid
      this._yesssLogin    = "http://localhost/cy_v/uebersicht.php.html";
      this._yesssVCalls    = "http://localhost/cy_v/gespraeche%.php.html";
      this._yesssVCallsP   = "http://localhost/cy_v/prepaid%.php.html";
      this._yesssVLogout   = "http://localhost/cy_v/logout.html";
      // webfon
      this._yesssWFLogin  = "http://localhost/wf/kundendaten.php.html";
      this._yesssWFCalls  = "http://localhost/wf/onlineegn3-%.csv";
      // internet
      this._yesssLogin    = "http://localhost/ds/login-fail-phone.html";*/
      this._yesssILogin   = "http://localhost/ds/uebersicht.html";
      this._yesssIConns   = "http://localhost/ds/datentransfers-%.html";
      this._yesssILogout  = "http://localhost/ds/logout.html";
      
      this._timeHour = this._timeMinute;
      this._timeMinute = this._timeSecond;
    }
    
    //this._checkTimer.initWithCallback(this, 10, 
    //                                  Components.interfaces.nsITimer.TYPE_ONE_SHOT);
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                              "Checkyesss YESSS-Manager started", "", false);
    
    this._firstTime = false;
  },
  
  /**************************************************************/
  
  _stop: function() {
    this._webAccess.endSession();
    
    // stop timer
    this._checkTimer.cancel();
    
    // clean up observers
    try {
      this._obs.removeObserver(this, "xpcom-shutdown");
    }
    catch(e) {}
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                              "Checkyesss YESSS-Manager stopped", "", false);
    
    //destroy variables
    this._webAccess = null;
    //this._busy = false;
    this._jumpWin = null;
    this._checkTimer = null;
    this._errtxt = null;
    //this._loginCallback = null;
    this._firstTime = true;
    this._accountsArray = null;
    this._utils = null;
    this._managerCenter = null;
  },
  
  /**************************************************************/
  
  get updateInterval() {
    return this._interval;
  },
  
  set updateInterval(aValue) {
    this._interval = (aValue <  4) ? 4 :
                     (aValue > 24) ? 24 :
                     aValue;
    return aValue;
  },
  
  /**************************************************************/
  
  get suspendNetAccessUntil() {
    return this._suspendUntil;
  },
  
  set suspendNetAccessUntil(aValue) {
    this._suspendUntil = aValue;
    return aValue;
  },
  
  /**************************************************************/
  
  updateClients: function() {
    this._refreshDisplay();
  },
  
  /**************************************************************/
  
  clearAccounts: function() {
    if (!this._accountsArray ||
        (this._accountsArray.length == 0))
      return;
    
    this.abortAccess();
    
    this._accountsArray = new Array();
    
    this._obs.notifyObservers(this, this._observerTopic, "reset_all");
  },

  /**************************************************************/
  
  addAccount: function(aAccountId, aUser, aPassword, aRatePlanId,
                       aPeriodReset, isEnabled, aMayStorePwd) {
    var result = true;
    
    if (!this._accountsArray)
      this._accountsArray = new Array();
    var index = this._accountsArray.length;
    
    var entry = new AccountDetails(aAccountId, aUser, aPassword, aRatePlanId, 
                                   aPeriodReset, isEnabled, aMayStorePwd);
    
    if (entry._rateplan == "auto" && entry._user.match(/^43780/))
      entry._rateplan = "webfon";
    
    var ratePlanOk = false;
    var ratePlanIds = this._managerCenter.getRatePlanList(this._providerId);
    while (!ratePlanOk && ratePlanIds.hasMore())
      ratePlanOk = (ratePlanIds.getNext() == aRatePlanId);
    
    if (!ratePlanOk) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesss.addAccount",
                                "Invalid rate plan ID '" + aRatePlanId + "'",
                                false);
      entry._en = false;
      entry._rateplan = "auto";
      result = false;
    }
    
    this._accountsArray[index] = entry;
    
    if (entry._en) {
      this._updateFromCache(index);
      
      // restart timer with a short timeout
      this._checkTimer.cancel();
      this._checkTimer.initWithCallback(this, 100, 
                                        Components.interfaces.nsITimer.TYPE_ONE_SHOT);
    }
    return result;
  },
  
  /**************************************************************/
  
  _assembleCreditInfo: function(aIndex, aDocument) {
    var entry;
    try {
      entry = this._accountsArray[aIndex];
    }
    catch (e) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "_assembleCreditInfo",
                                "Invalid internal index " + aIndex, false);
      return null;
    }
    
    if (!entry._en)
      return null;
    
    var expired = true;
    var age = -1;
    
    if (this._utils.existCache(entry._id)) {
      age = this._utils.getCacheAge(entry._id);
      expired = (this._utils.getCacheValidity(entry._id) == 0);
    }
    
    var container = Components.classes["@mozdev.org/checkyesss/container;1"].
                      createInstance(Components.interfaces.cyICheckyesssContainer);
    
    var tooltip = null;
    var pnum = entry._user;
    var maxDynColor;
    var pkdaydiff = -100;
    var addText;
    
    if (entry._rateplan == "webfon") {
      pnum = pnum.replace(/^43780/, "0780/");
    }
    else
    if (entry._rateplan == "wertkarte" || entry._rateplan == "V" ||
        entry._rateplan == "internet") {
      var matches = pnum.match(/^\s*(0043|\+43|0)(699|664|650|676|720|780|660|688|681|680)([0-9]+)\s*$/);
      if (matches)
        pnum = "0" + matches[2] + "/" + matches[3];
    }
    else {
      // the rate plan was not detected yet!
      if (!entry._error)
        entry._error = "Tarif wurde noch nicht erkannt,\nBitte den Betrag aktualisieren!";
    }
    
    var isError = (entry._error || (entry._credit.toString().length == 0)) ? true : false;
    
    if (!isError && (entry._rateplan == "wertkarte") && entry._intl_packet_type) {
      var now = (new Date()).valueOf();
      pkdaydiff = (entry._intl_packet_till - now) / (24 * 3600000);
      pkdaydiff += (pkdaydiff > 0) ? +1 : -1;
      pkdaydiff = parseInt(pkdaydiff, 10);
      if (pkdaydiff >= -7) {
        maxDynColor = (pkdaydiff <= 1) ? "low" :
                      (pkdaydiff <= 3) ? "ok" : "sup";
        addText = (pkdaydiff < 1) ? "0" : pkdaydiff.toString();
        addText = " / " + addText + "T";
      }
    }
    
    if (aDocument) {
      // the following should be done with some transformation really...
      var descr, ltext;
      var lines;
      var credit;
      var box;
      
      tooltip = aDocument.createDocumentFragment();
      
      if (entry._error) {
        lines = entry._error.split("\n");
        for (var i = 0; i < lines.length; i++) {
          descr = aDocument.createElement("description");
          descr.setAttribute("class", "checkyesss-class-ttval");
          ltext = aDocument.createTextNode(lines[i]);
          descr.appendChild(ltext);
          tooltip.appendChild(descr);
        }
      }
      else if (entry._credit.toString().length == 0) {
        descr = aDocument.createElement("description");
        ltext = aDocument.createTextNode("Keine Daten vorhanden.");
        descr.appendChild(ltext);
        tooltip.appendChild(descr);
      }
      else {
        var node;
        
        if (entry._rateplan == "webfon" || entry._rateplan == "V") {
          var is_V = (entry._rateplan == "V");
          var duration = this._utils.minutesToString(Math.round(entry._total_duration / 60));
          
          node = aDocument.createElement("label");
          node.setAttribute("value", "Geb\u00FChren seit " + entry._credit_till);
          node.setAttribute("class", "checkyesss-class-ttlab");
          tooltip.appendChild(node);
          
          if (is_V && ((entry._actual_rate == 49) ||
                       (entry._actual_rate == 29)))
          {
            var minRate = Math.round(entry._feeMinRate * 100) / 100;
            minRate = minRate.toString().replace(/\./, ",");
            
            node = aDocument.createElement("label");
            if (entry._actual_rate == 49)
              node.setAttribute("value", minRate + " von 4,9 \u20AC verbraucht");
            else
              node.setAttribute("value", minRate + " von 15 \u20AC verbraucht");
            node.setAttribute("style", "margin-left: 1.5em;");
            node.setAttribute("class", "checkyesss-class-ttval");
            tooltip.appendChild(node);
          }
          
          node = aDocument.createElement("label");
          node.setAttribute("value", entry._total_calls + " Telefonate");
          node.setAttribute("class", "checkyesss-class-ttval");
          node.setAttribute("style", "margin-left: 1.5em;");
          tooltip.appendChild(node);
          
          node = aDocument.createElement("label");
          node.setAttribute("value", duration + " Gespr\u00E4chsdauer");
          node.setAttribute("style", "margin-left: 1.5em;");
          
          if (is_V) {
            node.setAttribute("class", "checkyesss-class-ttval");
            tooltip.appendChild(node);
            
            node = aDocument.createElement("label");
            node.setAttribute("value", entry._total_sms + " SMS");
            node.setAttribute("style", "margin-left: 1.5em;");
          }
          
          node.setAttribute("class", "checkyesss-class-ttval checkyesss-class-ttboxhdr");
          tooltip.appendChild(node);

          if (entry._sms_packet_type) {
            node = aDocument.createElement("label");
            node.setAttribute("value", entry._sms_packet_type);
            node.setAttribute("class", "checkyesss-class-ttlab checkyesss-class-ttboxhdr");
            tooltip.appendChild(node);
          }
        }
        else
        if (entry._rateplan == "internet") {
          credit = parseFloat(entry._credit);

          // Grundoption: 20Euro/GB, Poweroption:10Euro/GB
          var data_rate = (entry._actual_rate == 1) ? 20 : 10;
          var data_vol = Math.round((credit * 1024) / data_rate);
          
          box = aDocument.createElement("hbox");
          box.setAttribute("class", "checkyesss-class-ttboxhdr");
          node = aDocument.createElement("label");
          node.setAttribute("value", "Tarif:");
          node.setAttribute("class", "checkyesss-class-ttlab");
          box.appendChild(node);
          node = aDocument.createElement("label");
          node.setAttribute("value", (entry._actual_rate == 1) ?
                                     "Grundoption" : "Poweroption");
          node.setAttribute("class", "checkyesss-class-ttval");
          box.appendChild(node);
          tooltip.appendChild(box);
          
          node = aDocument.createElement("label");
          node.setAttribute("value", "Guthaben entspricht");
          node.setAttribute("class", "checkyesss-class-ttlab");
          tooltip.appendChild(node);
          
          node = aDocument.createElement("label");
          node.setAttribute("value", data_vol + " MB");
          node.setAttribute("class", "checkyesss-class-ttval checkyesss-class-ttboxhdr");
          node.setAttribute("style", "margin-left: 1.5em;");
          tooltip.appendChild(node);
          
          box = aDocument.createElement("hbox");
          node = aDocument.createElement("label");
          node.setAttribute("value", "g\u00FCltig bis");
          node.setAttribute("class", "checkyesss-class-ttlab");
          box.appendChild(node);
          node = aDocument.createElement("label");
          node.setAttribute("value", (entry._credit_till.length) ? entry._credit_till : "?");
          node.setAttribute("class", "checkyesss-class-ttval");
          box.appendChild(node);
          tooltip.appendChild(box);
        }
        else {
          // credit: compute number of minutes and sms
          var call_rates = {
            9: 0.09,        // old scheme
            8: 0.08,
            6: 0.069,
            90: 0.09,       // new scheme
            80: 0.08,
            69: 0.069,
            68: 0.068
          };
          var sms_rates = {
            9: 0.13,        // old scheme
            8: 0.13,
            6: 0.15,
            90: 0.13,       // new scheme
            80: 0.13,
            69: 0.15,
            68: 0.068
          };
          credit = parseFloat(entry._credit);
          var call_rate = 0.069;       // default to prevent "divide by 0"
          var sms_rate = 0.15;         // default to prevent "divide by 0"
          if (entry._actual_rate in call_rates) {
            call_rate = call_rates[entry._actual_rate];
            sms_rate = sms_rates[entry._actual_rate];
          }
          var minutes = parseInt(credit / call_rate, 10);
          var sms = parseInt(credit / sms_rate, 10);
          var hours = this._utils.minutesToString(minutes);
          sms = sms + " SMS";
          
          node = aDocument.createElement("label");
          node.setAttribute("value", "Guthaben entspricht");
          node.setAttribute("class", "checkyesss-class-ttlab");
          tooltip.appendChild(node);
          
          node = aDocument.createElement("label");
          node.setAttribute("value", hours);
          node.setAttribute("class", "checkyesss-class-ttval");
          node.setAttribute("style", "margin-left: 1.5em;");
          tooltip.appendChild(node);
          
          node = aDocument.createElement("label");
          node.setAttribute("value", sms);
          node.setAttribute("class", "checkyesss-class-ttval checkyesss-class-ttboxhdr");
          node.setAttribute("style", "margin-left: 1.5em;");
          tooltip.appendChild(node);
          
          box = aDocument.createElement("hbox");
          node = aDocument.createElement("label");
          node.setAttribute("value", "g\u00FCltig bis");
          node.setAttribute("class", "checkyesss-class-ttlab");
          box.appendChild(node);
          node = aDocument.createElement("label");
          node.setAttribute("value", (entry._credit_till.length) ? entry._credit_till : "?");
          node.setAttribute("class", "checkyesss-class-ttval checkyesss-class-ttboxhdr");
          box.appendChild(node);
          tooltip.appendChild(box);

          if (entry._intl_packet_type && (pkdaydiff >= -7)) {
            var text;
            node = aDocument.createElement("label");
            node.setAttribute("value", entry._intl_packet_type + " International");
            node.setAttribute("class", "checkyesss-class-ttlab");
            tooltip.appendChild(node);

            if (pkdaydiff < 1) {
              text = "vor " + (-pkdaydiff) + " Tag";
              if (pkdaydiff < -1)
                text += "en";
              text += " ausgelaufen!";
            }
            else
            if (pkdaydiff == 1)
              text = "l\u00E4uft heute aus!";
            else {
              expiry = new Date(entry._intl_packet_till);
              text = "g\u00FCltig bis " + expiry.getDate() + ". " +
                     (expiry.getMonth() + 1) + ". " + expiry.getFullYear();
            }
            node = aDocument.createElement("label");
            node.setAttribute("value", text);
            node.setAttribute("style", "margin-left: 1.5em;");
            node.setAttribute("class", "checkyesss-class-ttval checkyesss-class-ttboxhdr");
            tooltip.appendChild(node);
          }

          if (entry._sms_packet_type) {
            node = aDocument.createElement("label");
            node.setAttribute("value", entry._sms_packet_type);
            node.setAttribute("class", "checkyesss-class-ttlab checkyesss-class-ttboxhdr");
            tooltip.appendChild(node);
          }
        }
      }
    }
    
    container.setPropertyAsString("phoneNumber", pnum);
    container.setPropertyAsString("phoneCredit", entry._credit.toString());
    container.setPropertyAsInt("dataAge", age);
    container.setPropertyAsBool("dataExpired", expired);
    container.setPropertyAsBool("dataError", isError);
    container.setPropertyAsISupports("docTooltip", tooltip);
    if (maxDynColor)
      container.setPropertyAsString("maxDynColor", maxDynColor);
    if (addText)
      container.setPropertyAsString("addText", addText);

    if (entry._rateplan == "V") {
      // fixed dynamic color limits */
      if (entry._actual_rate == 49)
        container.setPropertyAsString("fixedDynColor",
                                      (entry._feeMinRate < 4.9) ? "sup" : "ok");
      else
      if (entry._actual_rate == 29)
        container.setPropertyAsString("fixedDynColor",
                                      (entry._feeMinRate < 15) ? "sup" : "ok");
      else
        // no support for dynamic colors right now
        container.setPropertyAsBool("noDynColor", true);
    }
    else
    if (entry._rateplan == "webfon" || entry._rateplan == "V") {
      // no support for dynamic colors right now
      container.setPropertyAsBool("noDynColor", true);
    }
    
    return container;
  },
  
  /**************************************************************/
  
  getCreditInfo: function(aId, aDocument) {
    if (this._accountsArray) {
      for (var i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._id == aId)
          return this._assembleCreditInfo(i, aDocument);
      }
    }
    return null;
  },
  
  /**************************************************************/
  
  setManualAccountingStart: function(aAccountId, aStartTime) {
    //dump("setManualAccountingStart\n");
    var entry = null;
    var index = -1;
    
    if (this._accountsArray) {
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._id == aAccountId) {
          entry = this._accountsArray[i];
          index = i;
          break;
        }
      }
    }
    
    if (entry && (entry._rateplan == "V" || entry._rateplan == "webfon")) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "cyCheckyesss.setManualAccountingStart",
                                "Accounting period for '"+aAccountId+"' now starts on " +
                                (new Date(aStartTime)).toString(),
                                false);
      
      entry._manualperiod = aStartTime;
      this._recalcCurrentPeriod(index);
      
      if (entry._en)
        this._obs.notifyObservers(this, this._observerTopic,
                                  "update::" + entry._id);
    }
  },
  
  /**************************************************************/
  
  detectedRatePlan: function(aAccountId) {
    var result = null;
    var i;
    
    if (this._accountsArray) {
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._id == aAccountId) {
          if (this._accountsArray[i]._en)
            result = this._accountsArray[i]._rateplan;
          break;
        }
      }
    }
    
    return result;
  },
  
  /**************************************************************/
  
  _refreshDisplay: function() {
    if (this._accountsArray) {
      for (var i = 0; i < this._accountsArray.length; i++) {
        entry = this._accountsArray[i];
        if (entry._en)
          this._obs.notifyObservers(this, this._observerTopic,
                                    "update::" + entry._id);
      }
    }
  },
  
  /**************************************************************/
  
  reloadCredit: function(aId) {
    var index = -1;
    
    if (this._accountsArray) {
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._id == aId) {
          if (this._accountsArray[i]._en)
            index = i;
          break;
        }
      }
    }
    
    if (index >= 0) {
      var win = this._utils.isWebPageOpen(this._blockingUrlRegExp);
      if (win) {
        var ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
                   getService(Components.interfaces.nsIPromptService);    
        ps.alert(null, "Achtung",
                 "Das Guthaben kann derzeit nicht aktualisiert werden,\n"+
                 "da gerade ein YESSS!-Kontomanager angezeigt wird!\n\n"+
                 "Die Aktion wird verz\u00f6gert, bis alle Kontomanager\n"+
                 "geschlossen wurden.");
        
        // focus the mentioned window
        //win.focus();
        
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                  "cyCheckyesss.reloadCredit",
                                  "Conflicting browser window detected, action deferred",
                                  false);
      }
      
      this._accountsArray[index]._forceupdate = true;
      this._checkPending = true;
      
      // restart timer with a short timeout
      this._checkTimer.cancel();
      this._checkTimer.initWithCallback(this, 100, 
                                        Components.interfaces.nsITimer.TYPE_ONE_SHOT);
    }
  },
  
  /**************************************************************/
  
  updateCalls: function(aId, aForced) {
    var result = false;
    var index = -1;
    
    if (this._accountsArray) {
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._id == aId) {
          if (this._accountsArray[i]._en)
            index = i;
          break;
        }
      }
    }
    
    if (index >= 0) {
      var win = this._utils.isWebPageOpen(this._blockingUrlRegExp);
      if (win) {
        if (aForced) {
          var ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
                     getService(Components.interfaces.nsIPromptService);    
          ps.alert(null, "Achtung",
                   "Die Einzelgespr\u00E4che k\u00f6nnen derzeit nicht aktualisiert\n"+
                   "werden, da gerade ein YESSS!-Kontomanager angezeigt wird!\n\n"+
                   "Bitte zuerst ausloggen und dann Aktion wiederholen!");
          
          // focus the mentioned window
          //win.focus();
          
          this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                    "cyCheckyesss.updateCalls",
                                    "Conflicting browser window detected, action aborted",
                                    false);
        }
      }
      else {
        var entry = this._accountsArray[index];
        var now = (new Date()).valueOf();
        if (aForced ||
            (entry._dateCallsLoaded + (TO_MIN_CALLUPD * this._timeHour) < now)) {
          entry._forceupdate = true;
          entry._dateCallsLoaded = 0;
          this._checkPending = true;
          
          /* We need to circumvent the timer here as we have to make sure
             that the access is performed and a notification is returned!
             The usual 100ms timeout would open a time windows where
             conditions could change unexpectedly! */
          this._checkAccounts();
          
          result = true;      // expect a notification
        }
      }
    }
    return result;
  },
  
  /**************************************************************/
  
  getCallsRecords: function(aKey, aFromTime, aToTime) {
    var sum_credit = 0.0;
    var sum_fee = 0.0;         // fee for calls + sms
    var sum_download = 0.0;
    var sum_upload = 0.0;
    var sum_cfee = 0.0;        // fee for all calls
    var sum_duration = 0;
    var sum_charged = 0;       // duration charged for
    var is_data_product = 0;
    var callsArray = new Array();
    var call, duration, fee, credit, charged, type, number, download, upload;
    
    var calls = this._utils.getCallsRecords(aKey, aFromTime, aToTime, null);
    
    while (calls.hasMoreElements()) {
      call = calls.getNext().QueryInterface(Components.interfaces.cyICheckyesssContainer);
      
      number   = call.hasProperty("number") ? call.getPropertyAsString("number") : "";
      duration = call.hasProperty("duration") ? call.getPropertyAsInt("duration") : 0;
      fee      = call.hasProperty("fee") ? call.getPropertyAsString("fee") : "0";
      credit   = call.hasProperty("credit") ? call.getPropertyAsString("credit") : "0";
      download = call.hasProperty("download") ? call.getPropertyAsString("download") : "0";
      upload   = call.hasProperty("upload") ? call.getPropertyAsString("upload") : "0";
      type     = call.hasProperty("type") ? call.getPropertyAsString("type") : "";
      
      is_data_product = is_data_product || (type == "data");
      
      // replace Austrian country code by "0" to make numbers more readable
      number = number.replace(/^\+43/, "0");
      call.setPropertyAsString("number", number);
      
      fee    = Number(fee.replace(/[\.,]/, "."));
      credit = Number(credit.replace(/[\.,]/, "."));
      download = Number(download.replace(/[\.,]/, "."));
      upload   = Number(upload.replace(/[\.,]/, "."));
      
      if (type != "carryover") {
        sum_duration += duration;
        sum_fee      += fee;
        sum_credit   += credit;
        sum_download += download;
        sum_upload   += upload;
      }
      
      if (fee > 0 && type == "call") {
        // Now try to find the duration that's actually charged for.
        // THIS STRONGLY DEPENDS ON THE CURRENTLY AVAILABLE RATE PLANS!!!
        // Note: 9cent: <=60s -> 60s
        //       8cent: <=90s -> 90s
        // thereafter in increments of 30s (rounded up)
        charged = duration - ((duration + 29) % 30) + 29;
        if (charged <= 60)          // special handling of the first 90 seconds
          charged = (fee == 0.08) ? 90 : 60;
        call.setPropertyAsInt("tcharged", charged);
        sum_charged += charged;
        sum_cfee += fee;
      }
      if (type != "call" && type != "data")
        call.deleteProperty("duration");
      
      if (fee > 0 || is_data_product) {
        if (is_data_product) {      // yesss internet
          fee = Math.round(fee * 100.0);        // round to 1 cent
          fee = "000" + fee.toString();
          fee = fee.replace(/^0*(\d+?)(\d{2})$/, "$1,$2");
        }
        else {
          fee = Math.round(fee * 10000.0);        // round to 1/100 cent
          fee = "00000" + fee.toString();
          fee = fee.replace(/^0*(\d+?)(\d{4})$/, "$1,$2");
        }
        call.setPropertyAsString("fee", fee);
      }
      else if (call.hasProperty("fee"))
        call.deleteProperty("fee");
      
      if (credit > 0) {
        credit = credit.toString();
        call.setPropertyAsString("credit", credit);
      }
      else if (call.hasProperty("credit"))
        call.deleteProperty("credit");
      
      callsArray.push(call);
    }
    
    if (callsArray.length > 0) {
      call = Components.classes["@mozdev.org/checkyesss/container;1"].
               createInstance(Components.interfaces.cyICheckyesssContainer);
      call.setPropertyAsBool("statistic", true);
      call.setPropertyAsString("description", "Summe:");
      
      if (sum_duration > 0) {
        call.setPropertyAsInt("duration", sum_duration);
        call.setPropertyAsString("tcharged", this._utils.durationToString(sum_charged));
      }
      if (sum_fee > 0) {
        if (is_data_product) {      // yesss internet
          fee = Math.round(sum_fee * 100.0);         // round to 1 cent
          fee = "000" + fee.toString();
          fee = fee.replace(/^0*(\d+?)(\d{2})$/, "$1,$2");
        }
        else {
          fee = Math.round(sum_fee * 10000.0);      // round to 1/100 cent
          fee = "00000" + fee.toString();
          fee = fee.replace(/^0*(\d+?)(\d{4})$/, "$1,$2");
        }
        fee = "\u20AC " + fee;
        call.setPropertyAsString("fee", fee);
      }
      if (sum_credit > 0) {
        credit = "\u20AC " + sum_credit.toString();
        call.setPropertyAsString("credit", credit);
      }
      if (sum_download > 0) {
        download = "000" + Math.round(sum_download * 100.0);
        download = download.replace(/^0*(\d+?)(\d{2})$/, "$1,$2 MB");
        call.setPropertyAsString("download", download);
      }
      if (sum_upload > 0) {
        upload = "000" + Math.round(sum_upload * 100.0);
        upload = upload.replace(/^0*(\d+?)(\d{2})$/, "$1,$2 MB");
        call.setPropertyAsString("upload", upload);
      }
      callsArray.push(call);
      
      if (is_data_product) {
        if ((sum_download + sum_upload) > 0) {
          call = Components.classes["@mozdev.org/checkyesss/container;1"].
                   createInstance(Components.interfaces.cyICheckyesssContainer);
          call.setPropertyAsBool("statistic", true);
          var total_data = sum_download + sum_upload;
          total_data = Math.round((sum_fee * 102400.0) / total_data);
          fee = "000" + total_data.toString();
          fee = fee.replace(/^0*(\d+?)(\d{2})$/, "~$1,$2 \u20AC/GB");
          call.setPropertyAsString("fee", fee);
          callsArray.push(call);
        }
      }
      else
      if (sum_duration > 0 && sum_cfee > 0) {
        call = Components.classes["@mozdev.org/checkyesss/container;1"].
                 createInstance(Components.interfaces.cyICheckyesssContainer);
        call.setPropertyAsBool("statistic", true);
        call.setPropertyAsString("description", "Durchschnitt:");
        var average = (sum_cfee * 60000.0) / sum_duration;
        average = Math.round(average);                   // round to 1/10 cent
        fee = "0" + average.toString();
        fee = fee.replace(/^0*(\d+?)(\d{1})$/, "$1,$2");
        fee = fee + " \u00A2/min";                       // cent per minute
        call.setPropertyAsString("fee", fee);
        callsArray.push(call);
      }
    }
    
    return new ArrayEnumerator(callsArray);
  },
  
  /**************************************************************/
  
  enableReloadAvailableCalls: function(aAccountId) {
    //dump("enableReloadAvailableCalls\n");
    var entry = null;
    var index = -1;
    
    if (this._accountsArray) {
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._id == aAccountId) {
          entry = this._accountsArray[i];
          index = i;
          break;
        }
      }
    }
    
    if (entry) {
      entry._reloadAvailableCalls = true;
    }
  },
  
  /**************************************************************/
  
  _recalcCurrentPeriod: function(aIndex) {
    //dump("_recalcCurrentPeriod\n");
    var entry = this._accountsArray[aIndex];
    var now = new Date();
    var fromtime;
    
    if (entry._periodreset.match(/^\s*manual/)) {
      fromtime = entry._manualperiod;
      if (fromtime > 0) {
        var date = new Date(fromtime);
        entry._credit_till = date.getDate() + ". " + (date.getMonth()+1) +
                             ". " + date.getFullYear();
      }
    }
    else if (entry._periodreset.match(/^\d+/)) {
      var day = Number(entry._periodreset);
      var month = now.getMonth();
      var year = now.getFullYear();
      if (day > now.getDate()) {
        // switch to previous month
        month--;
        if (month < 0) {
          month = 11;
          year--;
        }
      }
      if ((month == 1) && (day > 28)) {
        // special handling of february
        if ((day > 29) || ((year % 4) != 0)) {
          // proceed to March 1st
          day = 1;
          month++;
        }
      }
      fromtime = (new Date(year, month, day)).valueOf();
      entry._credit_till = day + ". " + (month+1) + ". " + year;
    }
    else {
      // invalid value of entry._periodreset
      fromtime = 0;
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesss._recalcCurrentPeriod",
                                "Invalid type of accounting period '" +
                                entry._periodreset + "'",
                                false);
    }
    if (fromtime == 0)
      entry._credit_till = "Beginn";
    
    var type, ctype, duration, fee;
    var sum_fee = 0.0;         // fee for calls + sms
    var sum_duration = 0;
    var sum_sms = 0;
    var sum_calls = 0;
    var sum_minrate = 0;
    var calls = this._utils.getCallsRecords(entry._id, fromtime, now.valueOf(),
                                            "@type=\"sms\" or @type=\"call\"");
    
    while (calls.hasMoreElements()) {
      call = calls.getNext().QueryInterface(Components.interfaces.cyICheckyesssContainer);
      
      number   = call.hasProperty("number") ? call.getPropertyAsString("number") : "";
      type     = call.hasProperty("type") ? call.getPropertyAsString("type") : "";
      ctype    = call.hasProperty("ctype") ? call.getPropertyAsString("ctype") : "";
      duration = call.hasProperty("duration") ? call.getPropertyAsInt("duration") : 0;
      fee      = call.hasProperty("fee") ? call.getPropertyAsString("fee") : "0";
      fee      = Number(fee.replace(/[\.,]/, "."));
      
      sum_duration += duration;
      sum_fee      += fee;
      if (type == "sms")
        sum_sms++;
      else
        sum_calls++;
      
      // sum up relevant entries for the minimum fee:
      //   Only national calls, no SMS, no roaming
      if (type == "call" &&
          number.match(/^(?:\+43|\d|\*)/) &&
          !ctype.match(/roaming/i))
        sum_minrate += fee;
      
      //dump(call.toString() + "\n");
    }
    // done with calls
    this._utils.releaseCallsRecords();
    
    entry._credit = sum_fee;
    entry._feeMinRate = sum_minrate;
    entry._total_duration = sum_duration;
    entry._total_sms = sum_sms;
    entry._total_calls = sum_calls;
    
    var contents = this._utils.getCacheContents(entry._id);
    if (!contents)
      contents = Components.classes["@mozdev.org/checkyesss/container;1"].
                   createInstance(Components.interfaces.cyICheckyesssContainer);
    
    contents.setPropertyAsInt("manual-period", entry._manualperiod);
    contents.setPropertyAsString("credit", sum_fee);   // it's actually the total fee
    contents.setPropertyAsString("until", entry._credit_till); // abused for start of period
    
    this._utils.writeCache(entry._id,              // key
                           contents,
                           this._utils.getCacheValidity(entry._id),
                           false);                 // do not reset age
  },
  
  /**************************************************************/
  
  _updateFromCache: function(aIndex) {
    //dump("_updateFromCache\n");
    var entry = this._accountsArray[aIndex];
    var valid = false;
    
    if (entry._en &&
        this._utils.existCache(entry._id)) {
      var contents = this._utils.getCacheContents(entry._id);
      
      if (contents) {
        var err = (contents.hasProperty("error")) ? contents.getPropertyAsString("error") : "";
        // convert NL due to moz-bug #169521
        entry._error = err.replace(/\\n/g, "\n");
        
        entry._credit =
          (contents.hasProperty("credit")) ? contents.getPropertyAsString("credit") : "";
        
        entry._actual_rate =
          (contents.hasProperty("actual-rate")) ? contents.getPropertyAsInt("actual-rate") : 69;
        
        entry._credit_till =
          (contents.hasProperty("until")) ? contents.getPropertyAsString("until") : "";
        
        entry._intl_packet_type =
          (contents.hasProperty("intl-packet-type")) ? contents.getPropertyAsString("intl-packet-type") : "";
        entry._intl_packet_till =
          (contents.hasProperty("intl-packet-till")) ? contents.getPropertyAsInt("intl-packet-till") : 0;
        
        entry._sms_packet_type =
          (contents.hasProperty("sms-packet-type")) ? contents.getPropertyAsString("sms-packet-type") : "";
        entry._sms_packet_till =
          (contents.hasProperty("sms-packet-till")) ? contents.getPropertyAsInt("sms-packet-till") : 0;
        
        entry._dateCallsLoaded =
          (contents.hasProperty("dateCallsLoaded")) ? contents.getPropertyAsInt("dateCallsLoaded") : 0;
        
        entry._manualperiod =
          (contents.hasProperty("manual-period")) ? contents.getPropertyAsInt("manual-period") : 0;
        
        if ((entry._rateplan == "auto" || entry._rateplan == "wertkarte") &&
            contents.hasProperty("rateplan") &&
            !contents.hasProperty("actual-rate")) {
          // convert rate plan from pre-0.5.8
          entry._actual_rate = contents.getPropertyAsInt("rateplan") * 10;
        }
        
        if (entry._error.length || entry._credit.toString().length)
          valid = true;
      }
    }
    
    if (valid) {
      if (entry._rateplan == "webfon" || entry._rateplan == "V")
        this._recalcCurrentPeriod(aIndex);
    }
    
    if (!valid) {
      this._accountsArray[aIndex]._error = "Keine Daten vorhanden.";
      this._accountsArray[aIndex]._credit = "";
      this._accountsArray[aIndex]._credit_till = "";
    }
    
    this._obs.notifyObservers(this, this._observerTopic,
                              "update::" + entry._id);
  },
  
  /**************************************************************/
  
  _anyForcedUpdates: function() {
    if (this._accountsArray) {
      var i;
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._en &&
            this._accountsArray[i]._forceupdate)
          return i;
      }
    }
    return -1;
  },
  
  /**************************************************************/
  
  _anyQuickLinkUpdates: function() {
    if (this._accountsArray) {
      var i;
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._en &&
            this._accountsArray[i]._quickLinkPending)
          return i;
      }
    }
    return -1;
  },
  
  /**************************************************************/
  
  _needsUpdate: function(aIndex) {
    var entry;
    try {
      entry = this._accountsArray[aIndex];
    }
    catch (e) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "_needsUpdate",
                                "Invalid internal index " + aIndex, false);
      return "";
    }
    
    if (entry._en) {
      if (entry._quickLinkPending)
        return "link";
      
      var now = (new Date()).valueOf();
      if ((entry._dateCallsLoaded + (TO_CALLUPD * this._timeHour) < now) ||
          entry._reloadAvailableCalls)
        return "calls";
      
      if (entry._forceupdate ||
          !this._utils.existCache(entry._id) ||
          this._utils.getCacheValidity(entry._id) == 0)
        return "credit";
    }
    
    return "";
  },
  
  /**************************************************************/
  
  _checkAccounts: function() {
    //dump("_checkAccounts\n");
    try {
      this._checkTimer.cancel();
    }
    catch(e) { }
    
    this._checkPending = false;
    
    if (!this._accountsArray)
      return;
    
    if (this._webAccess.activeSession) {
      this._checkPending = true;
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesss._checkAccounts",
                                "Another access already active!",
                                false);
      return;
    }
    
    var type = "link";
    var index = this._anyQuickLinkUpdates();
    
    if (index < 0) {
      index = this._anyForcedUpdates();
      if (index >= 0)
        type = this._needsUpdate(index);
    }
    
    if (index < 0) {
      //dump("no forced updates\n");
      // no forced updates available
      var now = (new Date()).valueOf();
      if ((now - this._lastAccess) > (TO_MIN_PAUSE * this._timeMinute) &&
          now > this._suspendUntil) {
        // ensure a minimum break of TO_MIN_PAUSE minutes between accesses
        //dump("check for scheduled updates\n");
        var i;
        for (i = 0; i < this._accountsArray.length; i++) {
          type = this._needsUpdate(i);
          if (type) {
            //dump("type = " + type + "\n");
            index = i;
            break;
          }
        }
      }
    }
    
    if (index >= 0 && !this._utils.isWebPageOpen(this._blockingUrlRegExp)) {
      this._accessAccount = index;
      this._startWebPath(type);
    }
    else
      this._rescheduleCheck();
  },
  
  /**************************************************************/
  
  _rescheduleCheck: function() {
    //dump("_rescheduleCheck\n");
    if (!this._accountsArray)
      return;
    
    var timeout = 0;
    var blocked = this._utils.isWebPageOpen(this._blockingUrlRegExp);
    
    if (this._checkPending ||
        this._anyQuickLinkUpdates() >= 0 ||
        this._anyForcedUpdates() >= 0) {
      //dump("  immediate check\n");
      if (!blocked) {
        this._checkAccounts();
        return;
      }
    }
    else {
      //dump("  scheduled check\n");
      var now = (new Date()).valueOf();
      var mintimeout = Math.max((TO_MIN_PAUSE * this._timeMinute) - (now - this._lastAccess),
                                this._suspendUntil - now);
      if (mintimeout < 0)
        mintimeout = 0;
      
      timeout = 24 * this._timeHour;              // max. timeout
      for (i = 0; i < this._accountsArray.length; i++) {
        var entry = this._accountsArray[i];
        if (entry._en) {
          if (!this._utils.existCache(entry._id)) {
            //dump("  no cache found for " + entry._id + "\n");
            timeout = 0;
            break;
          }
          else {
            var valid = this._utils.getCacheValidity(entry._id);
            //dump("  cache for " + entry._id + " valid " + valid + "ms\n");
            if (valid < timeout)
              timeout = valid;
          }
        }
      }
      
      if (timeout < mintimeout)
        timeout = mintimeout;
      
      timeout += 1000;               // add 1 second tolerance
    }
    
    if (blocked && (timeout < (TO_BLOCKED * this._timeSecond))) {
      timeout = TO_BLOCKED * this._timeSecond;
      /*this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesss._rescheduleCheck",
                                "Scheduled check was delayed by open web page!",
                                false);*/
    }
    
    //dump("timeout = " + timeout + "\n");
    this._checkTimer.initWithCallback(this, timeout,
                                      Components.interfaces.nsITimer.TYPE_ONE_SHOT);
  },
  
  /**************************************************************/
  /**************************************************************/
  
  _webPath: {
    "credit": [ "throttle 1000",
                "getRate",
                "jump/webfon voip",
                "jump/internet internet",
                "login",
                "exit/err",
                "autoRate",
                "jump/wertkarte prepaid",
                "jump/V postpaid",
                "jump/internet internet",
                "invalidRate",
                "exit",
                "prepaid:",       //--------- PrePaid --------
                "loginDone",
                "exit/err",
                "logout",
                "exit",
                "postpaid:",      //-------- PostPaid --------
                "vLoginDone",
                "exit/err",
                "callsPrepare",
                "exit/err",
                "vCalls",             // fetch the calls       <-\
                "vCallsDone",         // parse calls table        |
                "jump/cont -2",       //                        _/
                "vSkipPrepaid",
                "jump/yes vstore",
                "vCallsPrepaid",      // fetch prepaid calls   <-\
                "vCallsDone",         // parse prepaid calls      |
                "jump/cont -2",       //                        _/
                "vstore:",
                "wfCallsStore",
                "vLogout",
                "exit",
                "voip:",          //---------  VoIP ----------
                "wfPrepare",
                "exit/err",
                "wfCalls",            // fetch the calls       <-\
                "wfCallsDone",        // parse calls table        |
                "jump/cont -2",       //                        _/
                "wfCallsStore",
                "exit",
                "internet:",      //--------- Internet --------
                "iLogin",
                "exit/err",
                "iLoginDone",
                "exit/err",
                "iLogout",
                "exit"
                ],
    "calls":  [ "throttle 1000",
                "getRate",
                "jump/webfon voip",
                "jump/internet internet",
                "login",
                "exit/err",
                "autoRate",
                "jump/wertkarte prepaid",
                "jump/V postpaid",
                "jump/internet internet",
                "invalidRate",
                "exit",
                "prepaid:",       //--------- PrePaid --------
                "loginDone",
                "exit/err",
                "callsPrepare",
                "exit/err",
              /*"deposits",
                "depositsDone",*/
                "calls",              // fetch the calls       <-\
                "callsDone",          // parse calls table        |
                "jump/cont -2",       //                        _/
                "callsStore",
                "logout",
                "exit",
                "postpaid:",      //-------- PostPaid --------
                "vLoginDone",
                "exit/err",
                "callsPrepare",
                "exit/err",
                "vCalls",             // fetch the calls       <-\
                "vCallsDone",         // parse calls table        |
                "jump/cont -2",       //                        _/
                "vSkipPrepaid",
                "jump/yes vstore",
                "vCallsPrepaid",      // fetch prepaid calls   <-\
                "vCallsDone",         // parse prepaid calls      |
                "jump/cont -2",       //                        _/
                "vstore:",
                "wfCallsStore",
                "vLogout",
                "exit",
                "voip:",          //---------  VoIP ----------
                "wfPrepare",
                "exit/err",
                "wfCalls",            // fetch the calls       <-\
                "wfCallsDone",        // parse calls table        |
                "jump/cont -2",       //                        _/
                "wfCallsStore",
                "exit",
                "internet:",      //--------- Internet --------
                "iLogin",
                "exit/err",
                "iLoginDone",
                "exit/err",
                "callsPrepare",
                "exit/err",
                "iConns",             // fetch the calls       <-\
                "iConnsDone",         // parse calls table        |
                "jump/cont -2",       //                        _/
                "callsStore",
                "iLogout",
                "exit"
                ],
    "link":   [ "getRate",
                "jump/webfon voip",
                "jump/internet internet",
                "login",
                "jump/err brerr",
                "autoRate",
                "jump/wertkarte prepaid",
                "jump/V postpaid",
                "jump/internet internet",
                "invalidRate",
                "exit",
                "prepaid:",       //--------- PrePaid --------
                "loginDone",
                "jump common",
                "postpaid:",      //-------- PostPaid --------
                "vLoginDone",
                "jump common",
                "voip:",          //---------  VoIP ----------
                "wfLogin",
                "jump/err brerr",
                "wfLoginDone",
                "jump common",
                "internet:",      //--------- Internet --------
                "iLogin",
                "jump/err brerr",
                "iLoginDone",
                "common:",        //- - - - - common - - - - -
                "jump/err brerr",
                "browserGoto",
                "exit",
                "brerr:",
                "browserError",
                "exit"
                ]
  },
  _webPathType: null,
  
  
  _prevFunctionName: "",
  
  webSeqCallback: function(aFunctionName) {
    //dump(">"+aFunctionName+"\n");
    var prefFN = this._prevFunctionName;
    this._prevFunctionName = aFunctionName;
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                              "cyCheckyesss.webSeqCallback", aFunctionName, false);
    
    switch (aFunctionName) {
      case "getRate":        return this._webGetRate();
      case "autoRate":       return this._webAutoRate();
      case "invalidRate":    return this._webInvalidRate();
      case "login":          return this._webLogin();
      case "wfLogin":        return this._webWFLogin();
      case "iLogin":         return this._webILogin();
      case "loginDone":      return this._webLoginDone();
      case "vLoginDone":     return this._webVLoginDone();
      case "wfLoginDone":    return this._webWFLoginDone();
      case "iLoginDone":     return this._webILoginDone();
      case "callsPrepare":   return this._webCallsPrepare();
      case "wfPrepare":      return this._webWFPrepare();
      case "calls":          return this._webCalls();
      case "vCalls":         return this._webVCalls();
      case "wfCalls":        return this._webWFCalls();
      case "iConns":         return this._webIConns();
      case "callsDone":      return this._webCallsDone();
      case "vCallsDone":     return this._webVCallsDone();
      case "wfCallsDone":    return this._webWFCallsDone();
      case "iConnsDone":     return this._webIConnsDone();
      case "vSkipPrepaid":   return this._webVSkipPrepaid();
      case "vCallsPrepaid":  return this._webVCallsPrepaid();
      case "callsStore":     return this._webCallsStore();
      case "wfCallsStore":   return this._webWFCallsStore();
      case "logout":         return this._webLogout();
      case "vLogout":        return this._webVLogout();
      case "iLogout":        return this._webILogout();
      case "browserGoto":    return this._webBrowserGoto(false);
      case "browserError":   return this._webBrowserGoto(true);
      case "exit":           return this._webExit();
      case "error":          return this._webError(prefFN);
      default:
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesss.webSeqCallback",
                                  "Unknown function name " + aFunctionName,
                                  false);
        return "";
    }
  },
  
  /**************************************************************/
  
  _startWebPath: function(aPathType) {
    //dump("_startWebPath\n");
    if (!aPathType in this._webPath) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesss._startWebPath",
                                "Invalid path type " + aPathType,
                                false);
      return;
    }
    
    if (this._webAccess.activeSession) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesss._startWebPath",
                                "Session already active!",
                                false);
      return;
    }
    this._webPathType = aPathType;
    
    this._webAccess.startRequestSequence(new StringEnumerator(this._webPath[aPathType]),
                                         this);
  },
  
  /**************************************************************/
  
  _webGetRate: function() {
    //dump("_webGetService\n");
    var entry = this._accountsArray[this._accessAccount];
    if (entry._rateplan == "auto" && entry._user.match(/^43780/))
      entry._rateplan = "webfon";
    
    entry._forceupdate = false;
    this._lastAccess = (new Date()).valueOf();
    
    return entry._rateplan;
  },
  
  /**************************************************************/
  
  _webLogin: function() {
    //dump("_webLogin\n");
    var entry = this._accountsArray[this._accessAccount];
    var result = "ok";
    
    this._obs.notifyObservers(this, this._observerTopic,
                                    "net-busy::" + entry._id);
    
    if (entry._pw == null) {
      var password = this._managerCenter.getPassword(this._providerId,
                                                     entry._id, entry._user,
                                                     !entry._mayStorePwd, true);
      if (password != null) {
        entry._pw = password;
      }
      else {
        this._webStoreError("Passwort nicht gefunden!");
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesss._webLogin",
                                  "Could not find password for account " + entry._id,
                                  false);
        result = "err";
      }
    }
    
    var check = new RegExp(this._userRegExp);
    if (!entry._user.match(check)) {
      this._webStoreError("Ung\u00FCltige Rufnummer!");
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesss._webLogin",
                                "Invalid format of phone number " + entry._user,
                                false);
      result = "err";
    }
    
    if (result == "ok") {
      // submit the HTTP request
      try {
        var cm = Components.classes["@mozilla.org/cookiemanager;1"].
                   getService(Components.interfaces.nsICookieManager);
        cm.remove("www.yesss.at", "PHPSESSID", "/", false);
        cm.remove("yesss.at", "PHPSESSID", "/", false);
      }
      catch (e) {
        dump(e.toString()+"\n");
      }
      var url = Components.classes["@mozilla.org/network/io-service;1"].
                  getService(Components.interfaces.nsIIOService).
                    newURI(this._yesssLogin, null, null);
      this._webAccess.startRequest(url, "POST", null,
                                   "rufnummer=" + encodeURIComponent(entry._user) +
                                   "&passwort=" + encodeURIComponent(entry._pw),
                                   null);
    }
    
    return result;
  },
  
  /**************************************************************/
  
  _webAutoRate: function() {
    //dump("_webAutoRate\n");
    var entry = this._accountsArray[this._accessAccount];
    if (entry._rateplan == "auto") {
      var text = this._webAccess.contentsText;
      if (text.match(/Login nicht .* Sie sich nur im/)) {
        // internet service
        entry._rateplan = "internet";
      }
      else
      if (text.match(/Login nicht erfolgreich/)) {
        // could not log in -> keep automatic detection active
        return "auto";
      }
      else
      if (this._webAccess.contentsHTML.match(/Aufladung/)) {
        // prepaid service
        entry._rateplan = "wertkarte";
      }
      else {
        // postpaid service
        entry._rateplan = "V";
      }
    }
    return entry._rateplan;
  },
  
  /**************************************************************/
  
  _webLoginDone: function() {
    //dump("_webLoginDone\n");
    var result = "ok";
    
    var response = this._webAccess.contentsText;
    var matches = response.match(/\bEUR\s*(([0-9]+)[,.]([0-9]+))/i);
    
    if (matches) {
      var entry = this._accountsArray[this._accessAccount];
      var expiry;
      
      var contents = this._utils.getCacheContents(entry._id);
      if (!contents)
        contents = Components.classes["@mozdev.org/checkyesss/container;1"].
                     createInstance(Components.interfaces.cyICheckyesssContainer);
      
      entry._credit = matches[2]+"."+matches[3];
      entry._error = "";
      entry._sms_packet_type = "";
      entry._sms_packet_till = 0;
      
      matches = response.match(/ltigkeit.*?([0-9]+\.\s*[0-9]+\.\s*20[0-9]+)/i);
      if (matches)
        entry._credit_till = matches[1];
      else {
        entry._credit_till = "";
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                  "cyCheckyesss._webLoginDone: Date for credit validity not found!",
                                  this._webAccess.contentsHTML, true);
      }
      
      matches = response.match(/(\d+-(?:Monat[es]?|Tages?)-Paket).*?ltigkeit.*?([0-9]+)\.\s*([0-9]+)\.\s*(20\d{2})/i);
      if (matches &&
          !matches[0].match(/\d+-SMS-Paket/)) { // ensure it belongs to the international package
        entry._intl_packet_type = matches[1];
        expiry = (new Date(matches[4], matches[3]-1, matches[2], 23, 59, 00)).valueOf();
        entry._intl_packet_till = expiry;
      }

      // the following regexp is just a guess at this point -> need to verify!
      matches = response.match(/(\d+-SMS-Paket).*?ltigkeit.*?([0-9]+)\.\s*([0-9]+)\.\s*(20\d{2})/i);
      if (matches) {
        entry._sms_packet_type = matches[1];
        expiry = (new Date(matches[4], matches[3]-1, matches[2])).valueOf();
        entry._sms_packet_till = expiry;
      }

      matches = response.match(/derzeitiger.*?\b(\d+(?:[,\.]\d+)?)\s+cent/i);
      if (matches) {
        entry._actual_rate = parseInt(Number(matches[1].replace(/,/, ".")) * 10, 10);
      }
      else {
        entry._error = "Tarif konnte nicht gefunden werden!";
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                  "cyCheckyesss._webLoginDone: Actual rate not found!",
                                  this._webAccess.contentsHTML, true);
      }
      
      this._utils.appendCreditRecord(entry._id, entry._credit.toString());
      contents.setPropertyAsString("credit", entry._credit.toString());
      contents.setPropertyAsString("actual-rate", entry._actual_rate);
      contents.setPropertyAsString("intl-packet-type", entry._intl_packet_type);
      contents.setPropertyAsInt("intl-packet-till", entry._intl_packet_till);
      contents.setPropertyAsString("sms-packet-type", entry._sms_packet_type);
      contents.setPropertyAsInt("sms-packet-till", entry._sms_packet_till);
      contents.setPropertyAsString("until", entry._credit_till);
      contents.setPropertyAsString("error", "");
      
      this._utils.writeCache(entry._id,
                             contents,
                             this._interval * this._timeHour,
                             true);
      
      this._obs.notifyObservers(this, this._observerTopic,
                                "update::" + entry._id);
    }
    else {
      response = this._webAccess.contentsHTML;
      var error;
      var fail = response.replace(/<br>/i, " ").
                   match(/>\s*([^<>]*?\bnicht\s+erfolgreich\b[^<>]*?)\s*</);
      if (fail) {
        var txt = this._utils.html2txt(fail[1]);
        txt = txt.replace(/([!\.,\?])\s*(\S)/g, "$1\n $2");
        txt = txt.replace(/([^\n]{38}.*?)\s+/g, "$1\n ");
        error = "Fehler!! Yesss sagt:\n\"" + txt + "\"";
        result = "err";
      }
      else if (response.match(/action\s*=\s*"\/?kontomanager.php"/)) {
        error = "Fehler beim Einloggen!!";
        result = "err";
      }
      else
        error = "Fehler!! Guthaben\nnicht gefunden.";
      
      this._webStoreError(error);
      
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesss._webLoginDone: "+error,
                                response, true);
    }
    
    return result;
  },
  
  /**************************************************************/
  
  _callsLastTime: 0,         // time of last stored call record
  _callsCheckTime: 0,        // check calls since _callsCheckTime (overlapping!)
  _callsLoadTime: 0,         // load calls since _callsCheckTime (overlapping!)
  _callsNewLast: 0,          // time of newest calls record loaded
  
  _callsPageNo: 0,           // page counter for URL
  
  _callsFound: null,         // array of loaded calls records
  
  _callsMissingFrom: 0,      // start of section missing calls
  _callsMissingTo: 0,        // end of section
  
  
  _webCallsPrepare: function() {
    //dump("_webCallsPrepare\n");
    var entry = this._accountsArray[this._accessAccount];
    var now = new Date();
    var date;
    var line;
    
    this._callsLastTime = this._utils.getLastCallTime(entry._id);
    
    if (entry._reloadAvailableCalls) {
      // we HAVE to reload all calls we can find
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesss._webCallsPrepare",
                                "Reloading all available calls "+
                                "for account '"+entry._id+"'",
                                false);
      entry._reloadAvailableCalls = false;
      this._callsLastTime = 0;
    }
    
    // load calls overlapping 3 days
    this._callsCheckTime = this._callsLastTime - (3 * 24 * this._timeHour);
    if (this._callsCheckTime > 0) {
      // round to midnight
      date = new Date(this._callsCheckTime);
      date = new Date(date.getFullYear(), date.getMonth(), date.getDate(), 0, 0, 0, 0);
    }
    else {
      /*if (entry._dateCallsLoaded > 0) {
        // apparently we already checked before but did not find any calls
        // -> recheck starting 3 days before that date
        // WARNING: This may lead to problems in the following situation:
        //   Someone mistakenly chose a different phone provider, where the
        //   calls update of course fails. Then she/he corrects the provider
        //   in the settings and now the calls will not be downloaded from
        //   Jan 2nd 2006 but from the date determined one line below!!!!!
        date = new Date(entry._dateCallsLoaded - (3 * 24 * this._timeHour));
        date = new Date(date.getFullYear(), date.getMonth(), date.getDate(), 0, 0, 0, 0);
      }
      else {*/
        // start checking at Jan 2nd 2006
        date = new Date(2006, 0, 2, 0, 0, 0, 0);
      /*}*/
    }
    //line = date.getDate()+". "+(date.getMonth()+1)+". "+date.getFullYear(); // debug
    this._callsCheckTime = date.valueOf();
    date = new Date(date.valueOf() - 1);
    this._callsLoadTime = new Date(date.getFullYear(), date.getMonth(), date.getDate(), 0, 0, 0, 0).valueOf();
    this._callsNewLast = 0;
    //dump("_callsCheckTime = "+this._callsCheckTime+"\n");
    
    /* debug
    date = new Date(this._callsLastTime);
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                              "cyCheckyesss._webCallsPrepare",
                              "Account '"+entry._id+"':\nLast call in cache: " +
                              date.getDate()+". "+(date.getMonth()+1)+". "+date.getFullYear()+
                              "\nConsidering new calls between "+line +
                              " and now",
                              false);
    */
    
    this._callsPageNo = 0;
    this._callsFound = new Array();
    
    this._callsMissingTo = now.valueOf();
    this._callsMissingFrom = (this._callsLastTime > 0) ? (this._callsLastTime + 1):
                                                         now.valueOf();
    
    // check whether we are able to calculate the download progress (in %)
    if (this._callsCheckTime > 0)
      this._obs.notifyObservers(this, this._observerTopic,
                                "calls-progress::" + entry._id + "::0");
    
    return "ok";
  },
  
  /*************/
  
  _webDeposits: function() {
    //dump("_webDeposits\n");
    var url = Components.classes["@mozilla.org/network/io-service;1"].
                getService(Components.interfaces.nsIIOService).
                  newURI(this._yesssDeposits, null, null);
    this._webAccess.startRequest(url, "GET", null, null, null);
    
    return "ok";
  },
  
  /*************/
  
  _webDepositsDone: function() {
    //dump("_webDepositsDone\n");
    var entry = this._accountsArray[this._accessAccount];
    var now = new Date();
    var maxFuture = now.valueOf() + this._timeHour;     // limit time to reasonable values
    
    this._obs.notifyObservers(this, this._observerTopic,
                              "calls-progress::" + entry._id + "::20");
    
    var deposits = this._webAccess.findTable(null, "Uhrzeit");
    if (!deposits) {
      entry._error = "Fehler beim Auslesen der Gutschriften!";
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesss._webDepositsDone: "+entry._error,
                                this._webAccess.contentsHTML, true);
      return "err";
    }
    
    var onedeposit, fields;
    var matches;
    var time = 0;
    var money = 0;
    var type = "";
    var parseError = null;
    var callbag;
    
    var depositsFound = new Array();
    var depositsFromTime = now.valueOf();
    
    while (deposits && deposits.hasMoreElements()) {
      onedeposit = deposits.getNext().QueryInterface(Components.interfaces.nsIStringEnumerator);
      fields = new Array();
      while (onedeposit.hasMore())
        fields.push(onedeposit.getNext());
      
      if (fields.length < 4)
        continue;
      
      matches = fields[0].match(/(\d{1,2})\.\s*(\d{1,2})\.\s*(2\d{1,3})\s+(\d{1,2}):(\d{2}):(\d{2})/);
      if (!matches)
        continue;
      else {
        time = (new Date(Number(matches[3]), Number(matches[2])-1, Number(matches[1]),
                         Number(matches[4]), Number(matches[5]), Number(matches[6]), 0)).valueOf();
        if (time > maxFuture)
          parseError = "reasonable time";
      }
      
      if (!parseError) {
        callbag = Components.classes["@mozdev.org/checkyesss/container;1"].
                        createInstance(Components.interfaces.cyICheckyesssContainer);
        
        callbag.setPropertyAsInt("time", time);
        
        type = fields[1] + " (#" + fields[3] + ")";
        callbag.setPropertyAsString("type", "reload");
        callbag.setPropertyAsString("ctype", type);
        
        matches = fields[2].match(/(\d+)[,.](\d+)/);
        if (!matches)
          parseError = "credit";
        else
          money = matches[1]+"."+matches[2];
        
        callbag.setPropertyAsString("credit", money);
        
        if (!parseError) {
          depositsFound.push(callbag);
          if (time < depositsFromTime)
            depositsFromTime = time;
        }
      }
      
      if (parseError) {
        entry._error = "Fehler beim Interpretieren der Aufladungen!";
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesss._webDepositsDone",
                                  "Error parsing "+parseError+":\n"+fields.toString(),
                                  false);
        return "err";
      }
    }
    
    if (depositsFound.length > 0) {
      depositsFromTime -= 2 * 60000;    // 2 minutes tolerance
      this._utils.storeCallsRecords(entry._id,
                                    new ArrayEnumerator(depositsFound),
                                    depositsFromTime, now,
                                    "@type=\"reload\"");
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "cyCheckyesss._webDepositsDone",
                                "Storing " + depositsFound.length + " deposits",
                                false);
    }
    
    return "ok";
  },
  
  /*************/
  
  _webCalls: function() {
    //dump("_webCalls\n");
    var entry = this._accountsArray[this._accessAccount];
    
    this._callsPageNo++;
    var urlstr = this._yesssCalls.replace(/%/, this._callsPageNo+"");
    var url = Components.classes["@mozilla.org/network/io-service;1"].
                getService(Components.interfaces.nsIIOService).
                  newURI(urlstr, null, null);
    this._webAccess.startRequest(url, "GET", null, null, null);
    
    return "ok";
  },
  
  /*************/
  
  _callsTypeMapping: {
    "sms":        "sms",
    "telefonat":  "call",
    "gutschrift": "refund",
    "geladen":    "reload",   //eg. "Guthaben geladen ueber Bankomat"
    "guthaben.bertrag": "carryover"
  },
  
  
  _parseCalls: function(aParam) {
    //dump("_parseCalls\n");
    /* aParam has the following properties:
       
       text:        web-page as plain text
       useUntil:    a limit for the oldest entry to add to calls
       checkUntil:  a limit for the oldest entry to check
       typeMapping: mapping between a patterns and the types
       
       exceed:      the limit was exceeded
       error:       optional error string, null if none
       oldest:      oldest entry found
       newest:      newest entry found
       calls:       array which gets extended
    */
    var count = 0;
    var callsTypeSign = {
      "sms":      -1,
      "call":     -1,
      "refund":   +1,
      "reload":   +1,
      "carryover": 0
    };
    
    aParam.exceed = false;
    aParam.oldest = (new Date()).valueOf();
    aParam.newest = 0;
    aParam.error = null;
    
    var calls = aParam.text.match(/\d{1,2}\.\s?\d{1,2}\.\s?2\d{1,3}\s+\d{1,2}:\d{2}:\d{2}\s+(?:\+?[0-9xX\*]+\s+)?\d{1,2}:\d{2}:\d{2}\s+\+?\d+[,\.]\d+\s+(?:(?!\d{1,2}\.|[<>]).)+/g);
    if (calls) {
      var onecall = null;
      var datum1 = "";
      for (var i = 0; i < calls.length; i++) {
        onecall = calls[i].match(/(\d{1,2})\.\s?(\d{1,2})\.\s?(2\d{1,3})\s+(\d{1,2}):(\d{2}):(\d{2})\s+((?:\+?[0-9xX\*]+\s+)?)(\d{1,2}):(\d{2}):(\d{2})\s+(\+?)(\d+)[,\.](\d+)\s+((?:(?!\d{1,2}\.|[<>]).)+)/);
        if (onecall) {
          var j;
          for (j = 1; j <= 6; j++)           // convert to numbers
            onecall[j] = Number(onecall[j]);
          for (j = 8; j <= 10; j++) 
            onecall[j] = Number(onecall[j]);
          
          var time = (new Date(onecall[3], onecall[2]-1, onecall[1],
                               onecall[4], onecall[5], onecall[6], 0)).valueOf();
          
          if (time >= aParam.useUntil) {
            
            /*! for now the deposits still seem to be mixed with the calls!!
                make this code active as soon as switching over to this._webDeposits
            if (onecall[11] == "+") {
              // old-style table containing credits!!! should not happen
              this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                        "cyCheckyesss._webCallsDone",
                                        "Positive amount in '"+calls[i]+"'",
                                        false);
              continue;
            }*/
            
            var callbag = Components.classes["@mozdev.org/checkyesss/container;1"].
                            createInstance(Components.interfaces.cyICheckyesssContainer);
            
            callbag.setPropertyAsInt("time", time);
            if (time < aParam.oldest)
              aParam.oldest = time;
            if (time > aParam.newest)
              aParam.newest = time;
            
            // change number to international format
            var pnumber = onecall[7];
            pnumber = pnumber.replace(/\s+$/,"");            // trailing spaces
            pnumber = pnumber.replace(/^00/,"+");            // intl number
            pnumber = pnumber.replace(/^0/,"+43");           // local number
            callbag.setPropertyAsString("number", pnumber);
            
            var money = onecall[12]+"."+onecall[13];
            var sign = (onecall[11] == "+") ? +1 : -1;
            var type = (onecall[11] == "+") ? "reload" : "call";
            
            // remove trailing spaces
            onecall[14] = onecall[14].replace(/\s+$/i, "");
            
            for (var t in aParam.typeMapping) {
              if (onecall[14].match(new RegExp(t, "i"))) {
                type = aParam.typeMapping[t];
                sign = callsTypeSign[type];
                break;
              }
            }
            
            callbag.setPropertyAsString("type", type);
            callbag.setPropertyAsString("ctype", onecall[14]);
            
            if (sign >= 0)
              callbag.setPropertyAsString("credit", money);
            else
              callbag.setPropertyAsString("fee", money);
            
            var duration = onecall[8]*3600 + onecall[9]*60 + onecall[10];
            callbag.setPropertyAsInt("duration", duration);
            
            aParam.calls.push(callbag);
            count++;
          }
          else if (time < aParam.checkUntil)
            aParam.exceed = true;
        }
        else {
          // strange parsing error...
          this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                    "cyCheckyesss._parseCalls",
                                    "Calls parsing error in '"+calls[i]+"'",
                                    false);
        }
      }
    }
    
    // if there were no calls to load, then do not load any further!
    if (count == 0)
      aParam.exceed = true;
    
    return;
  },
  
  /*************/
  
  _webCallsDone: function() {
    //dump("_webCallsDone\n");
    var entry = this._accountsArray[this._accessAccount];
    
    var response = this._webAccess.contentsText;
    
    var parseParam = {
      text:        response,
      useUntil:    this._callsCheckTime,
      checkUntil:  this._callsLoadTime,
      typeMapping: { "sms":        "sms",
                     "telefonat":  "call",
                     "gutschrift": "refund",
                     "geladen":    "reload",   //eg. "Guthaben geladen ueber Bankomat"
                     "guthaben.bertrag": "carryover" },
      calls:       this._callsFound
    };
    
    this._parseCalls(parseParam);
    
    /* debug
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                          "cyCheckyesss._webCallsDone",
                          "text = \"" + parseParam.text + "\"\ncalls = \"" +
                          parseParam.calls + "\"\nfertig = \""+parseParam.exceed+"\"",
                          false);
    */
    
    parseParam.calls = null;                 // remove reference, just to be sure
    
    if (parseParam.newest > 0) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "cyCheckyesss._webCallsDone",
                                "Loaded page " + this._callsPageNo + " with data from " +
                                (new Date(parseParam.oldest)).toString() + " to " +
                                (new Date(parseParam.newest)).toString(), false);
    }
    
    if (this._callsNewLast < parseParam.newest)
      this._callsNewLast = parseParam.newest;
    if (this._callsMissingTo >= parseParam.oldest)
      this._callsMissingTo = parseParam.oldest - 1;
    
    if (this._callsCheckTime > 0) {
      var progress = 100;
      if (!parseParam.exceed && 
          (this._callsNewLast > this._callsCheckTime) &&
          (this._callsMissingTo > this._callsCheckTime)) {
        var whole = this._callsNewLast - this._callsCheckTime;
        var diff = this._callsNewLast - this._callsMissingTo;
        /*! use with this._webDeposits():
        progress = 20 + parseInt((80 * diff) / whole, 10);
        if (progress < 20)
          progress = 20;*/
        progress = parseInt((100 * diff) / whole, 10);
        if (progress < 0)
          progress = 0;
        if (progress > 100)
          progress = 100;
      }
      //dump("progress = "+progress+"%\n");
      this._obs.notifyObservers(this, this._observerTopic,
                                "calls-progress::" + entry._id + "::" + progress);
    }
    
    //dump("exceed = "+!parseParam.exceed+"\n");
    return (parseParam.exceed) ? "done" : "cont";
  },
  
  /*************/
  
  _webCallsStore: function() {
    //dump("_webCallsStore\n");
    var now = (new Date()).valueOf();
    var entry = this._accountsArray[this._accessAccount];
    
    if (this._callsFound.length > 0) {
      if (this._callsMissingFrom < this._callsMissingTo) {
        this._utils.mergeCallsMissing(entry._id,
                                      this._callsMissingFrom,
                                      this._callsMissingTo);
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                  "cyCheckyesss._webCallsStore",
                                  "Missing calls from "+
                                  (new Date(this._callsMissingFrom)).toString() + " to " +
                                  (new Date(this._callsMissingTo)).toString(), false);
      }
      
      if (this._callsCheckTime < this._callsMissingTo)
        this._callsCheckTime = this._callsMissingTo;
      
      /*this._utils.mergeCallsRecords(entry._id,
                                    new ArrayEnumerator(this._callsFound),
                                    "time,duration,number,ctype");*/
      this._utils.storeCallsRecords(entry._id,
                                    new ArrayEnumerator(this._callsFound),
                                    this._callsCheckTime, now, null);
                                    // "@type!=\"reload\""); use with this._webDeposits()
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "cyCheckyesss._webCallsStore",
                                "Storing " + this._callsFound.length + " calls from " +
                                (new Date(this._callsCheckTime)).toString() + " to " +
                                (new Date()).toString(), false);
      
    }
    
    entry._dateCallsLoaded = now;
    var contents = this._utils.getCacheContents(entry._id);
    if (contents) {
      contents.setPropertyAsInt("dateCallsLoaded", entry._dateCallsLoaded);
      this._utils.writeCache(entry._id,              // key
                             contents,
                             this._utils.getCacheValidity(entry._id),
                             false);                 // do not reset age
    }
    
    this._obs.notifyObservers(this, this._observerTopic,
                              "calls::" + entry._id);
    
    // we are done -> release some memory
    this._utils.releaseCallsRecords();
    
    return "ok";
  },
  
  /**************************************************************/
  
  _webLogout: function() {
    //dump("_webLogout\n");
    var url = Components.classes["@mozilla.org/network/io-service;1"].
                getService(Components.interfaces.nsIIOService).
                  newURI(this._yesssLogout, null, null);
    this._webAccess.startRequest(url, "GET", null, "", null);
    
    return "ok";
  },
  
  /**************************************************************/
  
  _prepaidVPresent: false,     // true if there are still prepaid calls
  
  
  _webVLoginDone: function() {
    //dump("_webVLoginDone\n");
    var result = "ok";
    var response = this._webAccess.contentsHTML;
    
    if (response.match(/\bgespraeche\.php\b/g)) {
      // check if there are still prepaid calls we could download
      // note: they will only be downloaded if missing calls were detected!
      this._prepaidVPresent = (response.match(/\bprepaid\.php\b/)) ? true : false;
      var entry = this._accountsArray[this._accessAccount];
      entry._error = "";
      entry._sms_packet_type = "";
      entry._sms_packet_till = 0;
      
      var contents = this._utils.getCacheContents(entry._id);
      if (!contents)
        contents = Components.classes["@mozdev.org/checkyesss/container;1"].
                     createInstance(Components.interfaces.cyICheckyesssContainer);
      
      response = this._webAccess.contentsText;
      matches = response.match(/derzeitiger.*?\b(\d+(?:[,\.]\d+)?)\s+cent/i);
      if (matches) {
        entry._actual_rate = parseInt(Number(matches[1].replace(/,/, ".")) * 10, 10);

        // the following regexp is just a guess at this point -> need to verify!
        matches = response.match(/(\d+-SMS-Paket).*?ltigkeit.*?([0-9]+)\.\s*([0-9]+)\.\s*(20\d{2})/i);
        if (matches) {
          entry._sms_packet_type = matches[1];
          expiry = (new Date(matches[4], matches[3]-1, matches[2])).valueOf();
          entry._sms_packet_till = expiry;
        }
      }
      else {
        entry._error = "Tarif konnte nicht gefunden werden!";
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                  "cyCheckyesss._webVLoginDone: Actual rate not found!",
                                  this._webAccess.contentsHTML, true);
      }
      
      contents.setPropertyAsString("actual-rate", entry._actual_rate);
      contents.setPropertyAsString("error", "");
      contents.setPropertyAsString("sms-packet-type", entry._sms_packet_type);
      contents.setPropertyAsInt("sms-packet-till", entry._sms_packet_till);

      this._utils.writeCache(entry._id,
                             contents,
                             this._utils.getCacheValidity(entry._id),
                             false);                 // do not reset age
      
      this._obs.notifyObservers(this, this._observerTopic,
                                "update::" + entry._id);
    }
    else {
      // we do not seem to be logged in
      var error;
      var fail = response.replace(/<br>/i, " ").
                   match(/>\s*([^<>]*?\bnicht\s+erfolgreich\b[^<>]*?)\s*</);
      if (fail) {
        var txt = this._utils.html2txt(fail[1]);
        txt = txt.replace(/([!\.,\?])\s*(\S)/g, "$1\n $2");
        txt = txt.replace(/([^\n]{38}.*?)\s+/g, "$1\n ");
        error = "Fehler!! Yesss sagt:\n\"" + txt + "\"";
      }
      else {
        error = "Fehler beim Einloggen!!";
      }
      result = "err";
      
      this._webStoreError(error);
      
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesss._webVLoginDone: "+error,
                                response, true);
    }
    
    return result;
  },
  
  /*************/
  
  _webVCalls: function() {
    //dump("_webVCalls\n");
    var entry = this._accountsArray[this._accessAccount];
    
    this._callsPageNo++;
    var urlstr = this._yesssVCalls.replace(/%/, this._callsPageNo+"");
    var url = Components.classes["@mozilla.org/network/io-service;1"].
                getService(Components.interfaces.nsIIOService).
                  newURI(urlstr, null, null);
    this._webAccess.startRequest(url, "GET", null, null, null);
    
    return "ok";
  },
  
  /*************/
  
  _webVCallsPrepaid: function() {
    //dump("_webVCallsPrepaid\n");
    var entry = this._accountsArray[this._accessAccount];
    
    this._callsPageNo++;
    var urlstr = this._yesssVCallsP.replace(/%/, this._callsPageNo+"");
    var url = Components.classes["@mozilla.org/network/io-service;1"].
                getService(Components.interfaces.nsIIOService).
                  newURI(urlstr, null, null);
    this._webAccess.startRequest(url, "GET", null, null, null);
    
    return "ok";
  },
  
  /*************/
  
  _webVCallsDone: function() {
    //dump("_webCallsDone\n");
    var entry = this._accountsArray[this._accessAccount];
    
    var response = this._webAccess.contentsText;
    
    var parseParam = {
      text:        response,
      useUntil:    this._callsCheckTime,
      checkUntil:  this._callsLoadTime,
      typeMapping: { "sms":        "sms",
                     "telefonat":  "call",
                     "gutschrift": "refund", 
                     "geladen":    "reload",
                     "guthaben.bertrag": "carryover" },
      calls:       this._callsFound
    };
    
    this._parseCalls(parseParam);
    
    parseParam.calls = null;                 // remove reference, just to be sure
    parseParam.text  = null;
    
    if (parseParam.newest > 0) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "cyCheckyesss._webCallsDone",
                                "Loaded page " + this._callsPageNo + " with data from " +
                                (new Date(parseParam.oldest)).toString() + " to " +
                                (new Date(parseParam.newest)).toString(), false);
    }
    
    if (this._callsNewLast < parseParam.newest)
      this._callsNewLast = parseParam.newest;
    if (this._callsMissingTo >= parseParam.oldest)
      this._callsMissingTo = parseParam.oldest - 1;
    
    if (this._callsCheckTime > 0) {
      var progress = 100;
      if (!parseParam.exceed && 
          (this._callsNewLast > this._callsCheckTime) &&
          (this._callsMissingTo > this._callsCheckTime)) {
        var whole = this._callsNewLast - this._callsCheckTime;
        var diff = this._callsNewLast - this._callsMissingTo;
        /*! use with this._webDeposits():
        progress = 20 + parseInt((80 * diff) / whole, 10);
        if (progress < 20)
          progress = 20;*/
        progress = parseInt((100 * diff) / whole, 10);
        if (progress < 0)
          progress = 0;
        if (progress > 100)
          progress = 100;
      }
      //dump("progress = "+progress+"%\n");
      this._obs.notifyObservers(this, this._observerTopic,
                                "calls-progress::" + entry._id + "::" + progress);
    }
    
    //dump("exceed = "+!parseParam.exceed+"\n");
    return (parseParam.exceed) ? "done" : "cont";
  },
  
  /*************/
  
  _webVSkipPrepaid: function() {
    //dump("_webVSkipPrepaid");
    
    // reinitialize page number
    this._callsPageNo = 0;
    
    var skip = ((this._callsLastTime > 0) && (this._callsMissingTo < this._callsMissingFrom)) ||
               !this._prepaidVPresent;
    
    return skip ? "yes" : "no";
  },
  
  /**************************************************************/
  
  _webVLogout: function() {
    //dump("_webVLogout\n");
    var url = Components.classes["@mozilla.org/network/io-service;1"].
                getService(Components.interfaces.nsIIOService).
                  newURI(this._yesssVLogout, null, null);
    this._webAccess.startRequest(url, "GET", null, "", null);
    
    return "ok";
  },
  
  /**************************************************************/
  
  _callsPageTime: 0,           // the date of the current page to load
  
  _webWFPrepare: function() {
    //dump("_webWFPrepare\n");
    var entry = this._accountsArray[this._accessAccount];
    var now = new Date();
    var date;
    
    this._obs.notifyObservers(this, this._observerTopic,
                                    "net-busy::" + entry._id);
    
    if (entry._pw == null) {
      var password = this._managerCenter.getPassword(this._providerId,
                                                     entry._id, entry._user,
                                                     !entry._mayStorePwd, true);
      if (password != null) {
        entry._pw = password;
      }
      else {
        this._webStoreError("Passwort nicht gefunden!");
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesss._webWFPrepare",
                                  "Could not find password for account " + entry._id,
                                  false);
        return "err";
      }
    }
    
    var check = new RegExp(this._userRegExp);
    if (!entry._user.match(check)) {
      this._webStoreError("Ung\u00FCltige Rufnummer!");
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesss._webWFPrepare",
                                "Invalid format of phone number " + entry._user,
                                false);
      return "err";
    }
    
    this._callsLastTime = this._utils.getLastCallTime(entry._id);
    
    if (entry._reloadAvailableCalls) {
      // we HAVE to reload all calls we can find
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesss._webWFPrepare",
                                "Reloading all available calls "+
                                "for account '"+entry._id+"'",
                                false);
      entry._reloadAvailableCalls = false;
      this._callsLastTime = 0;
    }
    
    // load calls overlapping 3 days
    this._callsCheckTime = this._callsLastTime - (3 * 24 * this._timeHour);
    if (this._callsCheckTime > 0) {
      // round to midnight
      date = new Date(this._callsCheckTime);
      date = new Date(date.getFullYear(), date.getMonth(), date.getDate(), 0, 0, 0, 0);
    }
    else {
      if (entry._dateCallsLoaded > 0) {
        // apparently we already checked before but did not find any calls
        // -> recheck starting 3 days before that date
        date = new Date(entry._dateCallsLoaded - (3 * 24 * this._timeHour));
        date = new Date(date.getFullYear(), date.getMonth(), date.getDate(), 0, 0, 0, 0);
      }
      else {
        // start checking at Jan 2nd 2006
        date = new Date(2006, 0, 2, 0, 0, 0, 0);
      }
    }
    this._callsCheckTime = date.valueOf();
    date = new Date(date.valueOf() - 1);
    this._callsLoadTime = new Date(date.getFullYear(), date.getMonth(), 1, 0, 0, 0, 0).valueOf();
    this._callsNewLast = 0;
    
    this._callsPageTime = new Date(now.getFullYear(), now.getMonth(), 1, 0, 0, 0, 0).valueOf();
    this._callsFound = new Array();
    
    this._callsMissingTo = now.valueOf();
    this._callsMissingFrom = (this._callsLastTime > 0) ? (this._callsLastTime + 1):
                                                         now.valueOf();
    
    // check whether we are able to calculate the download progress (in %)
    if (this._callsCheckTime > 0)
      this._obs.notifyObservers(this, this._observerTopic,
                                "calls-progress::" + entry._id + "::0");
    
    entry._forceupdate = false;
    entry._error = "";
    
    return "ok";
  },
  
  /*************/
  
  _webWFCalls: function() {
    //dump("_webWFCalls\n");
    var entry = this._accountsArray[this._accessAccount];
    
    var date = new Date(this._callsPageTime);
    var month = date.getMonth() + 1;
    var year = date.getFullYear();
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                              "cyCheckyesss._webWFCalls",
                              "month="+month+", year="+year,
                              false);
    
    var urlstr = this._yesssWFCalls.replace(/%/, year+""+month);
    var url = Components.classes["@mozilla.org/network/io-service;1"].
                getService(Components.interfaces.nsIIOService).
                  newURI(urlstr, null, null);
    this._webAccess.startRequest(url, "POST", null,
                                 "login=" + encodeURIComponent(entry._user) +
                                 "&password=" + encodeURIComponent(entry._pw) +
                                 "&year=" + year + "&month=" + month +
                                 "&mode=csv&abfragen=Abrufen",
                                 "iso-8859-15");
    
    return "ok";
  },
  
  /*************/
  
  _webWFCallsDone: function() {
    //dump("_webWFCallsDone\n");
    var loadmore = false;
    var entry = this._accountsArray[this._accessAccount];
    var error = "";
    
    //dump("ct="+this._webAccess.contentType + "\n");
    if (this._webAccess.contentType == "text/csv" || DEBUG) {
      var response = this._webAccess.contentsAll;
      
      var calls = response.split("\n");
      if (calls) {
        var onecall = null;
        var datum1 = "";
        
        for (var i = 0; i < calls.length; i++) {
          onecall = calls[i].split(";");
          if (onecall && (onecall.length == 6)) {
            var j, callbag;
            var matches = onecall[0].match(/(20\d{2})\s*\-\s*(\d{1,2})\s*\-\s*(\d{1,2})\s+(\d{1,2}):(\d{2}):(\d{2})/);
            if (matches) {
              for (j = 1; j <= 6; j++)           // convert to numbers
                matches[j] = Number(matches[j]);
              var time = (new Date(matches[1], matches[2]-1, matches[3],
                                   matches[4], matches[5], matches[6], 0)).valueOf();
              
              if (time >= this._callsCheckTime) {
                callbag = Components.classes["@mozdev.org/checkyesss/container;1"].
                            createInstance(Components.interfaces.cyICheckyesssContainer);
                
                callbag.setPropertyAsInt("time", time);
                if (this._callsNewLast < time)
                  this._callsNewLast = time;
                if (time < this._callsMissingTo)
                  this._callsMissingTo = time - 1;
                
                // change number to international format
                var pnumber = onecall[3];
                pnumber = pnumber.replace(/^00/,"+");            // intl number
                pnumber = pnumber.replace(/^0/,"+43");           // local number
                pnumber = pnumber.replace(/^43/,"+43");          // the usual format!
                callbag.setPropertyAsString("number", pnumber);
                
                callbag.setPropertyAsString("type", "call");
                callbag.setPropertyAsString("ctype", onecall[4]);
                
                matches = onecall[5].match(/(\d+)[.,](\d+)/);
                if (matches) {
                  var money = Number(matches[1]+"."+matches[2]);
                  money *= 1.2;      // add VAT
                  callbag.setPropertyAsString("fee", money);
                }
                else
                  error = "money";
                
                matches = onecall[1].match(/(\d{1,2}):(\d{2}):(\d{2})/);
                if (matches) {
                  var duration = Number(matches[1])*3600 +
                                 Number(matches[2])*60 +
                                 Number(matches[3]);
                  callbag.setPropertyAsString("duration", duration);
                }
                else if (!error)
                  error = "duration";
                
                if (!error) {
                  this._callsFound.push(callbag);
                }
              }
            }
            else
              error = "date/time";
            
            if (error) {
              entry._error = "Fehler beim Auslesen\nder Einzelgespr\u00E4che";
              this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                        "cyCheckyesss._webWFCallsDone",
                                        "Error parsing " + error + "in '"+calls[i]+"'",
                                        false);
              error = null;
            }
          }
        }
      }
    }
    
    var now = new Date();
    var page1 = new Date(now.getFullYear(), now.getMonth(), 1, 0, 0, 0, 0).valueOf();
    var progress = Math.round(((this._callsPageTime - page1) * 100) /
                              (this._callsLoadTime - page1));
    //dump("progress = "+progress+"%\n");
    this._obs.notifyObservers(this, this._observerTopic,
                              "calls-progress::" + entry._id + "::" + progress);
    
    // prepare for previous month
    var date = new Date(this._callsPageTime - 1);
    this._callsPageTime = new Date(date.getFullYear(), date.getMonth(), 1, 0, 0, 0, 0).valueOf();
    
    loadmore = (this._callsPageTime >= this._callsLoadTime);
    
    //dump("loadmore = "+loadmore+"\n");
    return (loadmore) ? "cont" : "done";
  },
  
  /*************/
  
  _webWFCallsStore: function() {
    //dump("_webWFCallsStore\n");
    var now = (new Date()).valueOf();
    var entry = this._accountsArray[this._accessAccount];
    
    var contents = this._utils.getCacheContents(entry._id);
    if (!contents)
      contents = Components.classes["@mozdev.org/checkyesss/container;1"].
                   createInstance(Components.interfaces.cyICheckyesssContainer);
    
    if (this._callsFound.length > 0) {
      if (this._callsMissingFrom < this._callsMissingTo) {
        this._utils.mergeCallsMissing(entry._id,
                                      this._callsMissingFrom,
                                      this._callsMissingTo);
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                  "cyCheckyesss._webCallsStore",
                                  "Missing calls from "+
                                  (new Date(this._callsMissingFrom)).toString() + " to " +
                                  (new Date(this._callsMissingTo)).toString(), false);
      }
      
      if (this._callsCheckTime < this._callsMissingTo)
        this._callsCheckTime = this._callsMissingTo;
      
      /*this._utils.mergeCallsRecords(entry._id,
                                    new ArrayEnumerator(this._callsFound),
                                    "time,duration,number,ctype");*/
      this._utils.storeCallsRecords(entry._id,
                                    new ArrayEnumerator(this._callsFound),
                                    this._callsCheckTime, now, null);
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "cyCheckyesss._webWFCallsStore",
                                "Storing " + this._callsFound.length + " calls from " +
                                (new Date(this._callsCheckTime)).toString() + " to " +
                                (new Date()).toString(), false);
    }
    
    entry._dateCallsLoaded = now;
    
    contents.setPropertyAsInt("dateCallsLoaded", entry._dateCallsLoaded);
    // convert NL due to moz-bug #169521
    contents.setPropertyAsString("error", entry._error.replace(/\n/g, "\\n"));
    
    var interval = this._interval;         // default to configured interval
    if (this._utils.getCallsCount(entry._id) == 0) {
      // if there are no calls at all, recheck in 24 hours!
      interval = 24;
    }
    
    this._utils.writeCache(entry._id,
                           contents,
                           interval * this._timeHour,
                           true);
    
    this._recalcCurrentPeriod(this._accessAccount);
    this._utils.appendCreditRecord(entry._id, entry._credit.toString());
    
    this._obs.notifyObservers(this, this._observerTopic,
                              "update::" + entry._id);
    
    this._obs.notifyObservers(this, this._observerTopic,
                              "calls::" + entry._id);
    
    return "ok";
  },
  
  /*************/
  
  /* for now this is just needed for quickLinks */
  
  _webWFLogin: function() {
    //dump("_webWFLogin\n");
    var entry = this._accountsArray[this._accessAccount];
    
    /*this._obs.notifyObservers(this, this._observerTopic,
                                    "net-busy::" + entry._id);*/
    
    if (entry._pw == null) {
      var password = this._managerCenter.getPassword(this._providerId,
                                                     entry._id, entry._user,
                                                     !entry._mayStorePwd, true);
      if (password != null) {
        entry._pw = password;
      }
      else {
        this._webStoreError("Passwort nicht gefunden!");
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesss._webWFLogin",
                                  "Could not find password for account " + entry._id,
                                  false);
        return "err";
      }
    }
    
    var check = new RegExp(this._userRegExp);
    if (!entry._user.match(check)) {
      this._webStoreError("Ung\u00FCltige Rufnummer!");
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesss._webWFLogin",
                                "Invalid format of phone number " + entry._user,
                                false);
      return "err";
    }
    
    var url = Components.classes["@mozilla.org/network/io-service;1"].
                getService(Components.interfaces.nsIIOService).
                  newURI(this._yesssWFLogin, null, null);
    this._webAccess.startRequest(url, "POST", null,
                                 "rufnummer=" + encodeURIComponent(entry._user) +
                                 "&passwort=" + encodeURIComponent(entry._pw),
                                 null);
    return "ok";
  },
  
  /*************/
  
  _webWFLoginDone: function() {
    //dump("_webWFLoginDone\n");
    var result = "ok";
    
    var response = this._webAccess.contentsText;
    var matches = response.match(/Rechnungsadresse/i);
    
    if (!matches) {
      result = "err";
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesss._webWFLoginDone: Cannot log in",
                                this._webAccess.contentsHTML, true);
    }
    
    return result;
  },
  
  /**************************************************************/
  /**************************************************************/
  
  _webILogin: function() {
    //dump("_webILogin\n");
    var entry = this._accountsArray[this._accessAccount];
    var result = "ok";
    
    this._obs.notifyObservers(this, this._observerTopic,
                                    "net-busy::" + entry._id);
    
    if (entry._pw == null) {
      var password = this._managerCenter.getPassword(this._providerId,
                                                     entry._id, entry._user,
                                                     !entry._mayStorePwd, true);
      if (password != null) {
        entry._pw = password;
      }
      else {
        this._webStoreError("Passwort nicht gefunden!");
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesss._webILogin",
                                  "Could not find password for account " + entry._id,
                                  false);
        result = "err";
      }
    }
    
    var check = new RegExp(this._userRegExp);
    if (!entry._user.match(check)) {
      this._webStoreError("Ung\u00FCltige Rufnummer!");
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesss._webILogin",
                                "Invalid format of phone number " + entry._user,
                                false);
      result = "err";
    }
    
    if (result == "ok") {
      // submit the HTTP request
      try {
        var cm = Components.classes["@mozilla.org/cookiemanager;1"].
                   getService(Components.interfaces.nsICookieManager);
        cm.remove("www.yesss.at", "PHPSESSID", "/", false);
        cm.remove("yesss.at", "PHPSESSID", "/", false);
      }
      catch (e) {
        dump(e.toString()+"\n");
      }
      var url = Components.classes["@mozilla.org/network/io-service;1"].
                  getService(Components.interfaces.nsIIOService).
                    newURI(this._yesssILogin, null, null);
      this._webAccess.startRequest(url, "POST", null,
                                   "rufnummer=" + encodeURIComponent(entry._user) +
                                   "&passwort=" + encodeURIComponent(entry._pw),
                                   null);
    }
    
    return result;
  },
  
  /**************************************************************/
  
  _webILoginDone: function() {
    //dump("_webILoginDone\n");
    var result = "ok";
    
    var response = this._webAccess.contentsText;
    var matches = response.match(/\bGuthaben:.*?EUR\s*(\d+(?:[,\.]\d+)?)/i);
    
    if (matches) {
      var entry = this._accountsArray[this._accessAccount];
      
      var contents = this._utils.getCacheContents(entry._id);
      if (!contents)
        contents = Components.classes["@mozdev.org/checkyesss/container;1"].
                     createInstance(Components.interfaces.cyICheckyesssContainer);
      
      var credit = matches[1];
      entry._credit = credit.replace(/,/, ".");
      entry._error = "";
      
      // Grundoption: 20Euro/GB, Poweroption:10Euro/GB
      matches = response.match(/\b(Grund|Power)option\b/i);
      if (matches) {
        entry._actual_rate = (matches[1].match(/Grund/i)) ? 1 : 2;
      }
      else {
        entry._error = "Tarif konnte nicht gefunden werden!";
        entry._actual_rate = 0;
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                  "cyCheckyesss._webILoginDone: Actual rate not found!",
                                  this._webAccess.contentsHTML, true);
      }
      
      matches = response.match(/\bletzte\s+Aufladung\b.*?([0-9]+)\.\s*([0-9]+)\.\s*(20[0-9]+)/i);
      if (matches)
      {
        var reload = new Date(Number(matches[3]), 
                              Number(matches[2]) - 1,
                              Number(matches[1]));
        if (entry._actual_rate == 1) {
          // Grundoption
          // credit expires after one year
          reload.setFullYear(reload.getFullYear() + 1);
        }
        else {
          // Poweroption
          // credit expires after one month, however it is unclear
          // if this is always 30 days
          reload.setFullYear(reload.getFullYear(), reload.getMonth() + 1);
        }
        entry._credit_till = reload.getDate() + ". " +
                             (reload.getMonth()+1) + ". " +
                             reload.getFullYear();
      }
      else {
        entry._credit_till = "";
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                  "cyCheckyesss._webILoginDone: Date for credit validity not found!",
                                  this._webAccess.contentsHTML, true);
      }
      
      this._utils.appendCreditRecord(entry._id, entry._credit.toString());
      contents.setPropertyAsString("credit", entry._credit.toString());
      contents.setPropertyAsString("actual-rate", entry._actual_rate);
      contents.setPropertyAsString("until", entry._credit_till);
      contents.setPropertyAsString("error", entry._error);
      
      this._utils.writeCache(entry._id,
                             contents,
                             this._interval * this._timeHour,
                             true);
      
      this._obs.notifyObservers(this, this._observerTopic,
                                "update::" + entry._id);
    }
    else {
      response = this._webAccess.contentsHTML;
      var error;
      var fail = response.replace(/<br>/i, " ").
                   match(/>\s*([^<>]*?\bnicht\s+erfolgreich\b[^<>]*?)\s*</);
      if (fail) {
        var txt = this._utils.html2txt(fail[1]);
        txt = txt.replace(/([!\.,\?])\s*(\S)/g, "$1\n $2");
        txt = txt.replace(/([^\n]{38}.*?)\s+/g, "$1\n ");
        error = "Fehler!! Yesss sagt:\n\"" + txt + "\"";
        result = "err";
      }
      else if (response.match(/action\s*=\s*"\/?kontomanager.php"/)) {
        error = "Fehler beim Einloggen!!";
        result = "err";
      }
      else
        error = "Fehler!! Guthaben\nnicht gefunden.";
      
      this._webStoreError(error);
      
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesss._webILoginDone: "+error,
                                response, true);
    }
    
    return result;
  },
  
  /*************/
  
  _webIConns: function() {
    //dump("_webIConns\n");
    var entry = this._accountsArray[this._accessAccount];
    
    this._callsPageNo++;
    var urlstr = this._yesssIConns.replace(/%/, this._callsPageNo+"");
    var url = Components.classes["@mozilla.org/network/io-service;1"].
                getService(Components.interfaces.nsIIOService).
                  newURI(urlstr, null, null);
    this._webAccess.startRequest(url, "GET", null, null, null);
    
    return "ok";
  },
  
  /*************/
  
  _webIConnsDone: function() {
    //dump("_webIConnsDone\n");
    var entry = this._accountsArray[this._accessAccount];
    var now = new Date();
    var maxFuture = now.valueOf() + this._timeHour;     // limit time to reasonable values
    
    var calls = this._webAccess.findTable("", "Dauer:");
    if (!calls) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "cyCheckyesss._webIConnsDone",
                                "There are no calls available from " +
                                (new Date(this._callsLoadTime)).toString() + " to " +
                                now.toString(), false);
      return "done";
    }
    
    var onecall, fields;
    var matches;
    var time = 0;
    var money = 0;
    var duration = 0;
    var parseError = null;
    var callcount = 0;
    var date1 = 0;
    var exceed = 0;
    
    while (calls.hasMoreElements()) {
      onecall = calls.getNext().QueryInterface(Components.interfaces.nsIStringEnumerator);
      fields = new Array();
      while (onecall.hasMore())
        fields.push(onecall.getNext());
      
      if ((fields.length < 7) || (fields[1].match(/Dauer:?/)))
        continue;
      
      matches = fields[0].match(/^\s*(\d{1,2})\.\s*(\d{1,2})\.\s*(\d{4})\s+(\d{1,2}):(\d{2}):(\d{2})/);
      if (!matches)
        parseError = "date & time";
      else {
        time = (new Date(Number(matches[3]), Number(matches[2])-1, Number(matches[1]),
                         Number(matches[4]), Number(matches[5]), Number(matches[6]), 0)).valueOf();
        if (time > maxFuture)
          parseError = "reasonable time";
      }
      
      if (time >= this._callsCheckTime && !parseError) {
        var callbag = Components.classes["@mozdev.org/checkyesss/container;1"].
                        createInstance(Components.interfaces.cyICheckyesssContainer);
        
        callbag.setPropertyAsInt("time", time);
        callbag.setPropertyAsString("number", "");
        
        matches = fields[1].match(/(\d+):(\d+):(\d+)/);
        if (!matches) {
          // duration may be 'k.A.' if data volume is very low
          // (according to provider's web site)
          duration = 0;
        }
        else {
          duration = Number(matches[1])*3600 +
                     Number(matches[2])*60 +
                     Number(matches[3]);
        }
        callbag.setPropertyAsInt("duration", duration);
        
        callbag.setPropertyAsString("sessionid", fields[2]);
        
        matches = fields[3].match(/(\d+[,\.]\d+)\s*MB/);
        callbag.setPropertyAsString("upload", matches[1]);
        
        matches = fields[4].match(/(\d+[,\.]\d+)\s*MB/);
        callbag.setPropertyAsString("download", matches[1]);
        
        matches = fields[6].match(/(\d+)[,\.](\d+)/);
        if (!matches)
          parseError = "fee";
        else
          money = matches[1]+"."+matches[2];
        
        callbag.setPropertyAsString("type", "data");
        callbag.setPropertyAsString("ctype", fields[5]);
        callbag.setPropertyAsString("fee", money);
        
        if (!parseError) {
          this._callsFound.push(callbag);
          if (this._callsNewLast < time)
            this._callsNewLast = time;
          if (time < this._callsOldest)
            this._callsOldest = time;
          if (this._callsMissingTo >= time)
            this._callsMissingTo = time - 1;
          if (!callcount)               // store the first date for debugging
            date1 = (new Date(time)).toString();
          callcount++;
        }
      }
      else if (!parseError) {
        if (time < this._callsLoadTime)
          exceed = true;
      }
      
      if (parseError) {
        entry._error = "Fehler beim Interpretieren der Einzelverbindungen!";
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesss._webIConnsDone",
                                  "Error parsing "+parseError+":\n"+fields.toString(),
                                  false);
        return "err";
      }
    }
    if (callcount > 0)
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "cyCheckyesss._webIConnsDone",
                                "Loaded "+callcount+" connection records ranging from " +
                                (new Date(date1)).toString() + " to " +
                                (new Date(time)).toString(), false);
    else
      exceed = true;
    
    if (this._callsCheckTime > 0) {
      var progress = 100;
      if (!exceed && 
          (this._callsNewLast > this._callsCheckTime) &&
          (this._callsMissingTo > this._callsCheckTime)) {
        var whole = this._callsNewLast - this._callsCheckTime;
        var diff = this._callsNewLast - this._callsMissingTo;
        progress = parseInt((100 * diff) / whole, 10);
        if (progress < 0)
          progress = 0;
        if (progress > 100)
          progress = 100;
      }
      //dump("progress = "+progress+"%\n");
      this._obs.notifyObservers(this, this._observerTopic,
                                "calls-progress::" + entry._id + "::" + progress);
    }
    
    //dump("exceed = "+exceed+"\n");
    return (exceed) ? "done" : "cont";
  },
  
  /**************************************************************/
  
  _webILogout: function() {
    //dump("_webILogout\n");
    var url = Components.classes["@mozilla.org/network/io-service;1"].
                getService(Components.interfaces.nsIIOService).
                  newURI(this._yesssILogout, null, null);
    this._webAccess.startRequest(url, "GET", null, "", null);
    
    return "ok";
  },
  
  /**************************************************************/
  /**************************************************************/
  
  _webInvalidRate: function() {
    //dump("_webInvalidRate\n");
    this._webStoreError("Fehler bei der Tariferkennung!");
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                              "cyCheckyesss._webInvalidRate: Could not detect rate",
                              this._webAccess.contentsHTML, true);
    
    return "ok";
  },
  
  /**************************************************************/
  
  _webBrowserGoto: function(aError) {  
    //dump("_webBrowserGoto\n");
    var entry = this._accountsArray[this._accessAccount];
    entry._quickLinkPending = false;
    
    if (entry._quickLinkWindow) {
      if (!aError) {
        // display the remote html page
        entry._quickLinkWindow.location = entry._quickLinkTarget;
      }
      else {
        // replace the announcement by an error message
        var errorText = (entry._error) ? entry._error.replace(/\n/, " ") :
                                         "Unbekannter Fehler";
        var errorPage = "<h2><img src='chrome://checkyesss/skin/logo.png'/>"+
              "&nbsp;&nbsp; " + errorText + "</h2></body></html>";
        var body = entry._quickLinkWindow.document.getElementById("qlbody");
        if (body)
          body.innerHTML = errorPage;
      }
      entry._quickLinkWindow.focus();
      entry._quickLinkWindow = null;
      entry._quickLinkTarget = "";
    }
  },
  
  /**************************************************************/
  
  _webStoreError: function(aErrMsg, aFunctionName, aSerious) {
    aFunctionName = (typeof(aFunctionName) != "undefined") ? aFunctionName : "unknown";
    aSerious = (typeof(aSerious) != "undefined") ? aSerious : true;
  
    var entry = this._accountsArray[this._accessAccount];
    
    // A "serious" error is also written into the account's cache causing the
    // icon to turn gray (or similar). Non-serious error are just recorded in
    // the error log.
    if (aSerious) {
      entry._error = aErrMsg;
      
      var contents = this._utils.getCacheContents(entry._id);
      if (!contents)
        contents = Components.classes["@mozdev.org/checkyesss/container;1"].
                     createInstance(Components.interfaces.cyICheckyesssContainer);
      // convert NL due to moz-bug #169521
      contents.setPropertyAsString("error", entry._error.replace(/\n/g, "\\n"));
      
      this._utils.writeCache(entry._id,              // key
                             contents,
                             4 * this._timeHour,     // expires in 4 hours
                             false);                 // do not reset age
    }
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                              "cyCheckyesss._webStoreError",
                              "Account: " + entry._id + "\nCause: " + aErrMsg +
                              "\nFunction: " + aFunctionName + "\nSerious: " + aSerious,
                              false);
    
    this._obs.notifyObservers(this, this._observerTopic,
                              "update::" + entry._id);
  },
  
  /**************************************************************/
  
  _webError: function(aPrefFunctionName) {
    var error = "Unbekannter Fehler";
    var serious = true;
    
    if (this._webAccess.statusText)
      error = this._webAccess.statusText;
    else if (this._webAccess.status > 200)
      error = "Error " + this._webAccess.status;
    
    // Work around a bug (at least that's what it looks right now) in
    // Firefox 3.0b5 (as included in Ubuntu 8.04) where a 302 redirection from
    // a 'https' URL to a 'http' URL results in a security warning while
    // nsIXMLHttpRequest returns a 302 error.
    if (aPrefFunctionName.match(/logout/i))
      serious = false;
    
    this._webStoreError(error, aPrefFunctionName, serious);
    
    return "exit";
  },
  
  /**************************************************************/
  
  _webExit: function() {
    //dump("_webExit\n");
    this._lastAccess = (new Date()).valueOf();
    
    var entry = this._accountsArray[this._accessAccount];
    this._obs.notifyObservers(this, this._observerTopic,
                                    "net-idle::" + entry._id);
    
    if (this._webPathType == "calls") {
      if (entry._error) {
        this._obs.notifyObservers(this, this._observerTopic,
                                  "calls-error::" + entry._id + "::" + entry._error);
      }
      else {
        this._obs.notifyObservers(this, this._observerTopic,
                                  "calls::" + entry._id);
      }
    }
    
    if (this._webPathType == "link") {
      entry._quickLinkPending = false;
      entry._quickLinkWindow = null;
      entry._quickLinkTarget = "";
    }
    
    this._rescheduleCheck();
    return "ok";
  },
  
  /**************************************************************/
  
  abortAccess: function() {
    try {
      this._checkTimer.cancel();
    } catch(e) { }
    
    if (this._webAccess.activeSession) {
      //dump("abortAccess\n");
      this._webAccess.abortRequestSequence(false);
      this._lastAccess = (new Date()).valueOf();
      
      var entry = this._accountsArray[this._accessAccount];
      this._obs.notifyObservers(this, this._observerTopic,
                                      "net-idle::" + entry._id);
    }
  },
  
  /**************************************************************/
  
  ///////////////////////////
  // nsITimerCallback
  notify: function(timer) {
    //dump("notify\n");
    if (timer == this._checkTimer) {
      this._checkAccounts();
    }
  },
  
  /**************************************************************/
  
  observe: function(aSubject, aTopic, aData) {
    //do nothing until the first load
    if (this._firstTime)
      return;
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                              aTopic, aData, false);
    
    switch (aTopic) {
      case "xpcom-shutdown":
        this._stop();
        break;
      
      default:
        break;
    }
  },
  
  /**************************************************************/
  
  _quickLinkUrls: {
    "auto":
    [ "http://www.yesss.at/" ],
    "wertkarte":
    [ "http://www.yesss.at/",
      "https://www.yesss.at/kontomanager/kundendaten.php",
      "https://www.yesss.at/kontomanager/websms.php",
      "https://www.yesss.at/kontomanager/aufladen.php" ],
    "V":
    [ "http://www.yesss.at/",
      "https://www.yesss.at/kontomanager_v/uebersicht.php",
      "https://www.yesss.at/kontomanager_v/websms.php" ],
    "webfon":
    [ "http://www.yesss.at/webfon/",
      "https://www.yesss.at/webfon/kontomanager/uebersicht.php" ],
    "internet":
    [ "http://www.yesss.at/diskont-surfen/",
      "https://www.yesss.at/diskont-surfen/kontomanager/kundendaten.php" ] },
  
  jumpQuickLink: function(aId, aIndex, aTab) {
    var win;
    var entry = null;
    var i;
    if (this._accountsArray) {
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._id == aId) {
          if (this._accountsArray[i]._en)
            entry = this._accountsArray[i];
          break;
        }
      }
    }
    var rateplan = (entry) ? entry._rateplan : "auto";
    
    if (aIndex < 1) {
      // no need to log in
      win = this._utils.openUrlInBrowser(this._quickLinkUrls[rateplan][0], aTab);
      if (win)
        win.focus();
    }
    else if (!this._utils.isThunderbird) {
      // we have to log in before showing the page!
      if (entry) {
        var ps;
        win = this._utils.isWebPageOpen(this._blockingUrlRegExp);
        if (win) {
          ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
                     getService(Components.interfaces.nsIPromptService);    
          ps.alert(null, "Achtung",
                   "Die Seite kann nicht aufgerufen werden, da\n"+
                   "bereits ein YESSS!-Kontomanager angezeigt wird!\n\n"+
                   "Bitte zuerst ausloggen und dann Aktion wiederholen!");
          
          // focus the mentioned window
          //win.focus();
          
          this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                    "cyCheckyesss.jumpQuickLink",
                                    "Conflicting browser window detected, action aborted",
                                    false);
        }
        else if (this._anyQuickLinkUpdates() >= 0) {
          ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
                     getService(Components.interfaces.nsIPromptService);
          ps.alert(null, "Achtung",
                   "Die Seite kann nicht aufgerufen werden, da\n" +
                   "gerade ein anderer Quick-Link geladen wird!");
          
          // TODO: we could display the window of the active quick link...
          
          this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                    "cyCheckyesss.jumpQuickLink",
                                    "Another quick link request is already active!",
                                    false);
        }
        else {
          var names = this._managerCenter.getQuickLinkNames(this._providerId,
                                                            rateplan);
          var name = null;
          i = aIndex;
          while (names && names.hasMore()) {
            name = names.getNext();
            if (i == 0)
              break;
            i--;
          }
          names = null;
          if (name) {
            var jumpPage = "<html><head>"+
              "<title>" + name + " &ouml;ffnen...</title><head>\n"+
              "<body><h2><!--img src='chrome://checkyesss/skin/logo.png'/-->"+
              "&nbsp;&nbsp;Weiterleitung zu " + name +
              "...</h2></body></html>";
            
            win = this._utils.openUrlInBrowser("", aTab);
            try {
              // does does not work with Firefox 3.0b3
              win.document.open("text/html");
              win.document.write(jumpPage);
              win.document.close();
              win.focus();
            }
            catch (e) {
              dump(e.toString()+"\n");
            }
            
            entry._quickLinkPending = true;
            entry._quickLinkWindow = win;
            entry._quickLinkTarget = this._quickLinkUrls[rateplan][aIndex];
            this._checkPending = true;
            
            // restart timer with a short timeout
            this._checkTimer.cancel();
            this._checkTimer.initWithCallback(this, 100, 
                                              Components.interfaces.nsITimer.TYPE_ONE_SHOT);
          }
          else {
            ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
                       getService(Components.interfaces.nsIPromptService);
            ps.alert(null, "Achtung",
                     "Es ist ein interner Fehler aufgetreten!");
            
            this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                      "cyCheckyesss.jumpQuickLink",
                                      "Could not find name for quick-link #"+
                                      aIndex+"!", false);
          }
        }
      }
    }
  },
  
  /**************************************************************/
  /**************************************************************/
  
  QueryInterface: function (iid) {
    if (!iid.equals(Components.interfaces.cyICheckyesssManager) &&
        !iid.equals(Components.interfaces.cyICheckyesssWebCallback) &&
        !iid.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
  
};


/******************************************************************************
 * XPCOM Functions for construction and registration
 ******************************************************************************/
var gModule = {
  _firstTime: true,
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {
    if (this._firstTime) {
      this._firstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    }
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, CONTRACT_ID, aFileSpec, aLocation, aType);
  },
  
  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        
  },
  
  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return gFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) { return true; }
};

var gFactory = {
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    return (new CheckyesssYesssManager()).QueryInterface(aIID);
  }
};

function NSGetModule(aCompMgr, aFileSpec) { return gModule; }
