/* ***** BEGIN LICENSE BLOCK *****
   - Version: MPL 1.1/GPL 2.0/LGPL 2.1
   -
   - The contents of this file are subject to the Mozilla Public License Version
   - 1.1 (the "License"); you may not use this file except in compliance with
   - the License. You may obtain a copy of the License at
   - http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS IS" basis,
   - WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
   - for the specific language governing rights and limitations under the
   - License.
   -
   - The Original Code is CheckYesss.
   -
   - The Initial Developer of the Original Code is
   - Thomas Reitmayr <treitmayr@yahoo.com>.
   - Portions created by the Initial Developer are Copyright (C) 2007-2008
   - the Initial Developer. All Rights Reserved.
   -
   - Contributor(s):
   - - Christian Wallner <christian_wallner@gmx.at>
   -
   - Alternatively, the contents of this file may be used under the terms of
   - either the GNU General Public License Version 2 or later (the "GPL"), or
   - the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
   - in which case the provisions of the GPL or the LGPL are applicable instead
   - of those above. If you wish to allow use of your version of this file only
   - under the terms of either the GPL or the LGPL, and not to allow others to
   - use your version of this file under the terms of the MPL, indicate your
   - decision by deleting the provisions above and replace them with the notice
   - and other provisions required by the LGPL or the GPL. If you do not delete
   - the provisions above, a recipient may use your version of this file under
   - the terms of any one of the MPL, the GPL or the LGPL.
   -
   - ***** END LICENSE BLOCK ***** */

const CLASS_ID = Components.ID("{37c49c70-88d0-11dd-ad8b-0800200c9a66}");
const CLASS_NAME = "Checkyesss Bob Manager Component";
const CONTRACT_ID = "@mozdev.org/checkyesss/bob_manager;1";

const TO_HTTP = 120;              // 120 seconds until aborting stalled access
const TO_MIN_PAUSE = 10;          // 10 minutes pause between scheduled accesses
const TO_BLOCKED = 60;            // 60 seconds pause if access was blocked
const TO_MIN_CALLUPD = 6;         // 6 hours min. pause for calls (debug: set to 0!)
const TO_CALLUPD = 7*24;          // 7 days regular interval for calls

const MAX_CARDS = 10;

const DEBUG = 0;


/***************************************************************************/

/**
/* ArrayEnumerator courtesy of Brendan Eich <brendan@mozilla.org>
 */
function ArrayEnumerator(array) {
    this.array = array;
    this.index = 0;
}
ArrayEnumerator.prototype = {
    hasMoreElements: function () { return this.index < this.array.length; },
    getNext: function () { return (this.index < this.array.length) ?
                       this.array[this.index++] : null; }
};
// nsISupportsArrayEnumerator
//
function nsISupportsArrayEnumerator(array) {
    this.array = array;
    this.index = 0;
}
nsISupportsArrayEnumerator.prototype = {
    hasMoreElements: function() { return this.index < this.array.Count(); },
    getNext: function () { return (this.index < this.array.Count()) ?
                       this.array.GetElementAt(this.index++) : null; }
};

/**
/* StringEnumerator
 */
function StringEnumerator(array) {
    this.array = array;
    this.index = 0;
}
StringEnumerator.prototype = {
    hasMore: function() { return this.index < this.array.length; },
    getNext: function () { return (this.index < this.array.length) ?
                       this.array[this.index++] : null; }
};

/***************************************************************************/

/**
/* 
 */
function AccountDetails(aId, aUser, aPassword, aPlan, aEnabled, aMayStorePwd) {
  this._id = aId;
  this._user = aUser;            // user name
  this._pw = ((typeof(aPassword) == "string") && (aPassword.length > 0)) ?
               aPassword : null;          // password
  this._mayStorePwd = aMayStorePwd;   // may the password be stored in PW-Mng?
  this._en = aEnabled;           // account enabled
  this._rateplan = aPlan;        // selected plan
  this._phone = "??";            // phone number
  this._credit = "";             // current credit ("euro.cent")
  this._messages = 0;            // number of new messages
  this._dateCallsLoaded = 0;     // last time calls where updated
  this._error = "Daten noch nicht geladen";
  this._forceupdate = false;     // force update!!
  this._quickLinkPending = false;
  this._quickLinkWindow = null;
  this._quickLinkTarget = "";
  this._reloadAvailableCalls = false;
//  this._actual_rate = 1;         // 0..basic, 1..einheitstarif, 2..flat
}

/******************************************************************************
 * CheckyesssBobManager component
 ******************************************************************************/

function CheckyesssBobManager() {
  this.wrappedJSObject = this;
  this._start();
}


CheckyesssBobManager.prototype = {
  
  _providerId:    "BB001",
  
  _bobLogin:    "https://rechnung.bob.at/login.do",
  _bobLogout:   "https://rechnung.bob.at/logout.do",
  _bobCalls:    "https://rechnung.bob.at/ben.set.do?tcd=1",   // TODO
  
  _blockingUrlRegExp: "^https?://rechnung\\.bob\\.at",
  
  _timeHour:   3600000,
  _timeMinute: 60000,
  _timeSecond: 1000,
  
  _observerTopic: null,
  _userRegExp: null,
  
  _interval: 24,                     // interval [hours]
  _suspendUntil: 0,                  // suspend automatic update from internet
  
  _accountsArray: null,
  _accountsChanged: false,
  _accountsTransferred: null,
  
  _webAccess: null,
  _accessAccount: -1,
  _lastAccess: 0,     // date of last access [ms] (not counting logout)
  _checkPending: false,
  
  _firstTime: true,
  
  _checkTimer: null,
  _obs: null,
  _utils: null,
  _userAgent: null,
  _managerCenter: null,
  
  
  /**************************************************************/
  
  _start: function() {
    // set components
    this._utils = Components.classes["@mozdev.org/checkyesss/util_service;1"].
                getService(Components.interfaces.cyICheckyesssUtil);
    
    this._checkTimer = Components.classes["@mozilla.org/timer;1"].
                         createInstance(Components.interfaces.nsITimer); 
    
    this._obs = Components.classes["@mozilla.org/observer-service;1"].
                  getService(Components.interfaces.nsIObserverService);
    
    this._managerCenter = Components.classes["@mozdev.org/checkyesss/manager_center;1"].
                            getService(Components.interfaces.cyICheckyesssCenter);
    
    var providerProperties = this._managerCenter.getProviderProperties(this._providerId);
    this._observerTopic = providerProperties.getPropertyAsString("topic");
    this._userRegExp = providerProperties.getPropertyAsString("validUserRegExp");
    
    //add observer topics
    this._obs.addObserver(this, "xpcom-shutdown", false);
    
    this._webAccess = Components.classes["@mozdev.org/checkyesss/web_access;1"].
                        createInstance(Components.interfaces.cyICheckyesssWeb);
    this._webAccess.requestTimeout = TO_HTTP;
    
    try {
      this._userAgent = Components.classes["@mozilla.org/network/protocol;1?name=http"].
                          getService(Components.interfaces.nsIHttpProtocolHandler).
                            userAgent;
      this._userAgent += " Checkyesss/0.6";
    }
    catch (e) {
      this._userAgent = null;
    }
    
    if (DEBUG) {
      this._bobLogin  = "http://localhost/sy/index.html";
      this._bobLogout = "http://localhost/sy/abmelden.html";
      this._bobCalls = "http://localhost/sy/verbindungen-%.html";
      
      this._timeHour = this._timeMinute;
      this._timeMinute = this._timeSecond;
    }
    
    //this._checkTimer.initWithCallback(this, 10, 
    //                                  Components.interfaces.nsITimer.TYPE_ONE_SHOT);
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                              "Checkyesss Bob-Manager started", "", false);
    
    this._firstTime = false;
  },
  
  /**************************************************************/
  
  _stop: function() {
    this._webAccess.endSession();
    
    // stop timers
    this._checkTimer.cancel();
    
    // clean up observers
    try {
      this._obs.removeObserver(this, "xpcom-shutdown");
    }
    catch(e) {}
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                              "Checkyesss Bob-Manager stopped", "", false);
    
    //destroy variables
    this._webAccess = null;
    //this._busy = false;
    this._jumpWin = null;
    this._checkTimer = null;
    this._errtxt = null;
    //this._loginCallback = null;
    this._firstTime = true;
    this._accountsArray = null;
    this._utils = null;
    this._managerCenter = null;
  },
  
  /**************************************************************/
  
  get updateInterval() {
    return this._interval;
  },
  
  set updateInterval(aValue) {
    this._interval = (aValue <  4) ? 4 :
                     (aValue > 24) ? 24 :
                     aValue;
    return aValue;
  },
  
  /**************************************************************/
  
  get suspendNetAccessUntil() {
    return this._suspendUntil;
  },
  
  set suspendNetAccessUntil(aValue) {
    this._suspendUntil = aValue;
    return aValue;
  },
  
  /**************************************************************/
  
  updateClients: function() {
    this._refreshDisplay();
  },
  
  /**************************************************************/
  
  clearAccounts: function() {
    if (!this._accountsArray ||
        (this._accountsArray.length == 0))
      return;
    
    this.abortAccess();
    
    this._accountsArray = new Array();
    
    this._obs.notifyObservers(this, this._observerTopic, "reset_all");
  },

  /**************************************************************/
  
  addAccount: function(aAccountId, aUser, aPassword, aRatePlanId,
                       aPeriodReset, isEnabled, aMayStorePwd) {
    var result = true;
    
    if (!this._accountsArray)
      this._accountsArray = new Array();
    var index = this._accountsArray.length;

    var entry = new AccountDetails(aAccountId, aUser, aPassword, aRatePlanId, 
                                   isEnabled, aMayStorePwd);
    
    var ratePlanOk = false;
    var ratePlanIds = this._managerCenter.getRatePlanList(this._providerId);


    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
			      "cyCheckyesssBob.addAccount",
			      "ID '" + aAccountId + "' User '" + aUser + "'",
			      false);

    while (!ratePlanOk && ratePlanIds.hasMore())
      ratePlanOk = (ratePlanIds.getNext() == aRatePlanId);
    
    if (!ratePlanOk) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssBob.addAccount",
                                "Invalid rate plan ID '" + aRatePlanId + "'",
                                false);
      entry._en = false;
      entry._rateplan = "auto";
      result = false;
    }
    
    this._accountsArray[index] = entry;
    
    if (entry._en) {
      this._updateFromCache(index);
      
      // restart timer with a short timeout
      this._checkTimer.cancel();
      this._checkTimer.initWithCallback(this, 100, 
                                        Components.interfaces.nsITimer.TYPE_ONE_SHOT);
    }
    return result;
  },
  
  /**************************************************************/
  
  _assembleCreditInfo: function(aIndex, aDocument) {
    var entry;
    try {
      entry = this._accountsArray[aIndex];
    }
    catch (e) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssBob._assembleCreditInfo",
                                "Invalid internal index " + aIndex, false);
      return null;
    }
    
    if (!entry._en)
      return null;
    
    var expired = true;
    var age = -1;
    
    if (this._utils.existCache(entry._id)) {
      age = this._utils.getCacheAge(entry._id);
      expired = (this._utils.getCacheValidity(entry._id) == 0);
    }
    
    var container = Components.classes["@mozdev.org/checkyesss/container;1"].
                      createInstance(Components.interfaces.cyICheckyesssContainer);
    
    var tooltip = null;
    var pnum = entry._phone;
    var isError = (entry._error || (entry._credit.toString().length == 0)) ? true : false;
    
    if (aDocument) {
      // the following should be done with some transformation really...
      var descr, ltext;
      var lines;
      
      tooltip = aDocument.createDocumentFragment();
      
      if (entry._error) {
        lines = entry._error.split("\n");
        for (var i = 0; i < lines.length; i++) {
          descr = aDocument.createElement("description");
          descr.setAttribute("class", "checkyesss-class-ttval");
          ltext = aDocument.createTextNode(lines[i]);
          descr.appendChild(ltext);
          tooltip.appendChild(descr);
        }
      }
      else if (entry._credit.toString().length == 0) {
        descr = aDocument.createElement("description");
        ltext = aDocument.createTextNode("Keine Daten vorhanden.");
        descr.appendChild(ltext);
        tooltip.appendChild(descr);
      }
      else {
        // credit: compute number of minutes and sms
        var credit = parseFloat(entry._credit);
        var min_bob, min_rest;
        var sms = "";
        var hours_bob = "";
        var hours_rest = "";
        
/*        if (entry._actual_rate == 0) {   // 'basic' rate
          min_bob = parseInt(credit / 0.05, 10);
          min_rest = parseInt(credit / 0.15, 10);
          sms = parseInt(credit / 0.10, 10);
          
          hours_bob = this._utils.minutesToString(min_bob) + " zu bob";
          hours_rest = this._utils.minutesToString(min_rest) + " in andere Netze";
          sms = sms + " SMS";
        }
        else
        if (entry._actual_rate == 1) {   // 'einheitstarif' rate
          min_bob = parseInt(credit / 0.09, 10);
          sms = parseInt(credit / 0.09, 10);
          
          hours_bob = this._utils.minutesToString(min_bob) + " in alle Netze";
          sms = sms + " SMS";
        }
        else
        if (entry._actual_rate == 2) {   // 'flat' rate
          min_rest = parseInt(credit / 0.15, 10);
          sms = parseInt(credit / 0.10, 10);
          
          hours_bob = "unbegrenzt zu Bob & Festnetz";
          hours_rest = this._utils.minutesToString(min_rest) + " in andere Netze";
          sms = sms + " SMS";
        }
        else {
          hours_bob = "Fehler!";
          sms = "Unbekannter Tarif!";
          isError = true;
        }
        
        var node = aDocument.createElement("label");
        node.setAttribute("value", "Guthaben entspricht");
        node.setAttribute("class", "checkyesss-class-ttlab");
        tooltip.appendChild(node);
        
        node = aDocument.createElement("label");
        node.setAttribute("value", hours_bob);
        node.setAttribute("class", "checkyesss-class-ttval");
        node.setAttribute("style", "margin-left: 1.5em;");
        tooltip.appendChild(node);
        
        if (hours_rest) {
          node = aDocument.createElement("label");
          node.setAttribute("value", hours_rest);
          node.setAttribute("class", "checkyesss-class-ttval");
          node.setAttribute("style", "margin-left: 1.5em;");
          tooltip.appendChild(node);
        }
        
        node = aDocument.createElement("label");
        node.setAttribute("value", sms);
        node.setAttribute("class", "checkyesss-class-ttval checkyesss-class-ttboxhdr");
        node.setAttribute("style", "margin-left: 1.5em;");
        tooltip.appendChild(node);
*/      }
    }
    
    container.setPropertyAsString("phoneNumber", pnum);
    container.setPropertyAsString("phoneCredit", entry._credit.toString());
    container.setPropertyAsInt("dataAge", age);
    container.setPropertyAsBool("dataExpired", expired);
    container.setPropertyAsBool("dataError", isError);
    container.setPropertyAsISupports("docTooltip", tooltip);
    
    return container;
  },
  
  /**************************************************************/
  
  getCreditInfo: function(aId, aDocument) {
    if (this._accountsArray) {
      for (var i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._id == aId)
          return this._assembleCreditInfo(i, aDocument);
      }
    }
    return null;
  },
  
  /**************************************************************/
  
  setManualAccountingStart: function(aAccountId, aStartTime) {
    null;
  },
  
  /**************************************************************/
  
  detectedRatePlan: function(aAccountId) {
    return "auto";
  },
  
  /**************************************************************/
  
  _refreshDisplay: function() {
    if (this._accountsArray) {
      for (var i = 0; i < this._accountsArray.length; i++) {
        entry = this._accountsArray[i];
        if (entry._en)
          this._obs.notifyObservers(this, this._observerTopic,
                                    "update::" + entry._id);
      }
    }
  },
  
  /**************************************************************/
  
  reloadCredit: function(aId) {
    var index = -1;
    
    if (this._accountsArray) {
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._id == aId) {
          if (this._accountsArray[i]._en)
            index = i;
          break;
        }
      }
    }
    
    if (index >= 0) {
      var win = this._utils.isWebPageOpen(this._blockingUrlRegExp);
      if (win) {
        var ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
                   getService(Components.interfaces.nsIPromptService);    
        ps.alert(null, "Achtung",
                 "Das Guthaben kann derzeit nicht aktualisiert werden,\n"+
                 "da gerade ein Bob-Konto angezeigt wird!\n\n"+
                 "Die Aktion wird verz\u00f6gert, bis alle Web-Konten\n"+
                 "geschlossen wurden.");
        
        // focus the mentioned window
        //win.focus();
        
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                  "cyCheckyesssBob.reloadCredit",
                                  "Conflicting browser window detected, action deferred",
                                  false);
      }
      
      this._accountsArray[index]._forceupdate = true;
      this._checkPending = true;
      
      // restart timer with a short timeout
      this._checkTimer.cancel();
      this._checkTimer.initWithCallback(this, 100, 
                                        Components.interfaces.nsITimer.TYPE_ONE_SHOT);
    }
  },
  
  /**************************************************************/
  
  updateCalls: function(aId, aForced) {
    var result = false;
    var index = -1;
    
    if (this._accountsArray) {
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._id == aId) {
          if (this._accountsArray[i]._en)
            index = i;
          break;
        }
      }
    }
    
    if (index >= 0) {
      var entry = this._accountsArray[index];
      var ps;
      var win = this._utils.isWebPageOpen(this._blockingUrlRegExp);
      if (win) {
        if (aForced) {
	  ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
	                  getService(Components.interfaces.nsIPromptService);    
	  ps.alert(null, "Achtung",
		   "Die Einzelgespr\u00E4che k\u00f6nnen derzeit nicht aktualisiert\n"+
		   "werden, da gerade ein Bob-Konto angezeigt wird!\n\n"+
		   "Bitte zuerst ausloggen und dann Aktion wiederholen!");
            
	  // focus the mentioned window
	  //win.focus();
            
	  this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
				    "cyCheckyesssBob.updateCalls",
				    "Conflicting browser window detected, action aborted",
				    false);
	}
      }
      else {
        var now = (new Date()).valueOf();
	if (aForced ||
           (entry._dateCallsLoaded + (TO_MIN_CALLUPD * this._timeHour) < now)) {
          entry._forceupdate = true;
	  entry._dateCallsLoaded = 0;
	  this._checkPending = true;
            
	  /* We need to circumvent the timer here as we have to make sure
	     that the access is performed and a notification is returned!
	     The usual 100ms timeout would open a time windows where
	     conditions could change unexpectedly! */
	  this._checkAccounts();
            
	  result = true;      // expect a notification
	}
      }
    }
    return result;
  },
  
  /**************************************************************/
  
  getCallsRecords: function(aKey, aFromTime, aToTime) {
    var sum_credit = 0.0;
    var sum_fee = 0.0;         // fee for calls + sms
    var sum_cfee = 0.0;        // fee for all calls
    var sum_duration = 0;
    var sum_charged = 0;       // duration charged for
    var callsArray = new Array();
    var call, duration, fee, credit, charged, type, number;
    
    var calls = this._utils.getCallsRecords(aKey, aFromTime, aToTime, null);
    
    while (calls.hasMoreElements()) {
      call = calls.getNext().QueryInterface(Components.interfaces.cyICheckyesssContainer);
      
      number   = call.hasProperty("number") ? call.getPropertyAsString("number") : "";
      duration = call.hasProperty("duration") ? call.getPropertyAsInt("duration") : 0;
      fee      = call.hasProperty("fee") ? call.getPropertyAsString("fee") : "0";
      credit   = call.hasProperty("credit") ? call.getPropertyAsString("credit") : "0";
      type     = call.hasProperty("type") ? call.getPropertyAsString("type") : "";
      
      // replace Austrian country code by "0" to make numbers more readable
      number = number.replace(/^\+43/, "0");
      call.setPropertyAsString("number", number);
      
      fee    = Number(fee.replace(/[\.,]/, "."));
      credit = Number(credit.replace(/[\.,]/, "."));
      
      sum_duration += duration;
      sum_fee      += fee;
      sum_credit   += credit;
      
      if (type != "call")
        call.deleteProperty("duration");
      else
        sum_cfee += fee;
      
      if (type == "call" || type == "sms" || type == "data") {
        fee = Math.round(fee * 10000.0);        // round to 1/100 cent
        fee = "00000" + fee.toString();
        fee = fee.replace(/^0*(\d+?)(\d{4})$/, "$1,$2");
        call.setPropertyAsString("fee", fee);
      }
      else
        call.deleteProperty("fee");
      
      if (credit > 0) {
        credit = credit.toString();
        call.setPropertyAsString("credit", credit);
      }
      else if (call.hasProperty("credit"))
        call.deleteProperty("credit");
      
      callsArray.push(call);
    }
    
    if (callsArray.length > 0) {
      call = Components.classes["@mozdev.org/checkyesss/container;1"].
               createInstance(Components.interfaces.cyICheckyesssContainer);
      call.setPropertyAsBool("statistic", true);
      call.setPropertyAsString("description", "Summe:");
      
      if (sum_duration > 0) {
        call.setPropertyAsInt("duration", sum_duration);
        call.setPropertyAsString("tcharged", this._utils.durationToString(sum_charged));
      }
      if (sum_fee > 0) {
        fee = Math.round(sum_fee * 10000.0);        // round to 1/100 cent
        fee = "00000" + fee.toString();
        fee = fee.replace(/^0*(\d+?)(\d{4})$/, "$1,$2");
        fee = "\u20AC " + fee;
        call.setPropertyAsString("fee", fee);
      }
      if (sum_credit > 0) {
        credit = "\u20AC " + sum_credit.toString();
        call.setPropertyAsString("credit", credit);
      }
      callsArray.push(call);
      
      if (sum_duration > 0 && sum_cfee > 0) {
        call = Components.classes["@mozdev.org/checkyesss/container;1"].
                 createInstance(Components.interfaces.cyICheckyesssContainer);
        call.setPropertyAsBool("statistic", true);
        call.setPropertyAsString("description", "Durchschnitt:");
        var average = (sum_cfee * 600000.0) / sum_duration;
        average = Math.round(average);                   // round to 1/100 cent
        fee = "00" + average.toString();
        fee = fee.replace(/^0*(\d+?)(\d{2})$/, "$1,$2");
        fee = fee + " \u00A2/min";                       // cent per minute
        call.setPropertyAsString("fee", fee);
        callsArray.push(call);
      }
    }
    
    return new ArrayEnumerator(callsArray);
  },
  
  /**************************************************************/
  
  enableReloadAvailableCalls: function(aAccountId) {
    //dump("enableReloadAvailableCalls\n");
    var entry = null;
    var index = -1;
    
    if (this._accountsArray) {
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._id == aAccountId) {
          entry = this._accountsArray[i];
          index = i;
          break;
        }
      }
    }
    
    if (entry) {
      entry._reloadAvailableCalls = true;
    }
  },
  
  /**************************************************************/
  
  _updateFromCache: function(aIndex) {
    //dump("_updateFromCache\n");
    var entry = this._accountsArray[aIndex];
    var valid = false;
    
    if (entry._en &&
        this._utils.existCache(entry._id)) {
      var contents = this._utils.getCacheContents(entry._id);
      
      if (contents) {
        var err = (contents.hasProperty("error")) ? contents.getPropertyAsString("error") : "";
        // convert NL due to moz-bug #169521
        entry._error = err.replace(/\\n/g, "\n");
        
        entry._phone =
          (contents.hasProperty("phone")) ? contents.getPropertyAsString("phone") : "??";
        
        entry._credit =
          (contents.hasProperty("credit")) ? contents.getPropertyAsString("credit") : "";
        
//        entry._actual_rate =
//          (contents.hasProperty("actual-rate")) ? contents.getPropertyAsInt("actual-rate") : 1;
        
        entry._dateCallsLoaded =
          (contents.hasProperty("dateCallsLoaded")) ? contents.getPropertyAsInt("dateCallsLoaded") : 0;
        
        if (entry._error.length || entry._credit.toString().length)
          valid = true;
      }
    }
    
    if (!valid) {
      this._accountsArray[aIndex]._error = "Keine Daten vorhanden.";
      this._accountsArray[aIndex]._credit = "";
      this._accountsArray[aIndex]._credit_till = "";
    }
    
    this._obs.notifyObservers(this, this._observerTopic,
                              "update::" + entry._id);
  },
  
  /**************************************************************/
  
  _anyForcedUpdates: function() {
    if (this._accountsArray) {
      var i;
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._en &&
            this._accountsArray[i]._forceupdate)
          return i;
      }
    }
    return -1;
  },
  
  /**************************************************************/
  
  _anyQuickLinkUpdates: function() {
    if (this._accountsArray) {
      var i;
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._en &&
            this._accountsArray[i]._quickLinkPending)
          return i;
      }
    }
    return -1;
  },
  
  /**************************************************************/
  
  _needsUpdate: function(aIndex) {
    var entry;
    try {
      entry = this._accountsArray[aIndex];
    }
    catch (e) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "_needsUpdate",
                                "Invalid internal index " + aIndex, false);
      return "";
    }
    
    if (entry._en) {
      if (entry._quickLinkPending)
        return "link";
      
      var now = (new Date()).valueOf();
      /* Commented out to make sure this does not frequently try to
       * update call details!
      if (((entry._dateCallsLoaded + (TO_CALLUPD * this._timeHour) < now) ||
	   entry._reloadAvailableCalls))// && !entry._simchef)
        return "calls";
      */
      
      if (entry._forceupdate ||
          !this._utils.existCache(entry._id) ||
          this._utils.getCacheValidity(entry._id) == 0)
        return "credit";
    }
    
    return "";
  },
  
  /**************************************************************/
  
  _checkAccounts: function() {
    //dump("_checkAccounts\n");
    try {
      this._checkTimer.cancel();
    }
    catch(e) { }
    
    this._checkPending = false;
    
    if (!this._accountsArray)
      return;
    
    if (this._webAccess.activeSession) {
      this._checkPending = true;
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesssBob._checkAccounts",
                                "Another access already active!",
                                false);
      return;
    }
    
    var type = "link";
    var index = this._anyQuickLinkUpdates();
    
    if (index < 0) {
      index = this._anyForcedUpdates();
      if (index >= 0)
        type = this._needsUpdate(index);
    }
    
    if (index < 0) {
      //dump("no forced updates\n");
      // no forced updates available
      var now = (new Date()).valueOf();
      if ((now - this._lastAccess) > (TO_MIN_PAUSE * this._timeMinute) &&
          now > this._suspendUntil) {
        // ensure a minimum break of TO_MIN_PAUSE minutes between accesses
        //dump("check for scheduled updates\n");
        var i;
        for (i = 0; i < this._accountsArray.length; i++) {
          type = this._needsUpdate(i);
          if (type) {
            //dump("type = " + type + "\n");
            index = i;
            break;
          }
        }
      }
    }
    
    if (index >= 0 && !this._utils.isWebPageOpen(this._blockingUrlRegExp)) {
      this._accessAccount = index;
      this._startWebPath(type);
    }
    else
      this._rescheduleCheck();
  },
  
  /**************************************************************/
  
  _rescheduleCheck: function() {
    //dump("_rescheduleCheck\n");
    if (!this._accountsArray)
      return;
    
    var timeout = 0;
    var blocked = this._utils.isWebPageOpen(this._blockingUrlRegExp);
    
    if (this._checkPending ||
        this._anyQuickLinkUpdates() >= 0 ||
        this._anyForcedUpdates() >= 0) {
      //dump("  immediate check\n");
      if (!blocked) {
        this._checkAccounts();
        return;
      }
    }
    else {
      //dump("  scheduled check\n");
      var now = (new Date()).valueOf();
      var mintimeout = Math.max((TO_MIN_PAUSE * this._timeMinute) - (now - this._lastAccess),
                                this._suspendUntil - now);
      if (mintimeout < 0)
        mintimeout = 0;
      
      timeout = 24 * this._timeHour;              // max. timeout
      for (i = 0; i < this._accountsArray.length; i++) {
        var entry = this._accountsArray[i];
        if (entry._en) {
          if (!this._utils.existCache(entry._id)) {
            //dump("  no cache found for " + entry._id + "\n");
            timeout = 0;
            break;
          }
          else {
            var valid = this._utils.getCacheValidity(entry._id);
            //dump("  cache for " + entry._id + " valid " + valid + "ms\n");
            if (valid < timeout)
              timeout = valid;
          }
        }
      }
      
      if (timeout < mintimeout)
        timeout = mintimeout;
      
      timeout += 1000;               // add 1 second tolerance
    }
    
    if (blocked && (timeout < (TO_BLOCKED * this._timeSecond))) {
      timeout = TO_BLOCKED * this._timeSecond;
      /*this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesssBob._rescheduleCheck",
                                "Scheduled check was delayed by open web page!",
                                false);*/
    }
    
    //dump("timeout = " + timeout + "\n");
    this._checkTimer.initWithCallback(this, timeout,
                                      Components.interfaces.nsITimer.TYPE_ONE_SHOT);
  },
  
  /**************************************************************/
  /**************************************************************/
  
  _webPath: {
    "credit": [ "submit_login",             // log in
                "exit/err",                 // exit on error (eg. missing password)
                "parse_front_page",         // parse the front page
                "exit/err",                 // error parsing front page
                "logout",                   // log out (only if we had to log in!)
                "exit" ],
    
    "calls": [  "throttle 1000",            // pause for 1s after each http access
                "submit_login",             // log in
                "exit/err",                 // exit on error (eg. missing password)
                "parse_front_page",         // parse the front page
                "exit/err",                 // error parsing front page
/*                "prepare_calls",            // do some preparation
                "again:",                   //                   <-\
                "fetch_calls",              // request the list     |
                "exit/err",                 //                      |
                "parse_calls",              // extract the calls    |
                "jump/cont again",          //                    _/
                "store_calls",              // finally store the new calls
*/
                "logout",                   // log out (only if we had to log in!)
                "exit"
],
    
    "link":   [ "submit_login",             // then log in
                "jump/ok check",
                "browser_error",
                "exit",                     // exit on error (eg. missing password)
                "check:",
                "parse_front_page",         // parse the front page
                "jump/ok showpage",
                "browser_error",
                "exit",                     // exit on error
                "showpage:",
                "browser_goto",
                "exit" ]
  },
  _webPathType: null,
  
  
  _prevFunctionName: "",
  
  webSeqCallback: function(aFunctionName) {
    //dump(">"+aFunctionName+"\n");
    var prefFN = this._prevFunctionName;
    this._prevFunctionName = aFunctionName;
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                              "cyCheckyesssBob.webSeqCallback", aFunctionName, false);
    
    switch (aFunctionName) {
      case "submit_login":     return this._webLogin();
      case "parse_front_page": return this._webLoginDone();
      case "prepare_calls":    return this._webCallsPrepare();
      case "fetch_calls":      return this._webCalls();
      case "parse_calls":      return this._webCallsParse();
      case "store_calls":      return this._webCallsStore();
      case "browser_goto":     return this._webBrowserGoto(false);
      case "browser_error":    return this._webBrowserGoto(true);
      case "logout":           return this._webLogout();
      case "exit":             return this._webExit();
      case "error":            return this._webError(prefFN);
      default:
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesssBob.webSeqCallback",
                                  "Unknown function name " + aFunctionName,
                                  false);
        return "";
    }
  },
  
  /**************************************************************/
  
  _startWebPath: function(aPathType) {
    //dump("_startWebPath\n");
    if (!aPathType in this._webPath) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssBob._startWebPath",
                                "Invalid path type " + aPathType,
                                false);
      return;
    }
    
    if (this._webAccess.activeSession) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssBob._startWebPath",
                                "Session already active!",
                                false);
      return;
    }
    
    this._webPathType = aPathType;
    this._webAccess.startRequestSequence(new StringEnumerator(this._webPath[aPathType]),
					 this);
  },
  
  /**************************************************************/
  
  _webLogin: function() {
    //dump("_webLogin\n");
    var entry = this._accountsArray[this._accessAccount];
    var result = "ok";
    
    entry._forceupdate = false;
    this._lastAccess = (new Date()).valueOf();
    
    this._obs.notifyObservers(this, this._observerTopic,
                                    "net-busy::" + entry._id);
    if (entry._pw == null) {
      var password = this._managerCenter.getPassword(this._providerId,
                                                     entry._id, entry._user,
                                                     !entry._mayStorePwd, true);
      if (password != null) {
        entry._pw = password;
      }
      else {
        this._webStoreError("Passwort nicht gefunden!");
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesssBob._webLogin",
                                  "Could not find password for account " + entry._id,
                                  false);
        result = "err";
      }
    }
    
    var check = new RegExp(this._userRegExp);
    if (!entry._user.match(check)) {
      this._webStoreError("Ung\u00FCltige Telefonnummer!");
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesssBob._webLogin",
                                "Invalid format of user name " + entry._user,
                                false);
      result = "err";
    }
    
    if (result == "ok") {
      // remove any unneccessary characters, just to be sure...
      var cleanUser = entry._user.replace(/[\s\/]/g, "");
      cleanUser = cleanUser.replace(/^0/, "+43");
      // submit the HTTP request
      var url = Components.classes["@mozilla.org/network/io-service;1"].
                  getService(Components.interfaces.nsIIOService).
                    newURI(this._bobLogin, null, null);
      this._webAccess.startRequest(url, "POST", null,
                                   "logon=prc&username=" + encodeURIComponent(cleanUser) +
                                   "&password=" + encodeURIComponent(entry._pw),
                                   null);
    }
    return result;
  },
  
  /**************************************************************/
  
  _webLoginDone: function() {
    //dump("_webLoginDone\n");
    var result = "ok";
    var entry = this._accountsArray[this._accessAccount];
    
    var response = this._webAccess.contentsHTML;
    var matches;
    
    if (response.match(/Costmanager/)) {
	matches = response.match(/\s+([0-9]+)(?:[,.]([0-9]+))\s*\(inkl\.\sUSt\.\)/);
    }

    if (matches) {
      var contents = this._utils.getCacheContents(entry._id);
      if (!contents)
        contents = Components.classes["@mozdev.org/checkyesss/container;1"].
                     createInstance(Components.interfaces.cyICheckyesssContainer);
      
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "index.do - Match", matches[0], false);
      //dump("Match: "+matches[0]+"\n");
      if (!matches[2])
        matches[2] = "00";
      entry._credit = matches[1]+"."+matches[2];

//      get actual-rate
      
      entry._error = "";
      entry._phone = "??";
      
      matches = response.match(/Handynummer[^0-9]*(\d[\/\d\s]+)/i);
      
      entry._phone = (matches) ? matches[1] : entry._user;
      entry._phone = entry._phone.replace(/\s+$/, "");  // remove trailing spaces
      
      this._utils.appendCreditRecord(entry._id, entry._credit.toString());
      contents.setPropertyAsString("phone", entry._phone);
      contents.setPropertyAsString("credit", entry._credit.toString());
//      contents.setPropertyAsInt("actual-rate", entry._actual_rate);
      contents.setPropertyAsString("error", "");
      
      this._utils.writeCache(entry._id,
                             contents,
                             this._interval * this._timeHour,
                             true);
      
      this._obs.notifyObservers(this, this._observerTopic,
                                "update::" + entry._id);
    }
    else {
      var error;
      
      if (response.match(/\s+ist\s+falsch/)) {
        error = "Fehler beim Einloggen! Passwort oder Handynummer falsch!";
        result = "err";
      }
      else
        error = "Fehler!! Guthaben\nnicht gefunden.";
      
      this._webStoreError(error);
      
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesssBob._webLoginDone",
                                this._webAccess.contentsHTML.
                                  replace(/\btype="password"(.*?\bvalue=)".+?"/,
                                          " $1\"removed\""),
                                true);
    }

    return result;
  },
  
  /**************************************************************/
  
  _callsLastTime: 0,         // time of last stored call record
  _callsCheckTime: 0,        // check calls since _callsCheckTime (overlapping!)
  _callsLoadTime: 0,         // load calls since _callsLoadTime (overlapping!)
  _callsNewLast: 0,          // time of newest calls record loaded
  
  _callsPageNo: 0,           // parameters for loading the calls pages
  _callsMaxPageNo: -1,       //  :
  _callsMonthNo: 0,          //  :
  _callsYearNo: 0,           //  :
  
  _callsFound: null,         // array of loaded calls records
  
  _callsMissingFrom: 0,      // start of section missing calls
  _callsMissingTo: 0,        // end of section
  
  
  _webCallsPrepare: function() {
    //dump("_webCallsPrepare\n");
    var entry = this._accountsArray[this._accessAccount];
    var now = new Date();
    var date;
    
    if (entry._reloadAvailableCalls) {
      // we HAVE to reload all calls we can find
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesssBob._webCallsPrepare",
                                "Reloading all available calls "+
                                "for account '"+entry._id+"'",
                                false);
      entry._reloadAvailableCalls = false;
      this._callsLastTime = 0;
    }
    else {
      // this is the regular case
      this._callsLastTime = this._utils.getLastCallTime(entry._id);
    }
    
    // load calls overlapping 2 days
    this._callsCheckTime = this._callsLastTime - (2 * 24 * this._timeHour);
    if (this._callsCheckTime > 0) {
      // round to midnight
      date = new Date(this._callsCheckTime);
      date = new Date(date.getFullYear(), date.getMonth(), date.getDate(), 0, 0, 0, 0);
    }
    else {
      if (entry._dateCallsLoaded > 0) {
        // apparently we already checked before but did not find any calls
        // -> recheck starting 3 days before that date
        // WARNING: This may lead to problems in the following situation:
        //   Someone mistakenly chose a different phone provider, where the
        //   calls update of course fails. Then she/he corrects the provider
        //   in the settings and now the calls will not be downloaded from
        //   Jan 2nd 2006 but from the date determined one line below!!!!!
        date = new Date(entry._dateCallsLoaded - (3 * 24 * this._timeHour));
        date = new Date(date.getFullYear(), date.getMonth(), date.getDate(), 0, 0, 0, 0);
      }
      else {
        // start checking at Dec 2nd 2006
        date = new Date(2006, 11, 2, 0, 0, 0, 0);
      }
    }
    this._callsCheckTime = date.valueOf();
    date = new Date(date.valueOf() - 1);
    this._callsLoadTime = (new Date(date.getFullYear(), date.getMonth(), date.getDate(), 0, 0, 0, 0)).valueOf();
    this._callsNewLast = 0;
    //dump("_callsCheckTime = "+this._callsCheckTime+"\n");
    
    // parameters to start with - downloads go from newest to oldest calls
    this._callsPageNo = 0;
    this._callsMaxPageNo = -1;
    this._callsMonthNo = now.getMonth() + 1;
    this._callsYearNo = now.getFullYear();
    
    // here the calls will be filled in
    this._callsFound = new Array();
    
    this._callsMissingTo = now.valueOf();
    this._callsMissingFrom = (this._callsLastTime > 0) ? (this._callsLastTime + 1):
                                                         now.valueOf();
    
    this._obs.notifyObservers(this, this._observerTopic,
                              "calls-progress::" + entry._id + "::0");
    
    return "ok";
  },
  
  /*************/
  
  _webCalls: function() {
    //dump("_webCalls\n");
    var method, params;
    var entry = this._accountsArray[this._accessAccount];
    
    method = "GET";
    params = "selection=" + this._callsPageNo +
             "&x=" + Math.floor(Math.random()*70) +
             "&y=" + Math.floor(Math.random()*15);
    
    var urltxt = this._bobCalls;
    if (DEBUG) {
      urltxt = urltxt.replace(/%/, this._callsPageNo);
    }
    var url = Components.classes["@mozilla.org/network/io-service;1"].
                getService(Components.interfaces.nsIIOService).
                  newURI(urltxt, null, null);
    this._webAccess.startRequest(url, method, null, params, null);
    
    return "ok";
  },
  
  /*************/
  
  _webCallsParse: function() {
    //dump("_webCallsParse\n");
    var loadmore = false;
    var entry = this._accountsArray[this._accessAccount];
    
    this._webAccess.patchContents("(<[tT][dD]>(?:(?!</[tT][dD]).)*?)<!--",
                                  "$1!--");
    
    var calls = this._webAccess.findTable("Einzelgespr.*", "Datum");
    
    if (!calls) {
      if (!this._webAccess.contentsText.match(/\bGesamtkosten\b/)) {
        entry._error = "Fehler beim Auslesen der Einzelgespr\u00E4che!";
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesssBob._webCallsParse: "+entry._error,
                                  this._webAccess.contentsHTML, true);
        return "err";
      }
    }
    
    var onecall, fields, matches;
    var time = 0;
    var pnumber;
    var money = 0;
    var duration = 0;
    var ctype = "";
    var parseError = null;
    var callcount = 0;
    var now = (new Date()).valueOf();
    var oldest = now;
    var newest = 0;
    var finished = false;
    
    if (this._callsMaxPageNo < 0) {
      matches = this._webAccess.contentsHTML.match(/name="selection"(.*?)<\/select/i);
      if (matches) {
        matches = matches[1].match(/value="(\d)"/gi);
        if (matches) {
          var val;
          while (matches.length) {
            val = Number(matches.pop().match(/\d/)[0]);
            this._callsMaxPageNo = Math.max(this._callsMaxPageNo, val);
          }
        }
      }
    }
    
    while (calls && calls.hasMoreElements()) {
      onecall = calls.getNext().QueryInterface(Components.interfaces.nsIStringEnumerator);
      fields = new Array();
      while (onecall.hasMore())
        fields.push(onecall.getNext());
      
      if (fields.length != 5) {
        entry._error = "Tabelle der Einzelgespr\u00E4che ge\u00E4ndert!\nBitte den Entwickler informieren!";
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesssBob._webCallsParse: "+entry._error,
                                  this._webAccess.contentsHTML, true);
        
        return "err";
      }
      
      if (!fields[0].match(/^\d/))
        continue;
      
      //dump("fields = "+fields.join(" | ")+"\n");
      
      parseError = null;
      
      matches = (fields[0] + " " + fields[1]).
                  match(/(\d{1,2}).\s*(\d{1,2}).\s*(\d{2,4})\s+(\d{1,2}):(\d{2}):(\d{2})/);
      if (!matches)
        parseError = "date/time";
      else {
        var year = Number(matches[3]);
        if (year < 100)
          year += 2000;
        time = (new Date(year,                        // year
                         Number(matches[2]) - 1,      // month
                         Number(matches[1]),          // day
                         Number(matches[4]),          // hour
                         Number(matches[5]),          // minute
                         Number(matches[6]),          // second
                         0)).valueOf();
         // possibly adjust internal date tracking
         this._callsMonthNo = Number(matches[2]);
         this._callsYearNo = year;
      }
      
      if (time >= this._callsCheckTime && !parseError) {
        var callbag = Components.classes["@mozdev.org/checkyesss/container;1"].
                        createInstance(Components.interfaces.cyICheckyesssContainer);
        
        callbag.setPropertyAsInt("time", time);
        
        // change number to international format
        matches = fields[2].match(/^((?:(?!!--).)*)(.*)/);
        if (matches) {
          pnumber = matches[1];
          pnumber = pnumber.replace(/[\s\/]/g, "");
          if (pnumber.match(/^\+?\d*[\dxX*][xX*]*$/)) {
            pnumber = pnumber.replace(/^00/, "+");            // intl number
            pnumber = pnumber.replace(/^0/, "+43");           // local number
            if (matches[2].match(/\bTELEPHONY\b/))
              callbag.setPropertyAsString("type", "call");
            else if (matches[2].match(/\bBEC\b/))
              callbag.setPropertyAsString("type", "data");
            else
              callbag.setPropertyAsString("type", "sms");
          }
          else {
            pnumber = "";
            matches[2] = matches[1] + "," + matches[2];
            callbag.setPropertyAsString("type", "data");
          }
          callbag.setPropertyAsString("number", pnumber);
          ctype = matches[2].replace(/^!--\s*/, "").replace(/\s*-->$/, "");
          matches = ctype.split(/,\s*/);
          ctype = "";
          var ct, cs;
          for (ct = 0; ct < matches.length; ct++) {
            if (!matches[ct].match(/^\d+$/)) {
              cs = matches[ct];
              // translate known strings into german words
              if (cs == "TELEPHONY")
                cs = "Telefonat";
              else if (cs.match(/MESSAGE|SMS/))   //???
                cs = "SMS";
              else if (cs == "BEC")
                cs = "Datenverbindung";
              else if (cs == "MOBILE_ORIGINATED_CALL")
                cs = "";                      // skip this as it is the default
              else if (cs == "MOBILE_TO_MOBILE")
                cs = "Mobil";
              else if (cs == "NATIONAL")
                cs = "Inland";
              else if (cs == "INTERNATIONAL")
                cs = "Ausland";
              else if (cs == "SPECIAL_NUMBER")
                cs = "Sondernummer";
              if (cs.length) {
                if (ctype.length)
                  ctype += ", ";
                ctype += cs;
              }
            }
          }
          if (ctype.length)
            callbag.setPropertyAsString("ctype", ctype);
        }
        else
          parseError = "phone number";
        
        matches = fields[4].match(/(\d+)[,.](\d+)/);
        if (!matches)
          parseError = "fee";
        else
          money = matches[1]+"."+matches[2];
        
        callbag.setPropertyAsString("fee", money);
        
        duration = 0;
        matches = fields[3].match(/(?:(?:(\d+):)?(\d+):)?(\d+)/);
        if (matches) {
          if (typeof(matches[1]) == "undefined")
            matches[1] = "0";                // the hour may be missing -> 0
          if (typeof(matches[2]) == "undefined")
            matches[2] = "0";                // the minute may be missing -> 0
          duration = Number(matches[1])*3600 +
                     Number(matches[2])*60 +
                     Number(matches[3]);
        }
        if (duration > 0) {
          callbag.setPropertyAsInt("duration", duration);
        }
        
        if (!parseError) {
          this._callsFound.push(callbag);
          callcount++;
        }
      }
      
      if (!parseError) {
        if (oldest > time)
          oldest = time;
        if (newest < time)
          newest = time;
      }
      else {
        entry._error = "Fehler beim Interpretieren der Einzelgespr\u00E4che!";
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesssBob._webCallsParse",
                                  "Error parsing "+parseError+":\n"+fields.toString(),
                                  false);
        return "err";
      }
    }
    
    
    if (callcount > 0) {
      // found at least one call
      
      if (oldest <= this._callsMissingTo)
        this._callsMissingTo = oldest - 1;
      if (newest > this._callsNewLast)
        this._callsNewLast = newest;
      
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "cyCheckyesssBob._webCallsParse",
                                "Loaded page " + this._callsPageNo + " with data from " +
                                (new Date(oldest)).toString() + " to " +
                                (new Date(newest)).toString(), false);
    }
    else {
      // reached the beginning of the month
      oldest = (new Date(this._callsYearNo, this._callsMonthNo - 1, 1, 0, 0, 0, 0)).valueOf();
    }
    
    // proceed with next oldest month
    this._callsPageNo++;
    this._callsMonthNo--;
    if (this._callsMonthNo < 1) {
      this._callsYearNo--;
      this._callsMonthNo = 12;
    }
    
    finished = (this._callsPageNo > this._callsMaxPageNo) ||
               (oldest < this._callsLoadTime);
    
    if (this._callsCheckTime > 0) {
      var progress = 100;
      if (!finished) {
        var whole = now - this._callsLoadTime;
        var diff = now - oldest;
        progress = (whole > 0) ? parseInt((100 * diff) / whole, 10) : 0;
        if (progress < 0)
          progress = 0;
        if (progress > 100)
          progress = 100;
      }
      //dump("progress = "+progress+"%\n");
      this._obs.notifyObservers(this, this._observerTopic,
                                "calls-progress::" + entry._id + "::" + progress);
    }
    
    //dump("loadmore = "+!parseParam.exceed+"\n");
    return (finished) ? "done" : "cont";
  },
  
  /*************/
  
  _webCallsStore: function() {
    //dump("_webCallsStore\n");
    var now = (new Date()).valueOf();
    var entry = this._accountsArray[this._accessAccount];
    
    if (this._callsFound.length > 0) {
      if (this._callsMissingFrom < this._callsMissingTo) {
        this._utils.mergeCallsMissing(entry._id,
                                      this._callsMissingFrom,
                                      this._callsMissingTo);
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                  "cyCheckyesssBob._webCallsStore",
                                  "Missing calls from "+
                                  (new Date(this._callsMissingFrom)).toString() + " to " +
                                  (new Date(this._callsMissingTo)).toString(), false);
      }
      
      if (this._callsCheckTime < this._callsMissingTo)
        this._callsCheckTime = this._callsMissingTo;
      
      /*this._utils.mergeCallsRecords(entry._id,
                                    new ArrayEnumerator(this._callsFound),
                                    "time,duration,number,ctype");*/
      this._utils.storeCallsRecords(entry._id,
                                    new ArrayEnumerator(this._callsFound),
                                    this._callsCheckTime, now, null);
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "cyCheckyesssBob._webCallsStore",
                                "Storing " + this._callsFound.length + " calls from " +
                                (new Date(this._callsCheckTime)).toString() + " to " +
                                (new Date()).toString(), false);
    }
    
    entry._dateCallsLoaded = now;
    var contents = this._utils.getCacheContents(entry._id);
    if (contents) {
      contents.setPropertyAsInt("dateCallsLoaded", entry._dateCallsLoaded);
      this._utils.writeCache(entry._id,              // key
                             contents,
                             this._utils.getCacheValidity(entry._id),
                             false);                 // do not reset age
    }
    
    // we are done -> release some memory
    this._utils.releaseCallsRecords();
    
    return "ok";
  },
  
  /**************************************************************/
  
  _webLogout: function() {
    //dump("_webLogout\n");
    if (!this._webAlreadyLoggedIn) {
      var url = Components.classes["@mozilla.org/network/io-service;1"].
                  getService(Components.interfaces.nsIIOService).
                    newURI(this._bobLogout, null, null);
      this._webAccess.startRequest(url, "GET", null, "", null);
    }
    /*else {
      dump("logout skipped\n");
    }*/
    return "ok";
  },
  
  /**************************************************************/
  
  _webStoreError: function(aErrMsg, aFunctionName, aSerious) {
    aFunctionName = (typeof(aFunctionName) != "undefined") ? aFunctionName : "unknown";
    aSerious = (typeof(aSerious) != "undefined") ? aSerious : true;
  
    var entry = this._accountsArray[this._accessAccount];
    
    // A "serious" error is also written into the account's cache causing the
    // icon to turn gray (or similar). Non-serious error are just recorded in
    // the error log.
    if (aSerious) {
    entry._error = aErrMsg;
    
    var contents = this._utils.getCacheContents(entry._id);
    if (!contents)
      contents = Components.classes["@mozdev.org/checkyesss/container;1"].
                   createInstance(Components.interfaces.cyICheckyesssContainer);
    // convert NL due to moz-bug #169521
    contents.setPropertyAsString("error", entry._error.replace(/\n/g, "\\n"));
    
    this._utils.writeCache(entry._id,              // key
                           contents,
                           4 * this._timeHour,     // expires in 4 hours
                           false);                 // do not reset age
    }
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                              "cyCheckyesssBob._webStoreError",
                              "Account: " + entry._id + "\nCause: " + aErrMsg +
                              "\nFunction: " + aFunctionName + "\nSerious: " + aSerious,
                              false);
    
    this._obs.notifyObservers(this, this._observerTopic,
                              "update::" + entry._id);
  },
  
  /**************************************************************/
  
  _webError: function(aPrefFunctionName) {
    var error = "Unbekannter Fehler";
    var serious = true;
    
    if (this._webAccess.statusText)
      error = this._webAccess.statusText;
    else if (this._webAccess.status > 200)
      error = "Error " + this._webAccess.status;
    
    // Work around a bug (at least that's what it looks right now) in
    // Firefox 3.0b5 (as included in Ubuntu 8.04) where a 302 redirection from
    // a 'https' URL to a 'http' URL results in a security warning while
    // nsIXMLHttpRequest returns a 302 error.
    if (aPrefFunctionName.match(/logout/i))
      serious = false;
    
    this._webStoreError(error, aPrefFunctionName, serious);
    
    return "exit";
  },
  
  /**************************************************************/
  
  _webExit: function() {
    //dump("_webExit\n");
    this._lastAccess = (new Date()).valueOf();
    
    var entry = this._accountsArray[this._accessAccount];
    this._obs.notifyObservers(this, this._observerTopic,
                                    "net-idle::" + entry._id);
    
    if (this._webPathType == "calls") {
      if (entry._error) {
        this._obs.notifyObservers(this, this._observerTopic,
                                  "calls-error::" + entry._id + "::" + entry._error);
      }
      else {
        this._obs.notifyObservers(this, this._observerTopic,
                                  "calls::" + entry._id);
      }
    }
    
    if (this._webPathType == "link") {
      entry._quickLinkPending = false;
      entry._quickLinkWindow = null;
      entry._quickLinkTarget = "";
    }
    
    this._rescheduleCheck();
    return "ok";
  },
  
  /**************************************************************/
  
  _webBrowserGoto: function(aError) {  
    //dump("_webBrowserGoto\n");
    var entry = this._accountsArray[this._accessAccount];
    entry._quickLinkPending = false;
    
    if (entry._quickLinkWindow) {
      if (!aError) {
        // display the remote html page
        entry._quickLinkWindow.location = entry._quickLinkTarget;
      }
      else {
        // replace the announcement by an error message
        var errorText = (entry._error) ? entry._error.replace(/\n/, " ") :
                                         "Unbekannter Fehler";
        var errorPage = "<h2><img src='chrome://checkyesss/skin/logo.png'/>"+
              "&nbsp;&nbsp; " + errorText + "</h2></body></html>";
        var body = entry._quickLinkWindow.document.getElementById("qlbody");
        if (body)
          body.innerHTML = errorPage;
      }
      entry._quickLinkWindow.focus();
      entry._quickLinkWindow = null;
      entry._quickLinkTarget = "";
    }
  },
  
  /**************************************************************/
  /**************************************************************/
  
  abortAccess: function() {
    try {
      this._checkTimer.cancel();
    } catch(e) { }
    
    if (this._webAccess.activeSession) {
      //dump("abortAccess\n");
      this._webAccess.abortRequestSequence(false);
      this._lastAccess = (new Date()).valueOf();
      
      var entry = this._accountsArray[this._accessAccount];
      this._obs.notifyObservers(this, this._observerTopic,
                                      "net-idle::" + entry._id);
    }
  },
  
  /**************************************************************/
  
  ///////////////////////////
  // nsITimerCallback
  notify: function(timer) {
    //dump("notify\n");
    if (timer == this._checkTimer) {
      this._checkAccounts();
    }
  },
  
  /**************************************************************/
  
  observe: function(aSubject, aTopic, aData) {
    //do nothing until the first load
    if (this._firstTime)
      return;
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                              aTopic, aData, false);
    
    switch (aTopic) {
      case "xpcom-shutdown":
        this._stop();
        break;
      
      default:
        break;
    }
  },
  
  /**************************************************************/
  
  _quickLinkUrls: [
    "http://www.bob.at/",
    "https://rechnung.bob.at/ben.set.do?tcd=1"
  ],
  
  jumpQuickLink: function(aId, aIndex, aTab) {
    var win;
    var entry = null;
    var i;
    if (this._accountsArray) {
      for (i = 0; i < this._accountsArray.length; i++) {
        if (this._accountsArray[i]._id == aId) {
          if (this._accountsArray[i]._en)
            entry = this._accountsArray[i];
          break;
        }
      }
    }
    
    if (aIndex < 1) {
      // no need to log in
      win = this._utils.openUrlInBrowser(this._quickLinkUrls[0], aTab);
      if (win)
        win.focus();
    }
    else if (!this._utils.isThunderbird) {
      // we have to log in before showing the page!
      if (entry) {
        var ps;
        win = this._utils.isWebPageOpen(this._blockingUrlRegExp);
        if (win) {
          ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
                     getService(Components.interfaces.nsIPromptService);    
          ps.alert(null, "Achtung",
                   "Die Seite kann nicht aufgerufen werden, da\n"+
                   "bereits ein Bob-Kontomanager angezeigt wird!\n\n"+
                   "Bitte zuerst ausloggen und dann Aktion wiederholen!");
          
          // focus the mentioned window
          //win.focus();
          
          this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                    "cyCheckyesssBob.jumpQuickLink",
                                    "Conflicting browser window detected, action aborted",
                                    false);
        }
        else if (this._anyQuickLinkUpdates() >= 0) {
          ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
                     getService(Components.interfaces.nsIPromptService);
          ps.alert(null, "Achtung",
                   "Die Seite kann nicht aufgerufen werden, da\n" +
                   "gerade ein anderer Quick-Link geladen wird!");
          
          // TODO: we could display the window of the active quick link...
          
          this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                    "cyCheckyesssBob.jumpQuickLink",
                                    "Another quick link request is already active!",
                                    false);
        }
        else {
          var names = this._managerCenter.getQuickLinkNames(this._providerId,
                                                            "auto");
          var name = null;
          i = aIndex;
          while (names && names.hasMore()) {
            name = names.getNext();
            if (i == 0)
              break;
            i--;
          }
          names = null;
          if (name) {
            var jumpPage = "<html><head>"+
              "<title>" + name + " &ouml;ffnen...</title><head>\n"+
              "<body id='qlbody'><h2><img src='chrome://checkyesss/skin/logo.png'/>"+
              "&nbsp;&nbsp;Weiterleitung zu " + name +
              "...</h2></body></html>";
            
            win = this._utils.openUrlInBrowser("", aTab);
            try {
              // does does not work with Firefox 3.0b3
              win.document.open("text/html");
              win.document.write(jumpPage);
              win.document.close();
              win.focus();
            }
            catch (e) { 
              dump(e.toString()+"\n");
            }
            
            entry._quickLinkPending = true;
            entry._quickLinkWindow = win;
            entry._quickLinkTarget = this._quickLinkUrls[aIndex];
            this._checkPending = true;
            
            // restart timer with a short timeout
            this._checkTimer.cancel();
            this._checkTimer.initWithCallback(this, 100, 
                                              Components.interfaces.nsITimer.TYPE_ONE_SHOT);
          }
          else {
            ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
                       getService(Components.interfaces.nsIPromptService);
            ps.alert(null, "Achtung",
                     "Es ist ein interner Fehler aufgetreten!");
            
            this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                      "cyCheckyesssBob.jumpQuickLink",
                                      "Could not find name for quick-link #"+
                                      aIndex+"!", false);
          }
        }
      }
    }
  },
  
  /**************************************************************/
  /**************************************************************/
  
  QueryInterface: function (iid) {
    if (!iid.equals(Components.interfaces.cyICheckyesssManager) &&
        !iid.equals(Components.interfaces.cyICheckyesssWebCallback) &&
        !iid.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
  
};


/******************************************************************************
 * XPCOM Functions for construction and registration
 ******************************************************************************/
var gModule = {
  _firstTime: true,
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {
    if (this._firstTime) {
      this._firstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    }
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, CONTRACT_ID, aFileSpec, aLocation, aType);
  },
  
  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        
  },
  
  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return gFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) { return true; }
};

var gFactory = {
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    return (new CheckyesssBobManager()).QueryInterface(aIID);
  }
};

function NSGetModule(aCompMgr, aFileSpec) { return gModule; }
