/* ***** BEGIN LICENSE BLOCK *****
   - Version: MPL 1.1/GPL 2.0/LGPL 2.1
   -
   - The contents of this file are subject to the Mozilla Public License Version
   - 1.1 (the "License"); you may not use this file except in compliance with
   - the License. You may obtain a copy of the License at
   - http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS IS" basis,
   - WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
   - for the specific language governing rights and limitations under the
   - License.
   -
   - The Original Code is CheckYesss.
   -
   - The Initial Developer of the Original Code is
   - Thomas Reitmayr <treitmayr@yahoo.com>.
   - Portions created by the Initial Developer are Copyright (C) 2005-2008
   - the Initial Developer. All Rights Reserved.
   -
   - Contributor(s):
   - -
   -
   - Alternatively, the contents of this file may be used under the terms of
   - either the GNU General Public License Version 2 or later (the "GPL"), or
   - the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
   - in which case the provisions of the GPL or the LGPL are applicable instead
   - of those above. If you wish to allow use of your version of this file only
   - under the terms of either the GPL or the LGPL, and not to allow others to
   - use your version of this file under the terms of the MPL, indicate your
   - decision by deleting the provisions above and replace them with the notice
   - and other provisions required by the LGPL or the GPL. If you do not delete
   - the provisions above, a recipient may use your version of this file under
   - the terms of any one of the MPL, the GPL or the LGPL.
   -
   - ***** END LICENSE BLOCK ***** */

const CLASS_ID = Components.ID("{eab76df0-ca5d-11da-a94d-0800200c9a66}");
const CLASS_NAME = "Checkyesss Manager Center";
const CONTRACT_ID = "@mozdev.org/checkyesss/manager_center;1";

/***************************************************************************/

/**
/* ArrayEnumerator courtesy of Brendan Eich <brendan@mozilla.org>
 */
function ArrayEnumerator(array) {
  this.array = array;
  this.index = 0;
}
ArrayEnumerator.prototype = {
  hasMoreElements: function () { return this.index < this.array.length; },
  getNext: function () { return (this.index < this.array.length) ?
                       this.array[this.index++] : null; }
};
// nsISupportsArrayEnumerator
//
function nsISupportsArrayEnumerator(array) {
  this.array = array;
  this.index = 0;
}
nsISupportsArrayEnumerator.prototype = {
  hasMoreElements: function() { return this.index < this.array.Count(); },
  getNext: function () { return (this.index < this.array.Count()) ?
                       this.array.GetElementAt(this.index++) : null; }
};

/**
/* StringEnumerator
 */
function StringEnumerator(array) {
  this.array = array;
  this.index = 0;
}
StringEnumerator.prototype = {
  hasMore: function() { return this.index < this.array.length; },
  getNext: function () { return (this.index < this.array.length) ?
                       this.array[this.index++] : null; }
};

/******************************************************************************
 * CheckyesssManagerCenter component
 ******************************************************************************/

function CheckyesssManagerCenter() {
  this.wrappedJSObject = this;
  this._start();
}


CheckyesssManagerCenter.prototype = {
  
  _providerList: [
    "YS001",
    "SG001",
    "SG002",
    "VS001",
    "SY001",
    "BB001"
  ],
  
  _managerContractId: [
    "yesss_manager;1",
    "sipgate_manager;1",
    "sipgate_manager;2",
    "voipstation_manager;1",
    "simyo_manager;1",
    "bob_manager;1"
  ],
  
  _providerPropertyList: [
    "name",              // name of the provider
    "validUserRegExp",   // regexp for a valid login name
    "country"],          // country
  
  _providerProperties: [  // TODO: split this up for different rate plans, too!
    [ "YESSS!", "^(?:0(?:699|664|676|650|660|688|681|680)[ /]?[0-9]{6,8}|43780[0-9]{6})$", "Austria" ],
    [ "sipgate.at", "^\\S+$", "Austria" ],
    [ "sipgate.de", "^\\S+$", "Germany" ],
    [ "PepPhone-Free", "^.+$", "Germany" ],
    [ "simyo", "^\\s*\\d[\\d\\s]*(?:/\\s*\\d[\\d\\s]*)?$", "Germany" ],
    [ "bob", "^(?:\\+43|0)6\\S+", "Austria" ]
  ],
  
  _ratePlanList: [
    [ "auto", "wertkarte", "V", "webfon", "internet" ],
    [ "auto" ],
    [ "auto" ],
    [ "auto" ],
    [ "auto" ],
    [ "auto" ]
  ],
  
  _ratePlanPropertyList: [
    "name",                 // name of the rate plan
    "description",          // description
    "autoResetPeriod",      // can the billing period be determined automatically
    "userResetPeriod" ],    // may the user select when to reset the billing period
  
  _ratePlanProperties: [
    [ [ "automatisch", "Automatische Tariferkennung", 0, 0 ],
      [ "Wertkarte", "Wertkartentarif 6,9/8/9\u00A2/min", 0, 0 ],
      [ "+V!", "Vertragstarif", 0, 1 ],
      [ "WebFon", "WebFon VoIP", 0, 1 ],
      [ "Diskont-Surfen", "Diskont-Surfen", 0, 0 ] ],
    [ [ "automatisch", "", 0, 0 ] ],
    [ [ "automatisch", "", 0, 0 ] ],
    [ [ "automatisch", "", 0, 0 ] ],
    [ [ "automatisch", "", 0, 0 ] ],
    [ [ "automatisch", "", 0, 0 ] ]
  ],
  
  _quickLinkNames: [
    [ [ "Yesss! Homepage" ],
      [ "Yesss! Homepage", "Yesss! Kontomanager", "Yesss! WebSMS", "Yesss! Aufladen" ],
      [ "Yesss! Homepage", "Yesss!+V! Kontomanager", "Yesss!+V! WebSMS" ],
      [ "Yesss! Homepage", "Yesss! WebFon Kontomanager" ],
      [ "Yesss! Homepage", "Yesss! Diskont-Surfen Kontomanager" ] ],
    [ [ "sipgate Homepage", "sipgate Start", "sipgate Voicemail", "sipgate SMS" ] ],
    [ [ "sipgate Homepage", "sipgate Start", "sipgate Voicemail", "sipgate SMS" ] ],
    [ [ "PepPhone-Free Homepage", "PepPhone-Free Startseite", "PepPhone-Free 'Mein Account'" ] ],
    [ [ "simyo Homepage", "simyo Kontomanager" ] ],
    [ [ "Bob Homepage", "Bob Costmanager" ] ]
  ],
  
  _quickLinkPreferred: [
    [ "0", "2", "2", "1", "1" ],        // YESSS
    [ "3" ],                            // sipgate.at
    [ "3" ],                            // sipgate.de
    [ "1" ],                            // VoipStation
    [ "1" ],                            // simyo
    [ "1" ]                             // Bob
  ],
  
  _features: [  // add "datavol" here, but needs to be split for each rate plan
    [ "credit", "creditLog", "tooltip", "callDetails" ],
    [ "credit", "creditLog", "tooltip", "callDetails" ],
    [ "credit", "creditLog", "tooltip", "callDetails" ],
    [ "credit", "creditLog", "tooltip", "callDetails" ],
    [ "credit", "creditLog", "tooltip", "callDetails" ],
    [ "credit", "creditLog", "tooltip" ]  // Bob Einzelgespraeche nicht implementiert
  ],
  
/* TODO
  _customConfigList: [ "name", "label", "description", "type", "regexp" ],
  
  _customConfig: [
    null,
    [ [ "checkMissed", " */
  
  _headerPropertyList: [ "shortcut", "cleartext", "format",
                         "shortView", "extendedView", "fileView" ],
  
  _headerProperties: [
    /***** YESSS *****/
    [ // auto:
      [ [ "ctype", "Typ", "string", 1, 1, 1 ] ],
      // wertkarte:
      [ [ "ctype", "Typ", "string", 1, 1, 1 ],
        [ "tcharged", "Getaktet", "duration", 0, 1, 1 ] ],
      // V:
      [ [ "ctype", "Typ", "string", 1, 1, 1 ] ],
      // webfon:
      [ [ "ctype", "Typ", "string", 1, 1, 1 ],
        [ "credit", "x", "fixed", 0, 0, 0 ] ],
      // internet:
      [ [ "sessionid", "Session-ID", "string", 0, 1, 1 ],
        [ "upload", "Upload (MB)", "fixed", 1, 1, 1 ],
        [ "download", "Download (MB)", "fixed", 1, 1, 1 ],
        [ "ctype", "Tarif", "string", 0, 1, 1 ],
        [ "fee", "Entgelt", "fixed", 1, 1, 1 ],
        [ "number", "x", "", 0, 0, 0 ],
        [ "credit", "x", "", 0, 0, 0 ] ] ],
    /***** SIPGATE.AT *****/
    [ [ [ "stype", "Typ", "string", 1, 0, 0 ],
        [ "ctype", "Typ", "string", 0, 1, 1 ],
        [ "tcharged", "Getaktet", "duration", 0, 1, 1 ] ] ],
    /***** SIPGATE.DE *****/
    [ [ [ "stype", "Typ", "string", 1, 0, 0 ],
        [ "ctype", "Typ", "string", 0, 1, 1 ],
        [ "tcharged", "Getaktet", "duration", 0, 1, 1 ] ] ],
    /***** VOIPSTATION *****/
    [ [ [ "ctype", "Status", "string", 0, 1, 1 ],
        [ "credit", "x", "fixed", 0, 0, 0 ] ] ],
    /***** simyo (not used yet) *****/
    [ [ [ "ctype", "Status", "string", 0, 1, 1 ],
        [ "credit", "x", "fixed", 0, 0, 0 ] ] ],
    /***** Bob (not used yet) *****/
    [ [ [ "ctype", "Status", "string", 0, 1, 1 ],
        [ "credit", "x", "fixed", 0, 0, 0 ] ] ]
  ],
  
  _usedFeeDecimals: [
    [ "4", "4", "4", "4", "2" ],        // YESSS
    [ "4" ],                            // sipgate.at
    [ "4" ],                            // sipgate.de
    [ "4" ],                            // VoipStation
    [ "4" ],                            // simyo
    [ "4" ]                             // Bob
  ],
  
  _transferUrls: [
    [ "http://www.yesss.at",            // YESSS
      "https://www.yesss.at",
      "http://yesss.at",
      "https://yesss.at",
      "http://www.ewave.at"],
    [ "http://www.sipgate.at",          // sipgate.at
      "https://secure.sipgate.at",
      "http://sipgate.at" ],
    [ "http://www.sipgate.de",          // sipgate.de
      "https://secure.sipgate.de",
      "http://sipgate.de" ],
    [ "http://www.pepphone-free.de",    // VoipStation
      "http://pepphone-free.de" ],
    [ "http://www.simyo.de",            // simyo
      "https://www.simyo.de" ],
    [ "https://rechnung.bob.at" ]       // Bob
  ],
  
  
  _utils: null,
  _providerIndex: null,
  _providerManagers: null,
  
  
  /**************************************************************/
  
  _start: function() {
    // set components
    this._utils = Components.classes["@mozdev.org/checkyesss/util_service;1"].
                getService(Components.interfaces.cyICheckyesssUtil);
    
    // create an index for the provider IDs
    this._providerIndex = new Object();
    this._providerManagers = new Object();
    var id;
    for (var i = 0; i < this._providerList.length; i++)  {
      id = this._providerList[i];
      this._providerIndex[id] = i;
      this._providerManagers[id] = null;
    }
  },
  
  /**************************************************************/
  
  _stop: function() {
    this._providerManagers = null;
    this._providerIndex = null;
    this._utils = null;
  },
  
  /**************************************************************/
  
  get getProviderList() {
    return new StringEnumerator(this._providerList);
  },
  
  /**************************************************************/
  
  isValidProvider: function(aProviderId) {
    return (aProviderId in this._providerIndex);
  },
  
  /**************************************************************/
  
  getProviderProperties: function(aProviderId) {
    if (!(aProviderId in this._providerIndex)) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssCenter.getProviderProperties",
                                "Invalid provider ID '" + aProviderId + "'",
                                false);
      return null;
    }
    
    var index = this._providerIndex[aProviderId];
    
    var container = Components.classes["@mozdev.org/checkyesss/container;1"].
                      createInstance(Components.interfaces.cyICheckyesssContainer);
    for (var i = 0; i < this._providerPropertyList.length; i++)
      container.setPropertyAsString(this._providerPropertyList[i],
                                    this._providerProperties[index][i]);
    
    container.setPropertyAsString("topic", "checkyesss:"+aProviderId);
    
    return container;
  },
  
  /**************************************************************/
  
  getRatePlanList: function(aProviderId) {
    if (!(aProviderId in this._providerIndex)) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssCenter.getRatePlanList",
                                "Invalid provider ID '" + aProviderId + "'",
                                false);
      return null;
    }
    
    var index = this._providerIndex[aProviderId];
    return new StringEnumerator(this._ratePlanList[index]);
  },
  
  /**************************************************************/
  
  getRatePlanProperties: function(aProviderId, aRatePlanId) {
    if (!(aProviderId in this._providerIndex)) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssCenter.getRatePlanProperties",
                                "Invalid provider ID '" + aProviderId + "'",
                                false);
      return null;
    }
    
    var index = this._providerIndex[aProviderId];
    var rpindex = -1;
    var properties = this._ratePlanProperties[index];
    var i;
    
    for (i = 0; i < this._ratePlanList[index].length; i++) {
      if (this._ratePlanList[index][i] == aRatePlanId) {
        rpindex = i;
        break;
      }
    }
    
    if (rpindex < 0) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssCenter.getRatePlanProperties",
                                "Invalid rate plan ID '" + aRatePlanId + "'",
                                false);
      return null;
    }
    
    properties = properties[rpindex];
    
    var container = Components.classes["@mozdev.org/checkyesss/container;1"].
                      createInstance(Components.interfaces.cyICheckyesssContainer);
    for (i = 0; i < this._ratePlanPropertyList.length; i++)
      container.setPropertyAsString(this._ratePlanPropertyList[i], properties[i]);
    
    return container;
  },
  
  /**************************************************************/
  
  getQuickLinkNames: function(aProviderId, aRatePlanId) {
    if (!(aProviderId in this._providerIndex)) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssCenter.getQuickLinkNames",
                                "Invalid provider ID '" + aProviderId + "'",
                                false);
      return null;
    }
    
    var index = this._providerIndex[aProviderId];
    var rpindex = -1;
    var properties = this._quickLinkNames[index];
    var i;
    
    for (i = 0; i < this._ratePlanList[index].length; i++) {
      if (this._ratePlanList[index][i] == aRatePlanId) {
        rpindex = i;
        break;
      }
    }
    
    if (rpindex < 0) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssCenter.getRatePlanProperties",
                                "Invalid rate plan ID '" + aRatePlanId + "'",
                                false);
      return null;
    }
    
    return new StringEnumerator(properties[rpindex]);
  },
  
  /**************************************************************/
  
  hasFeature: function(aProviderId, aFeature) {
    if (!(aProviderId in this._providerIndex)) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssCenter.hasFeature",
                                "Invalid provider ID '" + aProviderId + "'",
                                false);
      return false;
    }
    
    var index = this._providerIndex[aProviderId];
    var myFeatures = this._features[index];
    for (var i = 0; i < myFeatures.length; i++) {
      if (myFeatures[i] == aFeature)
        return true;
    }
    return false;
  },
  
  /**************************************************************/
  
  getUsedFeeDecimals: function(aProviderId, aRatePlanId) {
    if (!(aProviderId in this._providerIndex)) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssCenter.getUsedFeeDecimals",
                                "Invalid provider ID '" + aProviderId + "'",
                                false);
      return null;
    }
    
    var index = this._providerIndex[aProviderId];
    var properties = this._usedFeeDecimals[index];
    var rpindex = -1;
    var i, j;
    
    for (i = 0; i < this._ratePlanList[index].length; i++) {
      if (this._ratePlanList[index][i] == aRatePlanId) {
        rpindex = i;
        break;
      }
    }
    
    if (rpindex < 0) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssCenter.getUsedFeeDecimals",
                                "Invalid rate plan ID '" + aRatePlanId + "'",
                                false);
      return null;
    }
    
    return properties[rpindex];
  },
  
  /**************************************************************/
  
  getCallsHeaderProperties: function(aProviderId, aRatePlanId) {
    if (!(aProviderId in this._providerIndex)) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssCenter.getCallsHeaderProperties",
                                "Invalid provider ID '" + aProviderId + "'",
                                false);
      return null;
    }
    
    var index = this._providerIndex[aProviderId];
    var properties = this._headerProperties[index];
    var rpindex = -1;
    var i, j;
    
    for (i = 0; i < this._ratePlanList[index].length; i++) {
      if (this._ratePlanList[index][i] == aRatePlanId) {
        rpindex = i;
        break;
      }
    }
    
    if (rpindex < 0) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssCenter.getRatePlanProperties",
                                "Invalid rate plan ID '" + aRatePlanId + "'",
                                false);
      return null;
    }
    
    properties = properties[rpindex];
    var prop2;
    var result = new Array();
    var container;
    
    for (i = 0; i < properties.length; i++) {
      prop2 = properties[i];
      container = Components.classes["@mozdev.org/checkyesss/container;1"].
                    createInstance(Components.interfaces.cyICheckyesssContainer);
      for (j = 0; j < this._headerPropertyList.length; j++)
        container.setPropertyAsString(this._headerPropertyList[j], prop2[j]);
      result.push(container);
    }
    
    return new ArrayEnumerator(result);
  },
  
  /**************************************************************/
  
  storePassword: function(aProviderId, aAccountId, aUserName, aPassword) {
    if (typeof(aPassword) == "string" && aPassword.length > 0) {
      if ("@mozilla.org/passwordmanager;1" in Components.classes) {
        // "old" password manager
        Components.classes["@mozilla.org/passwordmanager;1"].
          createInstance(Components.interfaces.nsIPasswordManager).
            addUser("Checkyesss " + aProviderId + " #" + aAccountId,
                    aUserName, aPassword);
      }
      else
      if ("@mozilla.org/login-manager;1" in Components.classes) {
        // login manager (Firefox 3)
        var loginManager = Components.classes["@mozilla.org/login-manager;1"].
                                getService(Components.interfaces.nsILoginManager);
        var realm = aProviderId + " #" + aAccountId;
        // first try to remove any existing entry
        var logins = loginManager.findLogins({}, "chrome://checkyesss", null, realm);
        for (var i = 0; i < logins.length; i++) {
          if (logins[i].username == aUserName) {
            loginManager.removeLogin(logins[i]);
            break;
          }
        }
        var nsLoginInfo = new Components.Constructor("@mozilla.org/login-manager/loginInfo;1",
                                Components.interfaces.nsILoginInfo, "init");
        var loginInfo = new nsLoginInfo("chrome://checkyesss",
                                null, realm,
                                aUserName, aPassword, "", "");
        loginManager.addLogin(loginInfo);
      }
    }
    else {
      this.removePassword(aProviderId, aAccountId, aUserName);
    }
  },
  
  /**************************************************************/
  
  getPassword: function(aProviderId, aAccountId, aUserName,
                        aAlwaysAskUser, aAskIfMissing) {
    if (!(aProviderId in this._providerIndex)) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssCenter.getPassword",
                                "Invalid provider ID '" + aProviderId + "'",
                                false);
      return null;
    }
    
    var title = "Passwort f\u00FCr Konto " + aAccountId;
    var msg1 = "Bitte das Passwort zum Benutzernamen\n'" + aUserName + "' eingeben";
    var msg2 = "Das Passwort permanent speichern?";
    var password = {value:""};
    var checked = {value:true};
    var ok, ps;
    
    if (aAlwaysAskUser) {
      // here we do not touch the password/login manager but just ask the user
      // for the password
      ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
                 getService(Components.interfaces.nsIPromptService);    
      ok = ps.promptPassword(null, title, msg1, password, null, checked);
      if (ok && password.value.length > 0)
        return password.value;
      return null;
    }
    else {
      // try to get the password from the password manager
      // if not found ask the user
      var index = this._providerIndex[aProviderId];
      var url = "Checkyesss " + aProviderId + " #" + aAccountId;
      var host = {value:""};
      var user =  {value:""};
      var found = false;
      var i;
      
      if ("@mozilla.org/login-manager;1" in Components.classes) {
        // login manager (Firefox 3)
        var loginManager = Components.classes["@mozilla.org/login-manager;1"].
                                getService(Components.interfaces.nsILoginManager);
        
        // first try to find old-style entries
        var logins = loginManager.findLogins({}, url, "", "");
        if (logins.length > 0) {
          for (i = 0; i < logins.length; i++) {
            if (logins[i].username == aUserName) {
              if (!found) {
                found = (logins[i].username.length > 0) && (logins[i].password.length > 0);
                // remove old-style entry
                loginManager.removeLogin(logins[i]);
                if (found) {
                  // re-add the entry in the new style
                  password.value = logins[i].password;
                  this.storePassword(aProviderId, aAccountId,
                                     logins[i].username, logins[i].password);
                }
              }
            }
          }
        }
        
        if (!found) {
          // try to find new-style entries
          logins = loginManager.findLogins({}, "chrome://checkyesss", "",
                                           aProviderId + " #" + aAccountId);
          if (logins.length > 0) {
            for (i = 0; i < logins.length; i++) {
              if ((logins[i].username == aUserName) &&
                  (logins[i].password.length > 0)) {
                found = true;
                password.value = logins[i].password;
                break;
              }
            }
          }
        }
      }
      
      if (!found && ("@mozilla.org/passwordmanager;1" in Components.classes)) {
        // try "old" password manager
        found = true;
        try {
          Components.classes["@mozilla.org/passwordmanager;1"].
            createInstance(Components.interfaces.nsIPasswordManagerInternal).
              findPasswordEntry(url, aUserName, null, host, user, password);
        }
        catch(e) { found = false; }
        found = found && (user.value.length > 0) && (password.value.length > 0);
        //dump("found="+found+", url="+url+", aUserName="+aUserName+", user="+user.value+", passwd="+password.value+"\n");
      }
      
      if (!found && aAskIfMissing) {
        password.value = "";
        ps = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
                   getService(Components.interfaces.nsIPromptService);    
        ok = ps.promptPassword(null, title, msg1, password, msg2, checked);
        found = ok && (password.value.length > 0);
        if (found && checked.value)
          this.storePassword(aProviderId, aAccountId, aUserName, password.value);
      }
      
      return (found) ? password.value : null;
    }
  },
  
  /**************************************************************/
  
  removePassword: function(aProviderId, aAccountId, aUserName) {
    if (!(aProviderId in this._providerIndex)) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssCenter.removePassword",
                                "Invalid provider ID '" + aProviderId + "'",
                                false);
      return;
    }
    
    var index = this._providerIndex[aProviderId];
    var url = "Checkyesss " + aProviderId + " #" + aAccountId;
    var host = {value:""};
    var user =  {value:""};
    var password = {value:""};
    var found = false;
    var i;
    
    if ("@mozilla.org/login-manager;1" in Components.classes) {
      // login manager (Firefox 3)
      var loginManager = Components.classes["@mozilla.org/login-manager;1"].
                              getService(Components.interfaces.nsILoginManager);
      
      // first try to find old-style entries
      var logins = loginManager.findLogins({}, url, "", "");
      if (logins.length > 0) {
        for (i = 0; i < logins.length; i++) {
          if (logins[i].username == aUserName) {
            loginManager.removeLogin(logins[i]);
          }
        }
      }
      // next try to find new-style entries
      logins = loginManager.findLogins({}, "chrome://checkyesss", "",
                                       aProviderId + " #" + aAccountId);
      if (logins.length > 0) {
        for (i = 0; i < logins.length; i++) {
          if (logins[i].username == aUserName) {
            loginManager.removeLogin(logins[i]);
          }
        }
      }
    }
    else
    if ("@mozilla.org/passwordmanager;1" in Components.classes) {
      var pwm = Components.classes["@mozilla.org/passwordmanager;1"].
                   createInstance(Components.interfaces.nsIPasswordManager);
    
      found = true;
      try {
        pwm.QueryInterface(Components.interfaces.nsIPasswordManagerInternal).
            findPasswordEntry(url, aUserName, null, host, user, password);
      }
      catch(e) { found = false; }
      found = found && (user.value.length > 0);
      
      if (found)
        pwm.removeUser(host.value, user.value);
    }
    return;
  },
  
  /**************************************************************/
  
  transferAccounts: function(aProviderId, aMaxAccounts) {
    if (!(aProviderId in this._providerIndex)) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssCenter.transferAccounts",
                                "Invalid provider ID '" + aProviderId + "'",
                                false);
      return null;
    }
    
    var urls = this._transferUrls[this._providerIndex[aProviderId]];
    var i, j, logins;
    var foundNames = new Array();
    var pwm;
    var host = {value:""};
    var user =  {value:""};
    var password = {value:""}; 
    var found;
    var id;

    if ("@mozilla.org/login-manager;1" in Components.classes)
      pwm = Components.classes["@mozilla.org/login-manager;1"].
              getService(Components.interfaces.nsILoginManager);
    else
    if ("@mozilla.org/passwordmanager;1" in Components.classes)
      pwm = Components.classes["@mozilla.org/passwordmanager;1"].
              createInstance(Components.interfaces.nsIPasswordManagerInternal);
    else
      return null;
    
    for (i = 0; i < urls.length; i++) {
      if (aMaxAccounts <= 0)
        break;
      if ("@mozilla.org/login-manager;1" in Components.classes) {
        // login manager (Firefox 3)
        logins = pwm.findLogins({}, urls[i], "", "");
        if (logins.length > 0) {
          for (j = 0; j < logins.length; j++) {
            if ((aMaxAccounts > 0) &&
                (logins[j].username.length > 0) &&
                (foundNames.indexOf(logins[j].username) < 0)) {
              foundNames.push(logins[j].username);
              id = "Tel-" + aProviderId + foundNames.length;
              aMaxAccounts--;
              this.storePassword(aProviderId, id,
                                 logins[j].username,
                                 logins[j].password);
            }
          }
        }
      }
      else
      if ("@mozilla.org/passwordmanager;1" in Components.classes) {
        host.value = "";
        user.value = "";
        password.value = "";
        found = true;
        try {
          pwm.findPasswordEntry(url[i], null, null, host, user, password);
        }
        catch(e) { found = false; }
        found = found && (user.value.length > 0);
        if (found && (foundNames.indexOf(user.value) < 0)) {
          foundNames.push(user.value);
          id = "Tel-" + aProviderId + foundNames.length;
          aMaxAccounts--;
          this.storePassword(aProviderId, id,
                             user.value,
                             password.value);
        }
      }
    }
    return new StringEnumerator(foundNames);
  },

  /**************************************************************/
  
  getManager: function(aProviderId) {
    if (!(aProviderId in this._providerIndex)) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssCenter.getManager",
                                "Invalid provider ID '" + aProviderId + "'",
                                false);
      return null;
    }
    
    var manager = this._providerManagers[aProviderId];
    
    if (!manager) {
      var index = this._providerIndex[aProviderId];
      var contractId = "@mozdev.org/checkyesss/"+this._managerContractId[index];
      try {
        manager = Components.classes[contractId].
                             getService(Components.interfaces.cyICheckyesssManager);
        this._providerManagers[aProviderId] = manager;
      }
      catch (e) {
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesssCenter.getManager",
                                  "Error starting a manager component " +
                                  aProviderId + ": " + e.toString(),
                                  false);
      }
    }
    return manager;
  },
  
  /**************************************************************/
  
  /**
   * clear the internal reference to the manager so it can unload;
   * call this when the manager is not needed in the near future
   */
  releaseManager: function(aProviderId) {
    if (!(aProviderId in this._providerIndex)) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssCenter.releaseManager",
                                "Invalid provider ID '" + aProviderId + "'",
                                false);
      return;
    }
    
    this._providerManagers[aProviderId] = null;
  },
  
  /**************************************************************/
  /**************************************************************/
  
  QueryInterface: function (iid) {
    if (!iid.equals(Components.interfaces.cyICheckyesssCenter) &&
        !iid.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
  
};


/******************************************************************************
 * XPCOM Functions for construction and registration
 ******************************************************************************/
var gModule = {
  _firstTime: true,
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {
    if (this._firstTime) {
      this._firstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    }
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, CONTRACT_ID, aFileSpec, aLocation, aType);
  },
  
  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        
  },
  
  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return gFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) { return true; }
};

var gFactory = {
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    return (new CheckyesssManagerCenter()).QueryInterface(aIID);
  }
};

function NSGetModule(aCompMgr, aFileSpec) { return gModule; }
