/* ***** BEGIN LICENSE BLOCK *****
   - Version: MPL 1.1/GPL 2.0/LGPL 2.1
   -
   - The contents of this file are subject to the Mozilla Public License Version
   - 1.1 (the "License"); you may not use this file except in compliance with
   - the License. You may obtain a copy of the License at
   - http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS IS" basis,
   - WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
   - for the specific language governing rights and limitations under the
   - License.
   -
   - The Original Code is CheckYesss.
   -
   - The Initial Developer of the Original Code is
   - Thomas Reitmayr <treitmayr@yahoo.com>.
   - Portions created by the Initial Developer are Copyright (C) 2005-2008
   - the Initial Developer. All Rights Reserved.
   -
   - Contributor(s):
   - -
   -
   - Alternatively, the contents of this file may be used under the terms of
   - either the GNU General Public License Version 2 or later (the "GPL"), or
   - the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
   - in which case the provisions of the GPL or the LGPL are applicable instead
   - of those above. If you wish to allow use of your version of this file only
   - under the terms of either the GPL or the LGPL, and not to allow others to
   - use your version of this file under the terms of the MPL, indicate your
   - decision by deleting the provisions above and replace them with the notice
   - and other provisions required by the LGPL or the GPL. If you do not delete
   - the provisions above, a recipient may use your version of this file under
   - the terms of any one of the MPL, the GPL or the LGPL.
   -
   - ***** END LICENSE BLOCK ***** */

const CLASS_ID = Components.ID("{3da2b5f0-22cb-11da-8cd6-0800200c9a66}");
const CLASS_NAME = "CheckYesss Utilities JS Component";
const CONTRACT_ID = "@mozdev.org/checkyesss/util_service;1";

const DEBUG = 0;



/***************************************************************************/

/**
/* ArrayEnumerator courtesy of Brendan Eich <brendan@mozilla.org>
 */
function ArrayEnumerator(array) {
    this.array = array;
    this.index = 0;
}
ArrayEnumerator.prototype = {
    hasMoreElements: function() { return this.index < this.array.length; },
    getNext: function () { return (this.index < this.array.length) ?
                       this.array[this.index++] : null; }
};

/***************************************************************************/

/**
 * The CheckYesss Util Service.
 */
function CheckyesssUtilServiceClass() {
  // allow unwrapping JS component - not needed here
  //this.wrappedJSObject = this;
  null;
}


CheckyesssUtilServiceClass.prototype = {

  /**********************************************************************/
  /*               P R E F E R E N C E   H A N D L I N G                */
  /**********************************************************************/

  /*****************************************************************
   * Read a boolean preference.
   * 
   * @param  {nsIPrefService} aBranch   the desired preferences branch
   * @param  {string}         aPrefId   the name of the preference
   * @param  {boolean}        aDefault  the default value if the preference does not exist
   * @return {boolean}        the value of the preference
   */
  readBoolPref: function(aBranch, aPrefId, aDefault) {
    var value;
    try {
      value = aBranch.getBoolPref(aPrefId);
    }
    catch(e) {
      aBranch.setBoolPref(aPrefId, aDefault);
      value = aDefault;
    }
    return value;
  },
  
  
  /*****************************************************************
   * Read an integer preference
   */
  readIntPref: function(aBranch, aPrefId, aMin, aMax, aDefault) {
    var value, valueOrg;
    var found = true;
    try {
      valueOrg = aBranch.getIntPref(aPrefId);
    } catch(e) { found = false; }
    //dump(aPrefId+"="+valueOrg+"\n");
    value = (!found) ? aDefault :
            (valueOrg < aMin) ? aMin :
            (valueOrg > aMax) ? aMax : valueOrg;
    if (!found || (value != valueOrg))
      aBranch.setIntPref(aPrefId, value);
    return value;
  },
  
  
  /*****************************************************************
   * Read a character preference
   */
  readCharPref: function(aBranch, aPrefId, aDefault) {
    var value;
    try {
      value = aBranch.getCharPref(aPrefId);
    }
    catch(e) {
      aBranch.setCharPref(aPrefId, aDefault);
      value = aDefault;
    }
    return value;
  },


  /*****************************************************************
   * Update a boolean preference, if changed
   */
  updateBoolPref: function(aBranch, aPrefId, aValue) {
    var valueOrg;
    try {
      valueOrg = aBranch.getBoolPref(aPrefId);
    }
    catch(e) {
      valueOrg = !aValue;
    }
    if (aValue != valueOrg)
      aBranch.setBoolPref(aPrefId, aValue);
  },
  
  
  /*****************************************************************
   * Update an integer preference, if changed
   */
  updateIntPref: function(aBranch, aPrefId, aValue) {
    var valueOrg;
    try {
      valueOrg = aBranch.getIntPref(aPrefId);
    }
    catch(e) {
      valueOrg = aValue + 1;
    }
    if (aValue != valueOrg)
      aBranch.setIntPref(aPrefId, aValue);
  },
  
  
  /*****************************************************************
   * Update a character preference, if changed
   */
  updateCharPref: function(aBranch, aPrefId, aValue) {
    var valueOrg;
    try {
      valueOrg = aBranch.getCharPref(aPrefId);
    }
    catch(e) {
      valueOrg = aValue + "x";
    }
    if (aValue != valueOrg)
      aBranch.setCharPref(aPrefId, aValue);
  },
  
  
  /**********************************************************************/
  /*                    T E X T   C O N V E R S I O N                   */
  /**********************************************************************/
  
  /*****************************************************************
   * Converts HTML to text.
   */
  
  html2txt: function(aString) {
    var txt = aString;
    if (txt.match(/&/)) {
      var h2tConvTable = {
        "auml":  "\u00E4",
        "ouml":  "\u00F6",
        "uuml":  "\u00FC",
        "Auml":  "\u00C4",
        "Ouml":  "\u00D6",
        "Uuml":  "\u00DC",
        "szlig": "\u00DF",
        "euro":  "\u20AC",
        "cent":  "\u00A2",
        "copy":  "\u00A9",
        "reg":   "\u00AE",
        "quot":  "\"",
        "amp":   "&",
        "nbsp":  " ",
        "gt":    ">",
        "lt":    "<" };
      var rfunc = function(aEntity) {
        if (aEntity.charAt(1) != "#") {
          // handle &<name>;
          var core = aEntity.substr(1, aEntity.length - 2);
          if (core in h2tConvTable)
            return h2tConvTable[core];
          else
            return aEntity;
        }
        else {
          // handle &#<ddd>; and &#x<hhh>;
          var num = (aEntity.charAt(2).toLowerCase() == "x") ?
                     parseInt(aEntity.substr(3, aEntity.length - 4), 16) :
                     parseInt(aEntity.substr(2, aEntity.length - 3), 10);
          return String.fromCharCode(num);
        }
      };
      txt = txt.replace(/&(?:[a-z]+|#\d+|#x[0-9a-f]+);/gi, rfunc);
    }
    return txt;
  },
  
  
  /*****************************************************************
   * Converts text to HTML.
   */
  txt2html: function(aString) {
    var rfunc = function(aChar) {
      switch (aChar) {
        case "<": return "&lt;";
        case ">": return "&gt;";
        case "&": return "&amp;";
        case "\n": return "<br>";
        default:  return "&#"+aChar.charCodeAt(0)+";";
      }
    };
    return aString.replace(/[<>&\u0000-\u001F\u007F-\uFFFF]/g, rfunc);
  },
  
  
  /*****************************************************************
   * Return the size of the file in a user-readable text format.
   */
  convertFileSize: function(aFileSize) {
    var fsize = aFileSize;
    var unit = " B";
    if (fsize >= 1024) {
      fsize /= 1024;
      unit = " KB";
    }
    if (fsize >= 1024) {
      fsize /= 1024;
      unit = " MB";
    }
    if (fsize >= 1024) {
      fsize /= 1024;
      unit = " GB";
    }
    fsize = Math.round(fsize * 10) / 10;
    fsize = fsize.toString() + unit;
    return fsize;
  },
  
  
  /*****************************************************************
   * Return number of hours (if applicable) and minutes
   */
  minutesToString: function(aMinutes) {
    var hours = "";
    if (aMinutes > 60) {
      hours = parseInt(aMinutes / 60, 10);
      aMinutes -= hours * 60;
      hours = hours.toString() + ((hours == 1) ? " Stunde, " : " Stunden, ");
    }
    return (hours + aMinutes.toString() +
            ((aMinutes == 1) ? " Minute" : " Minuten"));
  },
  
  
  /*****************************************************************
   * Convert seconds to h:mm:ss
   */
  durationToString: function(aDuration) {
    var result = "";
    if (aDuration > 0) {
      var hour = parseInt(aDuration / 3600, 10);
      var min = parseInt(aDuration / 60, 10) - (hour * 60);
      var sec = aDuration - (hour * 3600) - (min * 60);
      result = hour.toString() + ":" +
               min.toString().replace(/^(\d)$/, "0$1") + ":" +
               sec.toString().replace(/^(\d)$/, "0$1");
    }
    return result;
  },
  
  
  /*****************************************************************
   * Return a more digestable file name
   */
  
  convertFileName: function(aString) {
    var result = encodeURIComponent(aString);
    result = result.replace("'", "%27");
    result = result.replace("(", "%28");
    result = result.replace(")", "%29");
    result = result.replace("*", "%2A");
    result = result.replace("!", "%21");
    return result;
  },
  
  
  /**********************************************************************/
  /*                            L O G G I N G                           */
  /**********************************************************************/
  
  LOGLEVEL_OFF: 0,
  LOGLEVEL_ERROR: 1,
  LOGLEVEL_WARNING: 2,
  LOGLEVEL_INFORMATIONAL: 3,

  _logLevelNames: [ "none", "Error", "Warning", "Informational" ],
  _logLevelNClass: [ "", "nerr", "nwarn", "ninfo" ],
  _logLevelPClass: [ "", "perr", "pwarn", "pinfo" ],
  
  _logFile: null,
  _logLevel: this.LOGLEVEL_OFF,
  
  
  /*****************************************************************
   * Sets the log file.
   */
  set logFile(aFile) {
    this._logFile = aFile;
    this._logFileExists();      // if necessary create file and write header
    return this._logFile;
  },
  
  
  /**
   * Gets the log file.
   */
  get logFile() {
    return this._logFile;
  },
  
  
  /*****************************************************************
   * Empty the log file.
   */
  emptyLogFile: function() {
    if (this._logFile.exists())
      this._logFile.remove(false);
    this._logFileExists();      // create again and write header
  },
  
  
  /*****************************************************************
   * Gets the log file size
   */
  get logFileSize() {
    return this._logFile.clone().fileSize;
  },
  
  
  /*****************************************************************
   * Sets the log level.
   */
  set logLevel(aLevel) {
    this._logLevel = aLevel;
    return this._logLevel;
  },
  
  
  /**
   � Gets the log level.
   */
  get logLevel() {
    return this._logLevel;
  },
  
  
  /*****************************************************************
   * Appends a given string or html page to the log file.
   */
  writeLogEntry: function(aLevel, aHeader, aBody, isHtml) {
    if ((aLevel > this._logLevel) || (!this._logFileExists()))
      return;
    
    var flags = 0x02 | 0x10;
    var streamout = Components.classes["@mozilla.org/network/file-output-stream;1"].
                      createInstance(Components.interfaces.nsIFileOutputStream);
    streamout.init(this._logFile, flags, 0600, null);
    
    var msg = "<hr>\n<div class=\"" + this._logLevelPClass[aLevel] + "\">\n" +
              "<b>TOPIC:</b> " + aHeader + "<br>\n" +
              "<b>SEVERITY:</b> "+
              "<span class=\"" + this._logLevelNClass[aLevel] + "\">" +
              this._logLevelNames[aLevel] + "</span><br>\n" +
              "<b>TIMESTAMP:</b> " + (new Date()).toString() + "<br>\n";
    var msg2 = "";
    
    if (aBody.length > 0) {
      if (isHtml) {
        msg2 = aBody.replace(/\n|\r/g," ");
        msg2 = msg2.replace(/^.*<body\b[^>]*>/i, "");
        msg2 = msg2.replace(/<\/body\b\s*>.*/i, "");
        msg2 = msg2.replace(/<img\b[^>]*>/ig, "");
        msg2 = msg2.replace(/<\/?a\b[^>]*>/ig, "");
        msg2 = "<table class=\"htmlbox\"><tr><td>" +
              msg2 +
              "</table>\n";
      }
      else
        msg2 = "<div class=\"txtbox\">\n" +
               this.txt2html(aBody)+
               "</div>\n";
      
      msg2 = "<b>DETAILS:</b>\n" + msg2;
    }
    msg += msg2 + "</div>\n";
    streamout.write(msg, msg.length);
    
    streamout.close();
  },
  
  /*****************************************************************/
  
  _logFileExists: function() {
    if (this._logFile.exists()) {
      if (!this._logFile.isWritable()) 
        this._logFile.permissions |= 0600;
    }
    else {
      this._logFile.create(this._logFile.NORMAL_FILE_TYPE, 0600);
      if (!this._logFile.exists())     /* could not create file */
        return false;
    }
    
    if (this._logFile.clone().fileSize == 0) {
      var flags = 0x02 | 0x10;
      var streamout = Components.classes["@mozilla.org/network/file-output-stream;1"].
                        createInstance(Components.interfaces.nsIFileOutputStream);
      streamout.init(this._logFile, flags, 0600, null);
      var header = this.readProjectTextFile("chrome://checkyesss/locale/logfile_header.txt");
      streamout.write(header, header.length);
      streamout.close();
    }
    return true;
  },
  
  
  /**********************************************************************/
  /*                       F I L E   H A N D L I N G                    */
  /**********************************************************************/
  
  checkProjectDir: function(aProjectName) {
    var dirs = Components.classes["@mozilla.org/file/directory_service;1"].
                 getService(Components.interfaces.nsIProperties);
    var folder = dirs.get("ProfD", Components.interfaces.nsIFile);
    
    // append extension-specific subdirectory name
    folder.append(aProjectName);
    if (!folder.exists())
      folder.create(folder.DIRECTORY_TYPE, 0700);
  },
  
  /*****************************************************************/
  
  switchToBackup: function(aOrgFile) {
    var loglevel = this.LOGLEVEL_WARNING;
    if (aOrgFile.exists()) {
      aOrgFile.remove(false);
      loglevel = this.LOGLEVEL_ERROR;
    }
    var backupfile = Components.classes["@mozilla.org/file/local;1"].
                       createInstance(Components.interfaces.nsILocalFile);
    backupfile.initWithPath(aOrgFile.path + "~");
    if (backupfile.exists()) {
      backupfile.moveTo(null, aOrgFile.leafName);
      this.writeLogEntry(loglevel,
                         "cyCheckyesssUtil.switchToBackup",
                         "File " + aOrgFile.path + " had to be replaced " +
                         "by its backup copy",
                         false);
      return true;
    }
    return false;
  },
  
  /*****************************************************************/
  
  createBackup: function(aOrgFile) {
    if (aOrgFile.exists()) {
      var backupfile = Components.classes["@mozilla.org/file/local;1"].
                         createInstance(Components.interfaces.nsILocalFile);
      backupfile.initWithPath(aOrgFile.path + "~");
      if (backupfile.exists())
        backupfile.remove(false);
      aOrgFile.copyTo(null, backupfile.leafName);
    }
  },
  
  /*****************************************************************/
  
  readXMLDoc: function(aXmlFile) {
    
    if (aXmlFile.exists() && (aXmlFile.clone().fileSize > 0)) {
      var doc;
      var flags = 0x01;
      var parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].
                     createInstance(Components.interfaces.nsIDOMParser);
      var streamin = Components.classes["@mozilla.org/network/file-input-stream;1"].
                       createInstance(Components.interfaces.nsIFileInputStream);

      var bis = Components.classes["@mozilla.org/network/buffered-input-stream;1"].
                  createInstance(Components.interfaces.nsIBufferedInputStream);
      
      streamin.init(aXmlFile, flags, 00004, 0);
      bis.init(streamin, 4096);
      var fsize = bis.available();
      doc = parser.parseFromStream(bis, "UTF-8", fsize, "text/xml");
      streamin.close();
      
      var top = doc.documentElement;
      
      if ((top.tagName == "parserError") ||
          (top.namespaceURI == "http://www.mozilla.org/newlayout/xml/parsererror.xml")) {
        
        var converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
                          createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
        converter.charset = "UTF-8";
        var scriptableStream = Components.classes["@mozilla.org/scriptableinputstream;1"].
                                 createInstance(Components.interfaces.nsIScriptableInputStream);
        streamin.init(aXmlFile, flags, 00004, null);
        scriptableStream.init(streamin);
        var contents = converter.ConvertToUnicode(scriptableStream.read(fsize));
        streamin.close();
        
        this.writeLogEntry(this.LOGLEVEL_ERROR,
                           "cyCheckyesssUtil.readXMLDoc",
                           "XML parse error in aXmlFile " + aXmlFile.path + ":\n" +
                           contents,
                           false);
      }
      else
        return doc;
    }
    
    if (this.switchToBackup(aXmlFile)) {
      backupfile = null;                      // clean up
      return this.readXMLDoc(aXmlFile);       // retry with backup copy
    }
    return null;
  },
  
  /*****************************************************************/
  
  writeXMLDoc: function(aXmlFile, aDoc) {
    this.createBackup(aXmlFile);
    
    var serializer = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"].
                       createInstance(Components.interfaces.nsIDOMSerializer);
    
    var streamout = Components.classes["@mozilla.org/network/file-output-stream;1"].
                      createInstance(Components.interfaces.nsIFileOutputStream);
    var flags = (0x02 | 0x08 | 0x20);        // truncate or create file
    streamout.init(aXmlFile, flags, 0600, null);
    
    serializer.serializeToStream(aDoc, streamout, "UTF-8");
    //dump(aXmlFile.path + ": " + serializer.serializeToString(aDoc, "text/xml") + "\n");
    
    streamout.close();
  },
  
  /*****************************************************************/
  
  readProjectTextFile: function(aChromeURL) {
    var requestInst = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].
                        createInstance(Components.interfaces.nsIXMLHttpRequest);
    // this must be a synchronous request
    requestInst.open("GET", aChromeURL, false);
    requestInst.overrideMimeType('text/plain');
    requestInst.send(null);
    
    var result = requestInst.responseText;
    if (result !== null && result.length) {
      var converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
                        createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
      converter.charset = "UTF-8";
      result = converter.ConvertToUnicode(result);
    }
    return result;
  },
  
  
  /**********************************************************************/
  /*                     C A C H E   H A N D L I N G                    */
  /**********************************************************************/
  
  _cacheFile: null,
  _cacheAdminArray: null,
  _cacheContents: null,
  _cacheContentsKey: null,        // the key which _cacheContents belongs to
  
  _cacheXmlEmpty: "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"+
                  "<cachebag/>\n",
  
  /*****************************************************************/
  
  _getCacheInfo: function(aKey) {
    var fileError = false;
    
    if (this._cacheAdminArray == null)
      this._cacheAdminArray = new Array();
      
    if ((this._cacheContentsKey == aKey) && this._cacheAdminArray[aKey])
      return this._cacheAdminArray[aKey];
    
    this._cacheAdminArray[aKey] = null;
    this._cacheContentsKey = null;
    this._cacheContents = null;
    
    var doc = this.readXMLDoc(this._cacheFile);
    if (!doc)
      return null;
    
    var top = doc.documentElement;
    
    var adminnode = doc.evaluate("/cachebag/cache[@key=\""+aKey+"\"]",
                                 top,
                                 null,
                                 Components.interfaces.nsIDOMXPathResult.FIRST_ORDERED_NODE_TYPE,
                                 null).singleNodeValue;
    if (!adminnode) {
      this.writeLogEntry(this.LOGLEVEL_INFORMATIONAL,
                         "cyCheckyesssUtil._getCacheInfo",
                         "Cache miss for key '" + aKey + "'",
                         false);
      return null;
    }
    
    var datanode = doc.evaluate("/cachebag/cache[@key=\""+aKey+"\"]/data",
                                top,
                                null,
                                Components.interfaces.nsIDOMXPathResult.FIRST_ORDERED_NODE_TYPE,
                                null).singleNodeValue;
    
    if (!datanode) {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil._getCacheInfo",
                         "Missing data node for key '" + aKey + "'",
                         false);
      fileError = true;
    }
    
    var adminbag;
    
    if (!fileError) {
      adminbag = Components.classes["@mozdev.org/checkyesss/container;1"].
                   createInstance(Components.interfaces.cyICheckyesssContainer);
      adminbag.setPropertiesFromAttributes(adminnode);
      
      if (!adminbag.hasProperty("written") || !adminbag.hasProperty("expires")) {
        this.writeLogEntry(this.LOGLEVEL_ERROR,
                           "cyCheckyesssUtil._getCacheInfo",
                           "Missing attributes for key '" + aKey + "'",
                           false);
        fileError = true;
      }
    }
    
    if (!fileError) {
      this._cacheAdminArray[aKey] = adminbag;
      
      var contentsbag = Components.classes["@mozdev.org/checkyesss/container;1"].
                          createInstance(Components.interfaces.cyICheckyesssContainer);
      contentsbag.setPropertiesFromAttributes(datanode);

      this._cacheContents = contentsbag;
      this._cacheContentsKey = aKey;
          
      return this._cacheAdminArray[aKey];
    }
    
    if (this.switchToBackup(this._cacheFile)) {
      doc = null;                          // clean up
      adminnode = null;
      datanode = null;
      adminbag = null;
      return this._getCacheInfo(aKey);     // retry with backup copy
    }
    return null;
  },
  
  
  /*****************************************************************
   * Set the cache file.
   */
  set cacheFile(aValue) {
    this._cacheFile = aValue;
    this.writeLogEntry(this.LOGLEVEL_INFORMATIONAL,
                       "cyCheckyesssUtil.cacheFile",
                       "Using cache file " + aValue.path,
                       false);
    return this._cacheFile;
  },
  
  
  /*****************************************************************
   * Get the cache file.
   */
  get cacheFile() {
    return this._cacheFile;
  },
  
  
  /*****************************************************************
   * Check for the existence of a valid cache entry.
   */
  existCache: function(aKey) {
    if ((aKey == null) || (this._cacheFile == null))
      return false;
    
    if (!this._getCacheInfo(aKey))
      return false;
    
    return true;
  },
  
  
  /*****************************************************************
   * Write some contents to a cache file.
   */
  writeCache: function(aKey, aContents, aValidity, resetAge) {
    if (!this._cacheFile) {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.writeCache",
                         "Cache file name missing",
                         false);
      return;
    }
    
    var now = (new Date()).valueOf();
    var written = now;
    var expdate = now + aValidity;
    
    var entry = this._getCacheInfo(aKey);
    if (!resetAge) {
      // keep dates of modification and expiration
      if (entry && entry.hasProperty("written"))
        written = entry.getPropertyAsInt("written");
    }
    
    var adminbag = this._cacheAdminArray[aKey];
    if (!adminbag) {
      adminbag = Components.classes["@mozdev.org/checkyesss/container;1"].
                   createInstance(Components.interfaces.cyICheckyesssContainer);
      adminbag.setPropertyAsString("key", aKey);
    }
    
    adminbag.setPropertyAsInt("written", written);
    adminbag.setPropertyAsInt("expires", expdate);
    
    this._cacheAdminArray[aKey] = adminbag;
    
    // let's keep the contents in memory (for now)
    this._cacheContents = aContents;
    this._cacheContentsKey = aKey;
    
    var doc = this.readXMLDoc(this._cacheFile);
    
    if (!doc) {
      var parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].
                     createInstance(Components.interfaces.nsIDOMParser);
      
      doc = parser.parseFromString(this._cacheXmlEmpty, "text/xml");
    }
    
    var top = doc.documentElement;
    
    var adminnode = doc.evaluate("/cachebag/cache[@key=\""+aKey+"\"]",
                                 top,
                                 null,
                                 Components.interfaces.nsIDOMXPathResult.FIRST_ORDERED_NODE_TYPE,
                                 null).singleNodeValue;
    if (!adminnode) {
      var adminelem = doc.createElement("cache");
      adminbag.getPropertyAsAttribute("key", adminelem);
      top.appendChild(adminelem);
      adminnode = doc.evaluate("/cachebag/cache[@key=\""+aKey+"\"]",
                               top,
                               null,
                               Components.interfaces.nsIDOMXPathResult.FIRST_ORDERED_NODE_TYPE,
                               null).singleNodeValue;
    }
    adminbag.getPropertiesAsAttributes(adminnode);
    
    var datanode = doc.createElement("data");
    if (aContents)
      datanode = aContents.getPropertiesAsAttributes(datanode);
    
    var child = adminnode.firstChild;
    while (child) {
      if (child.nodeType == adminnode.ELEMENT_NODE &&
          child.nodeName == "data") {
        break;
      }
      child = child.nextSibling;
    }
    
    if (child)
      adminnode.replaceChild(datanode, child);
    else
      adminnode.appendChild(datanode);
    
    //dump(child ? "replaceChild\n" : "appendChild\n");
    
    this.writeXMLDoc(this._cacheFile, doc);
    
    this.writeLogEntry(this.LOGLEVEL_INFORMATIONAL,
                       "cyCheckyesssUtil.writeCache",
                       "Cache written for key " + aKey,
                       false);
  },
  
  
  /*****************************************************************
   * Get the age of the cached data.
   * @returns  number of ms since written
   */
  getCacheAge: function(aKey) {
    var age = -1;
    
    if (!this._cacheFile) {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.getCacheAge",
                         "Cache file name missing",
                         false);
      return -1;
    }
    
    var entry = this._getCacheInfo(aKey);
    if (entry && entry.hasProperty("written")) {
      var now = (new Date()).valueOf();
      var written = entry.getPropertyAsInt("written");
      age = now - written;
    }
    return age;
  },
  
  
  /*****************************************************************
   * Check if cache is expired.
   * 0: cache is expired
   * >0: number of ms until expired
   */
  getCacheValidity: function(aKey) {
    if (!this._cacheFile) {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.getCacheValidity",
                         "Cache file name missing",
                         false);
      return 0;
    }
    
    var entry = this._getCacheInfo(aKey);
    if (entry) {
      var now = (new Date()).valueOf();
      var expdate = entry.getPropertyAsInt("expires");
      var result = (now > expdate) ? 0 : (expdate - now);
      return result;
    }
    return 0;
  },
  
  
  /*****************************************************************
   * Returns the contents of the cache.
   */
  getCacheContents: function(aKey) {
    if (!this._cacheFile) {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.getCacheContents",
                         "Cache file name missing",
                         false);
      return null;
    }
    
    var entry = this._getCacheInfo(aKey);
    if (entry)
      return this._cacheContents;
    
    return null;
  },
  
  
  /*****************************************************************
   * Renames the key of the cache.
   * Returns true if successful,
   *         false if aOldKey does not exist or
   *               if aNewKey already exists and aForce is false.
   */
  renameCacheKey: function(aOldKey, aNewKey, aForce) {
    if (!this._cacheFile) {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.renameCacheKey",
                         "Cache file name missing",
                         false);
      return null;
    }
    
    var entry = this._getCacheInfo(aOldKey);
    if (!entry)
      return false;
    
    if (!aForce && this._cacheAdminArray[aNewKey])
      return false;
    
    var doc = this.readXMLDoc(this._cacheFile);
    if (!doc)
      return false;
    
    var top = doc.documentElement;
    var adminnode;
    
    if (this._cacheAdminArray[aNewKey]) {
      adminnode = doc.evaluate("/cachebag/cache[@key=\""+aNewKey+"\"]",
                               top,
                               null,
                               Components.interfaces.nsIDOMXPathResult.FIRST_ORDERED_NODE_TYPE,
                               null).singleNodeValue;
      adminnode.parentNode.removeChild(adminnode);
      delete this._cacheAdminArray[aNewKey];
    }
    adminnode = doc.evaluate("/cachebag/cache[@key=\""+aOldKey+"\"]",
                             top,
                             null,
                             Components.interfaces.nsIDOMXPathResult.FIRST_ORDERED_NODE_TYPE,
                             null).singleNodeValue;
    
    adminnode.setAttribute("key", aNewKey);
    
    this.writeXMLDoc(this._cacheFile, doc);
    
    // force reloading
    delete this._cacheAdminArray[aOldKey];
    this._cacheContentsKey = null;
    
    this.writeLogEntry(this.LOGLEVEL_INFORMATIONAL,
                       "cyCheckyesssUtil.renameCacheKey",
                       "Cache key renamed from " + aOldKey + " to " + aNewKey,
                       false);
    
    return true;
  },
  
  
  /*****************************************************************
   * Removes a cache entry.
   * Returns true if successful,
   *         false if aKey does not exist.
   */
  removeCache: function(aKey) {
    if (!this._cacheFile) {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.removeCache",
                         "Cache file name missing",
                         false);
      return null;
    }
    
    var entry = this._getCacheInfo(aKey);
    if (!entry)
      return false;
    
    var doc = this.readXMLDoc(this._cacheFile);
    if (!doc)
      return false;
    
    var top = doc.documentElement;
    
    var adminnode = doc.evaluate("/cachebag/cache[@key=\""+aKey+"\"]",
                                 top,
                                 null,
                                 Components.interfaces.nsIDOMXPathResult.FIRST_ORDERED_NODE_TYPE,
                                 null).singleNodeValue;
    adminnode.parentNode.removeChild(adminnode);
    delete this._cacheAdminArray[aKey];
    
    this.writeXMLDoc(this._cacheFile, doc);
    
    this.writeLogEntry(this.LOGLEVEL_INFORMATIONAL,
                       "cyCheckyesssUtil.removeCache",
                       "Cache entry removed for key " + aKey,
                       false);
    
    return true;
  },
  
  
  /* ****************************************************************** */
  /*                 T I M E   D A T A   L O G G I N G                  */
  /* ****************************************************************** */
  
  _creditLogFileTemplate: null,
  
  
  /*****************************************************************
   * Set the data log file.
   */
  set creditLogFileTemplate(aValue) {
    this._creditLogFileTemplate = aValue;
    this.writeLogEntry(this.LOGLEVEL_INFORMATIONAL,
                       "cyCheckyesssUtil.creditLogFileTemplate",
                       "Using credit log file template " + aValue,
                       false);
    return this._creditLogFileTemplate;
  },
  
  
  /*****************************************************************
   * Get the data log file.
   */
  get creditLogFileTemplate() {
    return this._creditLogFileTemplate;
  },
  
  
  /*****************************************************************
   * Write current time and aData to the log file
   */
  appendCreditRecord: function(aKey, aData) {
    if (this._creditLogFileTemplate) {
      var key = this.convertFileName(aKey);
      var filepath = this._creditLogFileTemplate.replace(/%s/, key);
      var file = Components.classes["@mozilla.org/file/local;1"].
                   createInstance(Components.interfaces.nsILocalFile);
      file.initWithPath(filepath);
      if (!file.exists())
        file.create(file.NORMAL_FILE_TYPE, 0600);
      else if (!file.isWritable())
        file.permissions |= 0600;
      if (!file.exists()) {
        this.writeLogEntry(this.LOGLEVEL_WARNING,
                           "cyCheckyesssUtil.appendCreditRecord",
                           "Cannot create file " + file.path,
                           false);
        return;
      }
      
      var now = (new Date()).valueOf();
      var logline = now + "\t" + aData + "\n";
      
      var converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
                        createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
      converter.charset = "UTF-8";
      var buffer = converter.ConvertFromUnicode(logline);
      buffer += converter.Finish();
      
      var flags = 0x02 | 0x10;
      var streamout = Components.classes["@mozilla.org/network/file-output-stream;1"].
                        createInstance(Components.interfaces.nsIFileOutputStream);
      streamout.init(file, flags, 0600, null);
      streamout.write(buffer, buffer.length);
      streamout.close();
    }
    else {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.appendCreditRecord",
                         "No file specified yet!",
                         false);
    }
  },
  
  /*****************************************************************/
  
  renameCreditKey: function(aOldKey, aNewKey) {
    if (this._creditLogFileTemplate) {
      var key = this.convertFileName(aOldKey);
      var oldpath = this._creditLogFileTemplate.replace(/%s/, key);
      key = this.convertFileName(aNewKey);
      var newpath = this._creditLogFileTemplate.replace(/%s/, key);
      
      try {         // try not to choke
        var oldfile = Components.classes["@mozilla.org/file/local;1"].
                        createInstance(Components.interfaces.nsILocalFile);
        oldfile.initWithPath(oldpath);
        
        if (oldfile.exists()) {
          var newfile = Components.classes["@mozilla.org/file/local;1"].
                          createInstance(Components.interfaces.nsILocalFile);
          newfile.initWithPath(newpath);
          
          oldfile.moveTo(null, newfile.leafName);
        }
      }
      catch (e) {
        this.writeLogEntry(this.LOGLEVEL_ERROR,
                           "cyCheckyesssUtil.renameCreditKey",
                           "aOldKey='"+aOldKey+"', aNewKey='"+aNewKey+"'\n"+
                           e.toString(),
                           false);
      }
    }
    else {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.renameCreditKey",
                         "No file specified yet!",
                         false);
    }
  },
  
  
  /*****************************************************************
   * Removes all entries of a certain aKey
   */
  removeCreditLog: function(aKey) {
    if (this._creditLogFileTemplate) {
      var key = this.convertFileName(aKey);
      var filepath = this._creditLogFileTemplate.replace(/%s/, key);
      var file = Components.classes["@mozilla.org/file/local;1"].
                   createInstance(Components.interfaces.nsILocalFile);
      file.initWithPath(filepath);
      if (file.exists())
        file.remove(false);
      // now the backup file
      filepath += "~";
      file.initWithPath(filepath);
      if (file.exists())
        file.remove(false);
    }
    else {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.removeCreditLog",
                         "No file specified yet!",
                         false);
    }
  },
  
  /*****************************************************************/
  
  readCreditLog: function(aKey, aFromTime, aToTime) {
    if (this._creditLogFileTemplate) {
      var key = this.convertFileName(aKey);
      var filepath = this._creditLogFileTemplate.replace(/%s/, key);
      var file = Components.classes["@mozilla.org/file/local;1"].
                   createInstance(Components.interfaces.nsILocalFile);
      file.initWithPath(filepath);
      
      // read input file as one chunk (ok until files grow big!)
      var converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
                        createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
      converter.charset = "UTF-8";
      var flags = 0x01;
      var streamin = Components.classes["@mozilla.org/network/file-input-stream;1"].
                       createInstance(Components.interfaces.nsIFileInputStream);
      var scriptableStream = Components.classes["@mozilla.org/scriptableinputstream;1"].
                               createInstance(Components.interfaces.nsIScriptableInputStream);
      streamin.init(file, flags, 00004, null);
      scriptableStream.init(streamin);
      var fsize = scriptableStream.available();
      var lines;
      try {
        lines = converter.ConvertToUnicode(scriptableStream.read(fsize)).split("\n");
      }
      catch (e) {
        this.writeLogEntry(this.LOGLEVEL_ERROR,
                           "cyCheckyesssUtil.readCreditLog",
                           "Could not read/convert file contents!\n" + e.toString(),
                           false);
        lines = new Array();
      }
      streamin.close();
      
      if (lines && (lines.length > 0) && (lines[lines.length-1].length == 0))
        lines.pop();
      
      // create a container
      var container = Components.classes["@mozdev.org/checkyesss/container;1"].
                        createInstance(Components.interfaces.cyICheckyesssContainer);
      
      // inspect all lines
      for (var i = 0; i < lines.length; i++) {
        var matches = lines[i].match(/^\s*(\d+)\t(.*)/);
        if (matches) {
          var time = Number(matches[1]);
          if (((aFromTime < 0) || (aFromTime <= time)) &&
              ((aToTime < 0) || (aToTime >= time)))
            container.setPropertyAsString(matches[1], matches[2]);
        }
      }
      return container;
    }
    else {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.readCreditLog",
                         "No file specified yet!",
                         false);
    }
    return null;
  },
  
  
  /* ****************************************************************** */
  /*                   C A L L   D E T A I L S   L O G                  */
  /* ****************************************************************** */
  
  _callsFileTemplate: null,
  _callsAdminArray: null,         // contains some statistics
  _callsContents: null,
  _callsContentsKey: null,        // the key which _callsContents belongs to
  
  _callsXmlEmpty: "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"+
                  "<callsbag written=\"0\"/>\n",
  
  /*****************************************************************/
  
  _getCallsInfo: function(aKey) {
    var fileError = false;
    
    if (this._callsAdminArray == null)
      this._callsAdminArray = new Array();
      
    if ((this._callsContentsKey == aKey) && this._callsAdminArray[aKey])
      return this._callsAdminArray[aKey];
    
    this._callsAdminArray[aKey] = null;
    this._callsContentsKey = null;
    this._callsContents = null;
    
    var key = this.convertFileName(aKey);
    var filepath = this._callsFileTemplate.replace(/%s/, key);
    var file = Components.classes["@mozilla.org/file/local;1"].
                 createInstance(Components.interfaces.nsILocalFile);
    file.initWithPath(filepath);
    
    var doc = this.readXMLDoc(file);
    if (!doc)
      return null;
    
    var top = doc.documentElement;
    
    var written = doc.evaluate("number(/callsbag/@written)",
                               top, null,
                               Components.interfaces.nsIDOMXPathResult.NUMBER_TYPE,
                               null).numberValue;
    
    var numCalls = doc.evaluate("count(/callsbag/call)",
                                top, null,
                                Components.interfaces.nsIDOMXPathResult.NUMBER_TYPE,
                                null).numberValue;
    
    var numMissing = doc.evaluate("count(/callsbag/missing)",
                                  top, null,
                                  Components.interfaces.nsIDOMXPathResult.NUMBER_TYPE,
                                  null).numberValue;
    
    /*
    var firstCall = doc.evaluate("min(/callsbag/call/@time)",
                                 top, null,
                                 Components.interfaces.nsIDOMXPathResult.NUMBER_TYPE,
                                 null).numberValue;
    
    var lastCall = doc.evaluate("max(/callsbag/call/@time)",
                                top, null,
                                Components.interfaces.nsIDOMXPathResult.NUMBER_TYPE,
                                null).numberValue;
    */
    
    var firstCall = (new Date()).valueOf();
    var lastCall = 0;
    var time;
    
    var callNodes = doc.evaluate("/callsbag/call", top, null,
                      Components.interfaces.nsIDOMXPathResult.UNORDERED_NODE_ITERATOR_TYPE,
                      null);
    
    var callNode = callNodes.iterateNext();
    while (callNode) {
      if (callNode.hasAttribute("time")) {
        time = Number(callNode.getAttribute("time"));
        if (time < firstCall)
          firstCall = time;
        if (time > lastCall)
          lastCall = time;
      }
      callNode = callNodes.iterateNext();
    }
    if (firstCall > lastCall)
      firstCall = lastCall;
    
    var adminbag;
    
    adminbag = Components.classes["@mozdev.org/checkyesss/container;1"].
                 createInstance(Components.interfaces.cyICheckyesssContainer);
    adminbag.setPropertyAsInt("written", written);
    adminbag.setPropertyAsInt("numCalls", numCalls);
    adminbag.setPropertyAsInt("numMissing", numMissing);
    adminbag.setPropertyAsInt("firstCall", firstCall);
    adminbag.setPropertyAsInt("lastCall", lastCall);
    //dump("written="+written+", numCalls="+numCalls+", numMissing="+numMissing+", firstCall="+(new Date(firstCall)).toString()+", lastCall="+(new Date(lastCall)).toString()+"\n");
    
    this._callsAdminArray[aKey] = adminbag;
    
    this._callsContents = doc;
    this._callsContentsKey = aKey;
          
    return this._callsAdminArray[aKey];
  },
  
  
  /*****************************************************************
   * Set the calls file name template.
   */
  set callsFileTemplate(aValue) {
    this._callsFileTemplate = aValue;
    this.writeLogEntry(this.LOGLEVEL_INFORMATIONAL,
                       "cyCheckyesssUtil.callsFileTemplate",
                       "Using credit calls file template " + aValue,
                       false);
    return this._callsFileTemplate;
  },
  
  
  /*****************************************************************
   * Get the calls file name template.
   */
  get callsFileTemplate() {
    return this._callsFileTemplate;
  },
  
  /*****************************************************************/
  
  getFirstCallTime: function(aKey) {
    if (!this._callsFileTemplate) {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.getFirstCallTime",
                         "File name template missing",
                         false);
      return 0;
    }
    
    var adminbag = this._getCallsInfo(aKey);
    
    if (adminbag && adminbag.hasProperty("firstCall"))
      return adminbag.getPropertyAsInt("firstCall");
    else
      return 0;
  },
  
  /*****************************************************************/
  
  getLastCallTime: function(aKey) {
    if (!this._callsFileTemplate) {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.getLastCallTime",
                         "File name template missing",
                         false);
      return 0;
    }
    
    var adminbag = this._getCallsInfo(aKey);
    
    if (adminbag && adminbag.hasProperty("lastCall"))
      return adminbag.getPropertyAsInt("lastCall");
    else
      return 0;
  },
  
  /*****************************************************************/
  
  getCallsWritten: function(aKey) {
    if (!this._callsFileTemplate) {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.getLastCallTime",
                         "File name template missing",
                         false);
      return 0;
    }
    
    var adminbag = this._getCallsInfo(aKey);
    
    if (adminbag && adminbag.hasProperty("written"))
      return adminbag.getPropertyAsInt("written");
    else
      return 0;
  },
  
  /*****************************************************************/
  
  getCallsCount: function(aKey) {
    if (!this._callsFileTemplate) {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.getLastCallTime",
                         "File name template missing",
                         false);
      return 0;
    }
    
    var adminbag = this._getCallsInfo(aKey);
    
    if (adminbag && adminbag.hasProperty("numCalls"))
      return adminbag.getPropertyAsInt("numCalls");
    else
      return 0;
  },
  
  /*****************************************************************/
  
  getCallsRecords: function(aKey, aFromTime, aToTime, aSelect) {
    if (!this._callsFileTemplate) {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.getCallsRecords",
                         "File name template missing",
                         false);
      return -1;
    }
    
    var callsArray = new Array();
    var adminbag = this._getCallsInfo(aKey);
    
    if (adminbag) {
      var top = this._callsContents.documentElement;
      
      var xpath = "/callsbag/call[@time>="+aFromTime+" and @time<="+aToTime;
      if (aSelect) {
        // add custom selection criteria
        xpath += " and ("+aSelect+")";
      }
      xpath += "]";
      
      var callNodes = this._callsContents.evaluate(
                        xpath, top, null,
                        Components.interfaces.nsIDOMXPathResult.UNORDERED_NODE_ITERATOR_TYPE,
                        null);
      
      var callNode = callNodes.iterateNext();
      var call;
      while (callNode) {
        call = Components.classes["@mozdev.org/checkyesss/container;1"].
                 createInstance(Components.interfaces.cyICheckyesssContainer);
        call.setPropertiesFromAttributes(callNode);
        callsArray.push(call);
        callNode = callNodes.iterateNext();
      }
    }
    
    // now sort the calls by time
    var sortByTime = function(a, b) {
      return a.getPropertyAsInt("time") - b.getPropertyAsInt("time");
    };
    callsArray.sort(sortByTime);
    
    return new ArrayEnumerator(callsArray);
  },
  
  /*****************************************************************/
  
  getCallsRecordsXML: function(aKey, aFromTime, aToTime, aSelect, aDomDoc) {
    if (!this._callsFileTemplate) {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.getCallsRecords",
                         "File name template missing",
                         false);
    }
    
    var callsNode = aDom.createNode("calls");
    var dest = aDom.documentElement.appendChild(callsNode);
    
    var callsArray = new Array();
    var adminbag = this._getCallsInfo(aKey);
    
    if (adminbag) {
      var top = this._callsContents.documentElement;
      
      var xpath = "/callsbag/call[@time>="+aFromTime+" and @time<="+aToTime;
      if (aSelect) {
        // add custom selection criteria
        xpath += " and ("+aSelect+")";
      }
      xpath += "]";
      
      var callNodes = this._callsContents.evaluate(
                        xpath, top, null,
                        Components.interfaces.nsIDOMXPathResult.UNORDERED_NODE_ITERATOR_TYPE,
                        null);
      
      var callNode = callNodes.iterateNext();
      var call;
      while (callNode) {
        importedNode = aDomDoc.importNode(callNode, true);
        dest.appendChild(importedNode);
        callNode = callNodes.iterateNext();
      }
    }
    
    var serializer = Components.classes["@mozilla.org/xmlextras/xmlserializer;1"].
                       createInstance(Components.interfaces.nsIDOMSerializer);
    dump(serializer.serializeToString(aDom, "text/xml") + "\n");
  },
  
  /*****************************************************************/
  
  renameCallsRecords: function(aOldKey, aNewKey) {
    var adminbag = this._getCallsInfo(aOldKey);
    if (!adminbag)
      return;
    
    if (this._callsAdminArray[aOldKey]) {
      this._callsAdminArray[aNewKey] = this._callsAdminArray[aOldKey];
      delete this._callsAdminArray[aOldKey];
    }
    
    if (this._callsContentsKey == aOldKey)
      this._callsContentsKey = aNewKey;
    
    if (this._callsFileTemplate) {
      var key = this.convertFileName(aOldKey);
      var oldpath = this._callsFileTemplate.replace(/%s/, key);
      key = this.convertFileName(aNewKey);
      var newpath = this._callsFileTemplate.replace(/%s/, key);
      
      try {         // try not to choke
        var oldfile = Components.classes["@mozilla.org/file/local;1"].
                        createInstance(Components.interfaces.nsILocalFile);
        oldfile.initWithPath(oldpath);
        
        if (oldfile.exists()) {
          var newfile = Components.classes["@mozilla.org/file/local;1"].
                          createInstance(Components.interfaces.nsILocalFile);
          newfile.initWithPath(newpath);
          
          oldfile.moveTo(null, newfile.leafName);
        }
      }
      catch (e) {
        this.writeLogEntry(this.LOGLEVEL_ERROR,
                           "cyCheckyesssUtil.renameCallsRecords",
                           e.toString(),
                           false);
      }
    }
    else {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.renameCallsRecords",
                         "No file specified yet!",
                         false);
    }
  },
  
  /*****************************************************************/
  
  removeCallsRecords: function(aKey) {
    var adminbag = this._getCallsInfo(aKey);
    if (!adminbag)
      return;
    
    if (this._callsAdminArray[aKey])
      delete this._callsAdminArray[aKey];
    
    if (this._callsContentsKey == aKey) {
      this._callsContentsKey = null;
      this._callsContents = null;
    }
    
    if (this._callsFileTemplate) {
      var key = this.convertFileName(aKey);
      var filepath = this._callsFileTemplate.replace(/%s/, key);
      var file = Components.classes["@mozilla.org/file/local;1"].
                   createInstance(Components.interfaces.nsILocalFile);
      file.initWithPath(filepath);
      if (file.exists())
        file.remove(false);
      // now the backup file
      filepath += "~";
      file.initWithPath(filepath);
      if (file.exists())
        file.remove(false);
    }
    else {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.removeCallsRecords",
                         "No file specified yet!",
                         false);
    }
  },
  
  /*****************************************************************/
  
  mergeCallsRecords: function(aKey, aCallsRecords, aComparePropertyNames) {
    if (!this._callsFileTemplate) {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.mergeCallsRecords",
                         "File name template missing",
                         false);
      return;
    }
    
    var callsArray = new Array();
    var adminbag = this._getCallsInfo(aKey);
    
    if (!adminbag) {
      var parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].
                     createInstance(Components.interfaces.nsIDOMParser);
      
      this._callsContents = parser.parseFromString(this._callsXmlEmpty, "text/xml");
      this._callsContentsKey = aKey;
    }
    
    var top = this._callsContents.documentElement;
    var changed = false;
    
    var props = aComparePropertyNames.split(/\s*,\s*/);
    
    while (aCallsRecords.hasMoreElements()) {
      var call = aCallsRecords.getNext().
                   QueryInterface(Components.interfaces.cyICheckyesssContainer);
      if (!call.hasProperty("time")) {
        this.writeLogEntry(this.LOGLEVEL_ERROR,
                           "cyCheckyesssUtil.mergeCallsRecords",
                           "No 'time' property specified",
                           false);
        continue;
      }
      
      var i;
      var xpath = "//call[@time='"+call.getPropertyAsInt("time")+"'";
      //var xpath = "//call[@time=\""+call.getPropertyAsInt("time")+"\"";
      for (i = 0; i < props.length; i++) {
        if (props[i].length &&
            props[i] != "time" &&
            call.hasProperty(props[i]))
          xpath += " and @"+props[i]+"='"+call.getPropertyAsString(props[i])+"'";
      }
      xpath += "]";
      //xpath = "//call";
      
      //dump(xpath+"\n");
      var duplicate = this._callsContents.evaluate(xpath, top, null,
                        Components.interfaces.nsIDOMXPathResult.FIRST_ORDERED_NODE_TYPE,
                        null).singleNodeValue;
                        //Components.interfaces.nsIDOMXPathResult.BOOLEAN_TYPE,
                        //null).booleanValue;
      if (!duplicate) {
        var callnode = this._callsContents.createElement("call");
        call.getPropertiesAsAttributes(callnode);
        top.appendChild(callnode);
        changed = true;
        //dump("adding\n");
      }
    }
    
    if (changed) {
      // remember time of modifications
      top.setAttribute("written", (new Date()).valueOf());
      
      var key = this.convertFileName(aKey);
      var filepath = this._callsFileTemplate.replace(/%s/, key);
      var file = Components.classes["@mozilla.org/file/local;1"].
                   createInstance(Components.interfaces.nsILocalFile);
      file.initWithPath(filepath);
      
      this.writeXMLDoc(file, this._callsContents);
      
      // force reanalysis next time the calls info is needed
      this.releaseCallsRecords();
      delete(this._callsAdminArray[aKey]);
    }
  },
  
  /*****************************************************************/
  
  storeCallsRecords: function(aKey, aCallsRecords, aFromTime, aToTime, aSelect) {
    if (!this._callsFileTemplate) {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.mergeCallsRecords",
                         "File name template missing",
                         false);
      return;
    }
    
    var callsArray = new Array();
    var adminbag = this._getCallsInfo(aKey);
    
    if (!adminbag) {
      var parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].
                     createInstance(Components.interfaces.nsIDOMParser);
      
      this._callsContents = parser.parseFromString(this._callsXmlEmpty, "text/xml");
      this._callsContentsKey = aKey;
    }
    
    var top = this._callsContents.documentElement;
    var changed = false;
    
    var xpath = "/callsbag/call[@time>="+aFromTime+" and @time<="+aToTime;
    if (aSelect) {
      // add custom selection criteria
      xpath += " and ("+aSelect+")";
    }
    xpath += "]";
    var callNode, call;
    while ((callNode = this._callsContents.evaluate(
                         xpath, top, null,
                         Components.interfaces.nsIDOMXPathResult.ANY_UNORDERED_NODE_TYPE,
                         null).singleNodeValue)) {
      //callNode = callNode.singleNodeValue;
      callNode.parentNode.removeChild(callNode);
      changed = true;
    }
    
    while (aCallsRecords.hasMoreElements()) {
      call = aCallsRecords.getNext().
               QueryInterface(Components.interfaces.cyICheckyesssContainer);
      if (!call.hasProperty("time")) {
        this.writeLogEntry(this.LOGLEVEL_ERROR,
                           "cyCheckyesssUtil.storeCallsRecords",
                           "No 'time' property specified",
                           false);
        continue;
      }
      
      time = call.getPropertyAsInt("time");
      if (time < aFromTime || time > aToTime) {
        continue;
      }
      
      callNode = this._callsContents.createElement("call");
      call.getPropertiesAsAttributes(callNode);
      top.appendChild(callNode);
      changed = true;
    }
    
    if (changed) {
      // remember time of modifications
      top.setAttribute("written", (new Date()).valueOf());
      
      var key = this.convertFileName(aKey);
      var filepath = this._callsFileTemplate.replace(/%s/, key);
      var file = Components.classes["@mozilla.org/file/local;1"].
                   createInstance(Components.interfaces.nsILocalFile);
      file.initWithPath(filepath);
      
      this.writeXMLDoc(file, this._callsContents);
      
      // force reanalysis next time the calls info is needed
      this.releaseCallsRecords();
      delete(this._callsAdminArray[aKey]);
    }
  },
  
  /*****************************************************************/
  
  mergeCallsMissing: function(aKey, aFromTime, aToTime) {
/*    if (!this._callsFileTemplate) {
      this.writeLogEntry(this.LOGLEVEL_ERROR,
                         "cyCheckyesssUtil.mergeCallsMissing",
                         "File name template missing",
                         false);
      return;
    }
    
    var adminbag = this._getCallsInfo(aKey);
    
    if (!adminbag) {
      var parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].
                     createInstance(Components.interfaces.nsIDOMParser);
      
      this._callsContents = parser.parseFromString(this._callsXmlEmpty, "text/xml");
      this._callsContentsKey = aKey;
    }
    
    var top = this._callsContents.documentElement;
    var changed = false;*/
    
    // TODO!!!  @@@
    // ...
    return;
  },
  
  /*****************************************************************/
  
  releaseCallsRecords: function() {
    this._callsContentsKey = null;
    this._callsContents = null;
  },
  
  /**********************************************************************/
  /*          B R O W S E R   W I N D O W   H A N D L I N G             */
  /**********************************************************************/
  
  /*****************************************************************
    Open the specified local file in a new window or tab
   */
  openFileInBrowser: function(aFileName, aUseTab) {
    var file = Components.classes["@mozilla.org/file/local;1"].
                 createInstance(Components.interfaces.nsILocalFile);
    file.initWithPath(aFileName);
    //convert to a URI
    var ios = Components.classes["@mozilla.org/network/io-service;1"].
                getService(Components.interfaces.nsIIOService);
    return this.openUrlInBrowser(ios.newFileURI(file).spec, aUseTab);
  },
  
  
  /*****************************************************************
    Open the specified URL in a new window or tab
   */
  openUrlInBrowser: function(aUrl, aUseTab) {
    if (!this.isThunderbird) {
      var windowMediator = Components.classes["@mozilla.org/appshell/window-mediator;1"].
        getService(Components.interfaces.nsIWindowMediator);
      var browserWindow = windowMediator.getMostRecentWindow("navigator:browser");
      
      if (!browserWindow)
        return null;
      
      var referrer = Components.classes["@mozilla.org/network/io-service;1"].
                       getService(Components.interfaces.nsIIOService).
                       newURI("http://checkyesss-extension/", null, null);
      var browser;
      
      if (aUseTab) {
        browser = browserWindow.getBrowser();
        var tab = browser.addTab(aUrl, referrer);
        browser.selectedTab = tab;
      }
      else {
        //return browserWindow.open(aUrl, referrer);
        browserWindow.open("about:blank");
        browserWindow = windowMediator.getMostRecentWindow("navigator:browser");
        browser = browserWindow.getBrowser();
        browser.loadURI(aUrl, referrer, null);
      }
      try {
        browserWindow.content.focus();
      } catch(e) {}
      return browser.contentWindow;
    }
    else {
      /* for Thunderbird open link in a remote browser */
      var messenger = Components.classes["@mozilla.org/messenger;1"].createInstance();
      messenger = messenger.QueryInterface(Components.interfaces.nsIMessenger);
      messenger.launchExternalURL(aUrl);
      return null;
    }
  },
  
  
  /*****************************************************************
    Check if a web page matching a certain regexp is open.
   */
  isWebPageOpen: function(aUrlRegExp) {
    var urlRegExp = new RegExp(aUrlRegExp);
    
    var ifaces = Components.interfaces;
    var mediator =
      Components.classes["@mozilla.org/appshell/window-mediator;1"].
        getService(ifaces.nsIWindowMediator);
    
    var winEnum = mediator.getXULWindowEnumerator(null);
    
    while (winEnum.hasMoreElements()) {
      var win = winEnum.getNext();
      
      if (!(win instanceof ifaces.nsIXULWindow)) {
        continue;
      }
      
      var iface = ifaces.nsIDocShellTreeItem;
      var docShellEnum = win.docShell.getDocShellEnumerator(iface.typeAll,
                                                            iface.ENUMERATE_FORWARDS);
      // now iterate through all documents in the window
      while (docShellEnum.hasMoreElements()) {
        var docShell = docShellEnum.getNext();
        if (docShell instanceof ifaces.nsIDocShell) {
          try {
            var domDocument = docShell.contentViewer.DOMDocument;
            if (domDocument instanceof ifaces.nsIDOMHTMLDocument) {
              if (domDocument.URL.match(urlRegExp)) {
                var requestor = docShell.QueryInterface(ifaces.nsIInterfaceRequestor);
                var itsWin = requestor.getInterface(ifaces.nsIDOMWindow);
                this.writeLogEntry(this.LOGLEVEL_INFORMATIONAL,
                                   "cyCheckyesssUtil.isWebPageOpen",
                                   "Checking for '" + aUrlRegExp +
                                   "', found open web page " +
                                   domDocument.URL,
                                   false);
                return itsWin;
              }
            }
          }
          catch (e) {}
        }
      }
    }
    return null;
  },
  
  /*****************************************************************/
  
  windowExists: function(aName) {
    var ifaces = Components.interfaces;
    var mediator =
      Components.classes["@mozilla.org/appshell/window-mediator;1"].
        getService(ifaces.nsIWindowMediator);
    
    var winEnum = mediator.getXULWindowEnumerator(null);
    
    while (winEnum.hasMoreElements()) {
      var win = winEnum.getNext();
      
      if (!(win instanceof ifaces.nsIXULWindow)) {
        continue;
      }
      
      var myDocShell = win.docShell.QueryInterface(ifaces.nsIDocShellTreeItem);
      if (myDocShell.nameEquals(aName)) {
        var requestor = myDocShell.QueryInterface(ifaces.nsIInterfaceRequestor);
        return requestor.getInterface(ifaces.nsIDOMWindow);
      }
    }
    return null;
  },
  
  /* ****************************************************************** */
  /*                 V E R S I O N   H A N D L I N G                    */
  /* ****************************************************************** */
  
  _oldVersion: null,
  _curVersion: null,
  _isUpdate: false,
  
  /*****************************************************************/
  
  get oldVersion() {
    return this._oldVersion;
  },
  
  get isUpdate() {
    return this._isUpdate;
  },
  
  /*****************************************************************/
  
  setVersion: function(aPrefName, aCurrentVersion) {
    var update = false;
    
    if (this._oldVersion == null) {
      var prefpath = aPrefName.match(/^(.+\.)([^\.]+)$/);
      var pbs = Components.classes["@mozilla.org/preferences-service;1"].
                  getService(Components.interfaces.nsIPrefService);
      var branch = pbs.getBranch(prefpath[1]);
      
      try {
        this._oldVersion = branch.getCharPref(prefpath[2]);
      }
      catch(e) {
        this._oldVersion = "";
      }
      if (this._oldVersion === null ||
          !this._oldVersion.length) {
        update = true;
        this._oldVersion = "0.0.0.0";
      }
      //dump("setVersion: _oldVersion='"+this._oldVersion+"', update="+update+"\n");
      
      branch.setCharPref(prefpath[2], aCurrentVersion);
      this._curVersion = aCurrentVersion;
      
      if (!update) {
        // could use nsIVersionComparator as well...
        oldarr = this._oldVersion.split(".");
        newarr = aCurrentVersion.split(".");
        for (var i = 0; i < newarr.length; i++) {
          if ((i >= oldarr.length) ||
              (Number(oldarr[i]) < Number(newarr[i]))) {
            update = true;
            break;
          }
        }
      }
      
      if (update) {
        this._isUpdate = true;
        
        this.writeLogEntry(this.LOGLEVEL_INFORMATIONAL,
                           "cyCheckyesssUtil.isUpdate",
                           "Updating from version " + this._oldVersion +
                           " to " + this._curVersion,
                           false);
      }
    }
    return update;
  },
  
  
  /*****************************************************************
   * if (compareVersions(appInfo.version, "1.5") >= 0) 
   *   // running under Version 1.5 or later
   */
  compareVersions: function(aVersion1, aVersion2) {
    var arr1 = aVersion1.split(".");
    var arr2 = aVersion2.split(".");
    var i;
    var result = 0;
    
    // fill up shorter array with '0'
    if (arr1.length > arr2.length) {
      for (i = 0; i < (arr1.length - arr2.length); i++)
        arr2.push("0");
    }
    else {
      for (i = 0; i < (arr2.length - arr1.length); i++)
        arr1.push("0");
    }
    
    for (i = 0; i < arr1.length; i++) {
      result = Number(arr1[i]) - Number(arr2[i]);
      if (result != 0)
        break;
    }
    
    return result;
  },
  
  
  /* ****************************************************************** */
  /*                 H O S T   A P P L I C A T I O N                    */
  /* ****************************************************************** */
  
  _checkedHostApp: false,
  _isThunderbird: false,
  
  get isThunderbird() {
    if (!this._checkedHostApp) {
      const FIREFOX_ID = "{ec8030f7-c20a-464f-9b0e-13a3a9e97384}";
      const THUNDERBIRD_ID = "{3550f703-e582-4d05-9a08-453d09bdfdc6}";
      var appInfo = Components.classes["@mozilla.org/xre/app-info;1"].
                          getService(Components.interfaces.nsIXULAppInfo);
      this._isThunderbird = (appInfo.ID == THUNDERBIRD_ID);
      this._checkedHostApp = true;
    }
    return this._isThunderbird;
  },
  
  /* ****************************************************************** */
  /*                   M U T U A L   E X C L U S I O N                  */
  /* ****************************************************************** */
  
  _firstTimeArr: null,
  
  /*****************************************************************/
  
  firstTime: function(aName) {
    if (this._firstTimeArr == null)
      this._firstTimeArr = new Array();
    if (!(" "+aName in this._firstTimeArr)) {
      this._firstTimeArr[" "+aName] = true;
      return true;
    }
    return false;
  },
  
  /*****************************************************************/
  
  //enterCS: function(aName)
  //leaveCS: function(aName)
  
  
  /**********************************************************************/
  /*          X P C O M - S P E C I F I C   M E T H O D S               */
  /**********************************************************************/

  QueryInterface: function (aIID) {
    if (!aIID.equals(Components.interfaces.cyICheckyesssUtil) &&
        !aIID.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
  
};


/******************************************************************************
 * XPCOM Functions for construction and registration
 ******************************************************************************/

var gModule = {
  _firstTime: true,
  
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType) {
    if (this._firstTime) {
      this._firstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    }
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID,
                                     CLASS_NAME,
                                     CONTRACT_ID,
                                     aFileSpec,
                                     aLocation,
                                     aType);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType) {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        
  },
  
  getClassObject: function(aCompMgr, aCID, aIID) {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return gFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) {
    return true;
  }
};

var gFactory = {
  _service: null,
  
  createInstance: function (aOuter, aIID) {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    
    if (!aIID.equals(Components.interfaces.cyICheckyesssUtil) &&
        !aIID.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    
    if (this._service == null)
      this._service = new CheckyesssUtilServiceClass();

    return this._service.QueryInterface(aIID);
  }
};

function NSGetModule(aCompMgr, aFileSpec) {
  return gModule;
}

