/* ***** BEGIN LICENSE BLOCK *****
   - Version: MPL 1.1/GPL 2.0/LGPL 2.1
   -
   - The contents of this file are subject to the Mozilla Public License Version
   - 1.1 (the "License"); you may not use this file except in compliance with
   - the License. You may obtain a copy of the License at
   - http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS IS" basis,
   - WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
   - for the specific language governing rights and limitations under the
   - License.
   -
   - The Original Code is CheckYesss.
   -
   - The Initial Developer of the Original Code is
   - Thomas Reitmayr <treitmayr@yahoo.com>.
   - Portions created by the Initial Developer are Copyright (C) 2006
   - the Initial Developer. All Rights Reserved.
   -
   - Contributor(s):
   - -
   -
   - Alternatively, the contents of this file may be used under the terms of
   - either the GNU General Public License Version 2 or later (the "GPL"), or
   - the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
   - in which case the provisions of the GPL or the LGPL are applicable instead
   - of those above. If you wish to allow use of your version of this file only
   - under the terms of either the GPL or the LGPL, and not to allow others to
   - use your version of this file under the terms of the MPL, indicate your
   - decision by deleting the provisions above and replace them with the notice
   - and other provisions required by the LGPL or the GPL. If you do not delete
   - the provisions above, a recipient may use your version of this file under
   - the terms of any one of the MPL, the GPL or the LGPL.
   -
   - ***** END LICENSE BLOCK ***** */

const CLASS_ID = Components.ID("{207759c0-8de9-11da-a72b-0800200c9a66}");
const CLASS_NAME = "Checkyesss Web Access Component";
const CONTRACT_ID = "@mozdev.org/checkyesss/web_access;1";

const TO_HTTP = 120;                  // default HTTP timeout is 120 seconds


/******************************************************************************/
/**
/* ArrayEnumerator courtesy of Brendan Eich <brendan@mozilla.org>
 */
function ArrayEnumerator(array) {
  this.array = array;
  this.index = 0;
}
ArrayEnumerator.prototype = {
  hasMoreElements: function () { return this.index < this.array.length; },
  getNext: function () { return (this.index < this.array.length) ?
                         this.array[this.index++] : null; }
};

/**
/* StringEnumerator
 */
function StringEnumerator(array) {
  this.array = array;
  this.index = 0;
}
StringEnumerator.prototype = {
  hasMore: function() { return this.index < this.array.length; },
  getNext: function () { return (this.index < this.array.length) ?
                       this.array[this.index++] : null; },
  QueryInterface: function (iid) {
    if (!iid.equals(Components.interfaces.nsIStringEnumerator) &&
        !iid.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
};

/******************************************************************************
 * cyWebAccess component
 ******************************************************************************/

function CheckyesssWebClass() {
  //this.wrappedJSObject = this;
  this._start();
}


CheckyesssWebClass.prototype = {
  
  _activeSession: false,
  _activeRequest: false,
  _requestTimeout: TO_HTTP,
  
  _requestTimer: null,
  _requestInst: null,
  _obs: null,
  _utils: null,
  
  _topic: null,
  _userAgent: null,
  
  _urlRequest: null,
  _urlResponse: null,
  _defaultCharset: null,
  
  _status: 0,
  _statusText: null,
  _contentType: null,
  
  _contentsAll: null,
  _contentsHTML: null,
  _contentsText: null,
  _contentsLinks: null,
  
  
  /**************************************************************/
  
  _start: function() {
    // set components
    this._utils = Components.classes["@mozdev.org/checkyesss/util_service;1"].
                    getService(Components.interfaces.cyICheckyesssUtil);
    
    this._requestTimer = Components.classes["@mozilla.org/timer;1"].
                           createInstance(Components.interfaces.nsITimer); 
    this._obs = Components.classes["@mozilla.org/observer-service;1"].
                  getService(Components.interfaces.nsIObserverService);
    
    try {
      this._userAgent = Components.classes["@mozilla.org/network/protocol;1?name=http"].
                          getService(Components.interfaces.nsIHttpProtocolHandler).
                            userAgent;
      this._userAgent += " Checkyesss/0.6";
    }
    catch (e) {
      this._userAgent = null;
    }
  },
  
  /**************************************************************/
  
  get observerTopic() {
    return this._topic;
  },
  
  set observerTopic(value) {
    this._topic = value;
    return value;
  },
  
  /**************************************************************/
  
  get requestTimeout() {
    return this._requestTimeout;
  },
  
  set requestTimeout(value) {
    this._requestTimeout = value;
    return value;
  },
  
  /**************************************************************/
  
  startSession: function() {
    if (this._activeSession)
      throw new Error("cyCheckyesssWeb: Session already active!");
    
    this._activeSession = true;
  },
  
  
  get activeSession() {
    return this._activeSession;
  },
  
  
  endSession: function() {
    //dump("endSession\n");
    if (this._activeSession) {
      if (this._activeRequest)
        this.abortRequest();
      
      this._cleanUp();
      // TODO: clean up more
      this._activeSession = false;
    }
  },
  
  /**************************************************************/
  
  startRequest: function(aURL, aMethod, aReferrer, aFormData, aDefaultCharset) {
    // prepare the form data for debug output
    var debug_formdata = "";
    if (aFormData)
      debug_formdata = aFormData;
    else {
      var matches = aURL.path.match(/\?(.+)/);
      if (matches)
        debug_formdata = matches[1];
    }
    
    if (debug_formdata)
      debug_formdata = "\nFormData: " +
                       debug_formdata.replace(/\b((?:(?:login)?passw(?:or[dt])?|webpw|pw)=)([^&]+)/i,
                                              "$1<hidden>");
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                              "cyCheckyesssWeb.startRequest",
                              "URL: " + aURL.spec + "\nMethod: " + aMethod +
                              debug_formdata, false);
    
    if (this._activeRequest) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesssWeb.startRequest",
                                "Another request is already in progress",
                                false);
      this._status = 1;
      this._statusText = "Eine andere Server-Anfrage ist bereits aktiv";
      if (this._activeSession)
        this._reqSeqWebResult("error");
      else if (this._topic)
        this._obs.notifyObservers(this, this._topic, "error");
      return;
    }
    
    // reset some variables
    this._cleanUp();
    this._urlRequest = aURL;
    this._defaultCharset = (aDefaultCharset) ? aDefaultCharset : "iso-8859-1";
    // more to come
    this._activeRequest = true;
    
    this._requestInst = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].
                          createInstance(Components.interfaces.nsIXMLHttpRequest);
    var url = aURL.spec;
    var send = null;
    if (aMethod.toUpperCase() == "POST") {
      aMethod = "POST";
      send = aFormData;
    }
    else {
      aMethod = "GET";
      if ((aFormData != null) && aFormData.length)
        url += "?" + aFormData;
    }
    this._requestInst.open(aMethod, url, true);
    if (aMethod == "POST")
      this._requestInst.setRequestHeader("Content-Type",
                                         "application/x-www-form-urlencoded");
    if (this._userAgent)
      this._requestInst.setRequestHeader("User-Agent", this._userAgent);
    
    if (aReferrer)
      this._requestInst.setRequestHeader("Referer", aReferrer);
    
    try {
      // we need this to get binary data!!
      // see http://mgran.blogspot.com/2006/08/downloading-binary-streams-with.html
      // HOWEVER, only firefox and seamonkey are confirmed to work with this,
      // Mozilla Suite does not support it (results in incomplete downloads!)
      var appname = Components.classes["@mozilla.org/xre/app-info;1"].
                      getService(Components.interfaces.nsIXULAppInfo).name.toLowerCase();
      if (appname == "seamonkey" ||
          appname == "firefox") {
        this._requestInst.overrideMimeType('text/plain; charset=x-user-defined');
      }
    }
    catch (e) { null; }
    
    var self = this;
    this._requestInst.onload  = function() { self._requestDone(); };
    this._requestInst.onerror = function() { self._requestDone(); };
    this._requestTimer.initWithCallback(this,
                                        this._requestTimeout * 1000,
                                        Components.interfaces.nsITimer.TYPE_ONE_SHOT);
    try {
      this._requestInst.send(send);
    }
    catch (e) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesssWeb.startRequest",
                                e.toString(), false);
      this._status = 2;
      this._statusText = "Fehler beim Verbindungsaufbau:\n" + e.toString();
      if (this._activeSession)
        this._reqSeqWebResult("error");
      else if (this._topic)
        this._obs.notifyObservers(this, this._topic, "error");
    }
  },
  
  /**************************************************************/
  
  get activeRequest() {
    return this._activeRequest;
  },
  
  /**************************************************************/
  
  abortRequest: function() {
    if (this._activeRequest) {
      this._requestTimer.cancel();
      this._requestInst.abort();
      this._requestInst = null;
      this._cleanUp();
      
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "cyCheckyesssWeb.abortRequest",
                                "HTTP access had to be aborted.", false);
      //...
      
      this._activeRequest = false;
      // do NOT send "error" or "done" notification
    }
  },
  
  /**************************************************************/
  
  _requestDone: function() {
    var error = false;
    var req = this._requestInst;
    
    this._requestTimer.cancel();
    
    if (!this._activeRequest) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssWeb._requestDone",
                                "Unexpected server response", false);
      this._status = 3;
      this._statusText = "Unerwartete Antwort vom Server!";
      error = true;
    }
    this._activeRequest = false;
    
    if (!error) {
      try {
        req.status;
      }
      catch(e) {
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                  "cyCheckyesssWeb._requestDone",
                                  "No connection to server", false);
        this._status = 4;
        this._statusText = "Keine Verbindung zum Server!";
        error = true;
      }
    }
    
    if (!error && (req.status > 200)) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesssWeb._requestDone",
                                "Communication error #" +
                                req.status + "\n" +
                                req.statusText, false);
      this._status = req.status;
      this._statusText = req.statusText;
      error = true;
    }
    
    if (!error && (req.responseText.length == 0)) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesssWeb._requestDone",
                                "Received empty web page", false);
      this._status = 5;
      this._statusText = "Leere Seite empfangen";
      error = true;
    }
    
    this._contentType = null;
    
    if (!error) {
      this._status = req.status;
      this._statusText = req.statusText;
      
      var contentType = req.getResponseHeader("Content-Type");
      if (contentType) {
        var matches = contentType.match(/^([^\s;]+)/);
        if (matches)
          this._contentType = matches[1];
      }
      /* convert according to specified character-set */
      var charset = contentType;
      if (!charset || !charset.match(/\bcharset=\w/)) {
        // try to find a meta tag specifying the character set
        // Note: We just parse the raw contents but do not mask off comments
        //       or other stuff -> might not work for all HTML code
        var metatag = req.responseText.match(/<meta\s+[^>]*\bhttp-equiv\s*=\s*["']*Content-Type[^>]*>/i);
        if (metatag)
          charset = metatag[0];
        else
          charset = "charset=" + this._defaultCharset;
      }
      if (charset && charset.match(/\bcharset=\w/)) {
        charset = charset.match(/\bcharset=([\w\-]+)/);
        var converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
                          createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
        try {
          converter.charset = charset[1];
          this._contentsAll = converter.ConvertToUnicode(req.responseText);
        }
        catch (e) {
          // could not convert -> just take the text as it is
          this._contentsAll = req.responseText;
          this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                    "cyCheckyesssWeb._requestDone",
                                    "Could not convert from character-set '" +
                                    charset[1] + "':\n" + e.toString(), false);
        }
      }
      else
        this._contentsAll = req.responseText;
      
      /* works only for mozilla -> don't use it
      try {
        this._urlResponse = req.channel.URI;
      }
      catch (e) {
        this._urlResponse = null;
      }*/
    }
    
    this._requestInst = null;
    req = null;
    
    // either notify the request sequencer or the application
    if (this._activeSession)
      this._reqSeqWebResult((error) ? "error" : "done");
    else if (this._topic)
      this._obs.notifyObservers(this, this._topic, (error) ? "error" : "done");
  },
  
  /**************************************************************/
  
  get urlRequest() {
    return this._urlRequest;
  },
  
  /**************************************************************/
  
  get status() {
    return this._status;
  },
  
  get statusText() {
    return this._statusText;
  },
  
  get contentType() {
    return this._contentType;
  },
  
  get contentsAll() {
    return this._contentsAll;
  },
  
  get contentsHTML() {
    if ((this._contentsHTML == null) &&
        (this._contentsAll != null)) {
      this._contentsHTML = this._contentsAll.replace(/\n|\r/g," ");
      this._contentsHTML = this._contentsHTML.replace(/<!--.*?-->/g, "");
      this._contentsHTML = this._contentsHTML.replace(/<(script|style)\b.*?<\/\s*\1\s*>/gi, "");
      this._contentsHTML = this._contentsHTML.replace(/\s{2,}/g, " ");
    }
    return this._contentsHTML;
  },
  
  get contentsText() {
    if ((this._contentsText == null) &&
        (this._contentsAll != null)) {
      // first replace certain tags by a space
      this._contentsText = this.contentsHTML.
        replace(/<\/?(?:div|p|br|img|table|tr|td|th|dl|li|dd)(?:(?:\s+[\w\-]+(?:\s*=\s*(?:".*?"|'.*?'|[^'">\s]+))?)+\s*|\s*)\/?>/ig, " ");
      // now remove the remaining tags
      this._contentsText = this._contentsText.
        replace(/<[\/!]?\w+(?:(?:\s+[\w\-]+(?:\s*=\s*(?:".*?"|'.*?'|[^'">\s]+))?)+\s*|\s*)\/?>/ig, "");
      // get rid of excess spaces
      this._contentsText = this._contentsText.replace(/\s{2,}/g, " ");
      this._contentsText = this._utils.html2txt(this._contentsText);
    }
    return this._contentsText;
  },
  
  /**************************************************************/
  
  patchContents: function(aPattern, aReplacement) {
    var regexp = RegExp(aPattern, "g");
    this._contentsAll = this._contentsAll.replace(regexp, aReplacement);
    this._contentsHTML = null;
    this._contentsText = null;
  },
  
  /**************************************************************/
  
  findLink: function(aPattern) {
    if (this._contentsLinks == null) {
      var text = this.contentsHTML.
        replace(/<\/?(?!a)\w+(?:(?:\s+\w+(?:\s*=\s*(?:".*?"|'.*?'|[^'">\s]+))?)+\s*|\s*)\/?>/ig, "");
      this._contentsLinks = text.
        match(/(<a(?:(?:\s+\w+(?:\s*=\s*(?:".*?"|'.*?'|[^'">\s]+))?)+\s*|\s*)\/?>[^<]*<\/a\s*>)/ig);
      text = null;
    }
    
    var pattern = new RegExp(aPattern, "i");
    for (var i = 0; i < this._contentsLinks.length; i++) {
      if (this._contentsLinks[i].match(pattern)) {
        var addr = this._contentsLinks[i].match(/\bhref\s*=\s*(".*?"|'.*?'|[^'">\s]+)/i);
        if (addr)
          return addr[1].replace(/^(["'])(.*)\1$/, "$2");
        else
          break;
      }
    }
    return null;
  },
  
  /**************************************************************/
  
  findTable: function(aRegionPattern, aPattern) {
    var text;
    if (aRegionPattern) {
      text = this.contentsHTML.match(new RegExp(aRegionPattern, "i"));
      if (text)
        text = text[0];
    }
    else
      text = this.contentsHTML;
    
    if (text === null || text.length == 0) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesssWeb.findTable",
                                "Could not find aRegionPattern '" +
                                aRegionPattern + "'", false);
      return null;
    }
    
    var pattern = "<table\\b((?!<table).)*"+aPattern+".*?</table\\s*>";
    pattern = new RegExp(pattern, "i");
    
    text = text.match(pattern);
    if (text) {
      text = text[0];
      text = text.replace(/<(?!t[dhr]\b)\/?\w+(?:(?:\s+\w+(?:\s*=\s*(?:".*?"|'.*?'|[^'">\s]+))?)+\s*|\s*)\/?>/ig, "");
      text = text.replace(/(<t[dhr])\b[^>]*/ig, "$1");
      
      var rows, cells;
      var x, y;
      var table = new Array();
      rows = text.split(/<tr>/i);
      if (rows.length > 1) {
        for (y = 1; y < rows.length; y++) {
          cells = rows[y].split(/<(?:th|td)>/i);
          if (cells.length > 1) {
            cells.shift();
            for (x = 0; x < cells.length; x++) {
              cells[x] = this._utils.html2txt(cells[x]).replace(/^\s*(.*?)\s*$/, "$1");
            }
            table.push(new StringEnumerator(cells));
          }
        }
      }
      
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "cyCheckyesssWeb.findTable",
                                "Extracted " + table.length + " rows", false);
      
      return new ArrayEnumerator(table);
    }
    else {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                "cyCheckyesssWeb.findTable",
                                "Could not find table containing '" +
                                aPattern + "'", false);
      return null;
    }
  },
  
  /**************************************************************/
  
  _cleanUp: function() {
    this._statusText = null;
    this._contentsAll = null;
    this._contentsHTML = null;
    this._contentsText = null;
    this._contentsLinks = null;
    this._urlResponse = null;
  },
  
  /**************************************************************/
  /********************   REQUEST SEQUENCER   *******************/
  
  _reqSeq: null,
  _reqSeqIndex: 0,
  _reqSeqResult: "",
  _reqSeqStop: false,
  _reqSeqThrottle: 0,
  _reqSeqCallback: null,
  _reqSeqLastFunction: null,   // stores the name of the last function called
  
  
  startRequestSequence: function(aSequence, aCallback) {
    if (this._activeSession) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssWeb.startRequestSequence",
                                "Session already active!",
                                false);
      return;
    }
    
    this._activeSession = true;
    this._reqSeq = new Array();
    this._reqSeqIndex = 0;
    this._reqSeqResult = "";
    this._reqSeqStop = false;
    this._reqSeqThrottle = 0;
    this._reqSeqCallback = aCallback;
    this._reqSeqLastFunction = "";
    
    // copy the complete program into an array
    var index = 0;
    while (aSequence.hasMore()) {
      this._reqSeq[index++] = aSequence.getNext();
    }
    
    // now start the sequencer
    this._requestSequencer();
  },
  
  
  _requestSequencer: function() {
    var jumpIndex;
    
    while (!this._reqSeqStop) {
      var action = this._reqSeq[this._reqSeqIndex];
      //dump("reqSeq: " + action + "\n");
      if (typeof(action) != "string") {
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesssWeb._requestSequencer",
                                  "Unexpeced end of command sequence in line " + 
                                  (this._reqSeqIndex + 1),
                                  false);
        break;
      }
      
      var cmd = action.split(/\s+/);
      
      if (cmd[0].lastIndexOf(":") == (cmd[0].length -1)) {
        // this is a label, just skip
        this._reqSeqIndex++;
        continue;
      }
      
      // check for conditional execution
      var cond = cmd[0].split(/\//);
      if (cond.length > 1) {
        if (cond[1] != this._reqSeqResult) {
          // condition not satisfied
          this._reqSeqIndex++;
          continue;                  // --> process next command
        }
        cmd[0] = cond[0];
      }
      
      // now process the command
      switch (cmd[0]) {
        
        case "throttle":
          this._reqSeqIndex++;
          this._reqSeqThrottle = Number(cmd[1]);
          break;
          
        case "jump":
          var jump = cmd[1];
          if (jump.match(/^[+\-]\d+/))        // relative jump
            this._reqSeqIndex += parseInt(jump, 10);
          else if (jump.match(/^\d+/))        // absolute jump
            this._reqSeqIndex = parseInt(jump, 10);
          else {
            jumpIndex = this._reqSeqFindLabel(jump);
            if (jumpIndex >= 0)
              this._reqSeqIndex = jumpIndex;
            else
              this._reqSeqIndex++;           // just continue for unknown labels
          }
          break;
          
        case "wait":
          this._reqSeqIndex++;
          var time = Number(cmd[1]);
          if (time > 0) {
            if (this._waitTimer)
              this._waitTimer.cancel();
            else
              this._waitTimer = Components.classes["@mozilla.org/timer;1"].
                                  createInstance(Components.interfaces.nsITimer);
            this._waitTimer.initWithCallback(this, time, 
                                             Components.interfaces.nsITimer.TYPE_ONE_SHOT);
            // leave the loop until timer expires
            return;
          }
          break;
          
        case "exit":
          this._reqSeqStop = true;
          break;
          
        case "error":
          // skip any explicit calls to error
          this._reqSeqIndex++;
          this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                    "cyCheckyesssWeb._requestSequencer",
                                    "Command 'error' must not be called explicitly " +
                                    "in line " + this._reqSeqIndex,
                                    false);
          break;
          
        default:
          this._reqSeqIndex++;
          this._reqSeqLastFunction = cmd[0];
          try {
            this._reqSeqResult = this._reqSeqCallback.webSeqCallback(cmd[0]);
          }
          catch (e) {
            this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                      "cyCheckyesssWeb._requestSequencer",
                                      "Exception in command '" + cmd[0] +
                                      "' (line " + this._reqSeqIndex + "):\n" +
                                      e.toString(),
                                      false);
            this._reqSeqResult = "";
          }
          // if a http transfer was initiated by the function, wait for it to finish
          if (this._activeRequest) {
            return;
          }
          break;
      }
    }
    
    // finally call 'exit'
    this._reqSeqLastFunction = "exit";
    this._activeSession = false;
    try {
      this._reqSeqCallback.webSeqCallback("exit");
    }
    catch (e) {
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssWeb._requestSequencer",
                                "Exception in command 'exit':\n" +
                                e.toString(),
                                false);
    }
  },
  
  /**************************************************************/
  
  _reqSeqFindLabel: function(aLabel) {
    var i;
    var label = aLabel + ":";
    
    for (i = 0; i < this._reqSeq.length; i++) {
      if (typeof(this._reqSeq[i]) == "string" && this._reqSeq[i] == label)
        return i;
    }
    return -1;
  },
  
  /**************************************************************/
  
  _reqSeqWebResult: function(aStatus) {
    if (aStatus != "done") {
      // there was an error when loading the web page!
      this._reqSeqLastFunction = "error";
      this._reqSeqResult = "exit";
      try {
        this._reqSeqResult = this._reqSeqCallback.webSeqCallback("error");
      }
      catch (e) {
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesssWeb._requestSequencer",
                                  "Exception in command 'error' (line " +
                                  this._reqSeqIndex + "):\n" +
                                  e.toString(),
                                  false);
        this._reqSeqResult = "exit";
      }
      // a return value "exit" stops the sequencer!
      if (this._reqSeqResult == "exit")
        this._reqSeqStop = true;
      else
        aStatus = "done";                // continue as usual
    }

    if (aStatus == "done" &&
        this._reqSeqThrottle > 0  &&
        this._reqSeqLastFunction != "logout") {
      // throttle the http requests by waiting the specified amount of time
      if (this._waitTimer)
        this._waitTimer.cancel();
      else
        this._waitTimer = Components.classes["@mozilla.org/timer;1"].
                            createInstance(Components.interfaces.nsITimer);
      this._waitTimer.initWithCallback(this, this._reqSeqThrottle, 
                                       Components.interfaces.nsITimer.TYPE_ONE_SHOT);
    }
    else {
      // no throttling, so just continue
      this._requestSequencer();
    }
  },
  
  /**************************************************************/
  
  abortRequestSequence: function(doLogout) {
    //dump("abortRequestSequence\n");
    if (this._activeSession) {
      if (this._activeRequest)
        this.abortRequest();
      
      if (this._waitTimer)
        this._waitTimer.cancel();
      
      this._cleanUp();
      // TODO: clean up more
      this._activeSession = false;
    }
  },
  
  /**************************************************************/
  
  /** nsITimerCallback **/
  
  notify: function(timer) {
    //dump("notify\n");
    if (timer == this._requestTimer) {
      this._requestInst.abort();
      this._requestInst = null;
      this._cleanUp();
      this._activeRequest = false;
      
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
                                "cyCheckyesssWeb.notify",
                                "Connection timeout", false);
      
      this._status = 10;
      this._statusText = "Timeout beim Verbindungsaufbau";
      if (this._activeSession)
        this._reqSeqWebResult("error");
      else if (this._topic)
        this._obs.notifyObservers(this, this._topic, "error");
    }
    else if (timer == this._waitTimer) {
      // waited long enough, now continue with the program
      this._requestSequencer();
    }
  },
  
  /**************************************************************/
  
  QueryInterface: function (iid) {
    if (!iid.equals(Components.interfaces.cyICheckyesssWeb) &&
        !iid.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
  
};


/******************************************************************************
 * XPCOM Functions for construction and registration
 ******************************************************************************/

var gModule = {
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType) {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID,
                                     CLASS_NAME,
                                     CONTRACT_ID,
                                     aFileSpec,
                                     aLocation,
                                     aType);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType) {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        
  },
  
  getClassObject: function(aCompMgr, aCID, aIID) {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return gFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) {
    return true;
  }
};

var gFactory = {
  
  createInstance: function (aOuter, aIID) {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    
    if (!aIID.equals(Components.interfaces.cyICheckyesssWeb) &&
        !aIID.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    
    return ((new CheckyesssWebClass()).QueryInterface(aIID));
  }
};

function NSGetModule(aCompMgr, aFileSpec) {
  return gModule;
}

